#include "GRCMPCGen.h"
#include <iostream>
#include <sstream>
#include <fstream>
#include <windows.h>
#include "GReaTUtils.h" // for for_each_ref()

///////////////////////////////////////////////////////////////////////////////
bool loadCodeTemplate( const std::string& module_name, const std::string& resource_name, 
					   const std::string& resource_type, std::string& codeTemplate)
{
	HMODULE hDllModule= GetModuleHandle( module_name.c_str());
	if ( 0 == hDllModule)
		return false;
//	char temp[ 100]= {'\0'};
//	DWORD ret= GetModuleFileName( hDllModule, temp, 100);
	// Find Resource
	HRSRC hRes= FindResource( hDllModule, resource_name.c_str(), resource_type.c_str());
	if ( 0 == hRes)
		return false;
	// Load Resource
	/* Daniel Fix 3-22-06 */

	HGLOBAL hTempl= LoadResource( hDllModule, hRes);
	if ( 0 == hTempl)
		return false;

	int resSize = SizeofResource(hDllModule, hRes);
	char *tempResString = new char[resSize+1];
	memcpy(tempResString, hTempl, resSize);
	tempResString[resSize] = '\0';
	// Lock Resource
	LPVOID pTempl= LockResource( hTempl);
	if ( 0 == pTempl) {
		delete [] tempResString;
		return false;
	}
	// get copy of the resource into 'script'
	//std::istringstream templ_stream( static_cast<char*>( pTempl));
	//codeTemplate= templ_stream.str();	
	codeTemplate = tempResString;
	return true;
}

std::string loadMpcTemplate( const std::string& tmplName) {
#ifndef _DEBUG
	const char app_name[]= "GRCC.dll";
#else
	const char app_name[]= "GRCCd.dll";
#endif
	std::string tmpl;
	const std::string MCP_RES= "MCP_TMPL";
	bool success= loadCodeTemplate( app_name, tmplName, MCP_RES, tmpl);
	return tmpl;
}

//////////////////////////////////////////////////////////////////////////////
class FileNameDecorator {
public:
	FileNameDecorator( const std::string& prefix, const std::string& postfix, const StrList_t& filenames, const std::string& sep= "\n") 
		: _prefix( prefix), _postfix( postfix), _fns( filenames), _sep( sep)
	{}
	std::string print()  {
		_out.clear();
		for_each_ref( _fns.begin(), _fns.end(), *this);
		return _out;
	}
	void operator()( const std::string& name) {
		_out+= _prefix + name + _postfix + _sep;
	}
private:
	std::string _prefix;
	std::string _postfix;
	StrList_t _fns;
	std::string _sep;
	std::string _out;
};

///////////////////////////////////////////////////////////////////////////////
void GRCMPCDynamicGen::generate( const std::string& fn, const std::string& metaDir, const StrList_t& metas) {
	char wrapperCode[ 5000] = {'\0'};
	std::ofstream os( (fn + ".mpc").c_str());
	std::string mpctmpl= loadMpcTemplate( "GRC_dynamic.MPC.TMPL");
	FileNameDecorator cppDecorator( "\""+ metaDir, ".cpp\"", metas);
	FileNameDecorator hDecorator( "\""+ metaDir, ".h\"", metas);
	std::string cppFiles= cppDecorator.print() + fn + ".cpp\n" + fn + ".def\n" + fn + "_genops.cpp\n";
	std::string baseProjects= ": " + fn+ "_codeinc"+ ", "+ fn+ "_codelib";
	sprintf( wrapperCode, mpctmpl.c_str(), fn.c_str(), baseProjects.c_str() ,fn.c_str(), (putPathIntoCmdLine( metaDir) + " "+ getStlDir()).c_str() , fn.c_str(), cppFiles.c_str(), hDecorator.print().c_str());
	os << wrapperCode;
	os.close();

	
}

void GRCMPCStaticGen::generate( const std::string& fn, const std::string& metaDir, const StrList_t& metas) {
	char wrapperCode[ 5000] = {'\0'};
	std::string mpctmpl= loadMpcTemplate( "GRC_static.MPC.TMPL");
	std::string baseProject= ": " + fn+ "_codeinc";
	// generate code mpc
	std::ofstream os_c( (fn + "_code.mpc").c_str());
	memset( wrapperCode, '\0', 5000);
	sprintf_s( wrapperCode, mpctmpl.c_str(), (fn+"_code").c_str(), baseProject.c_str(), (fn+"_code").c_str(), (putPathIntoCmdLine( metaDir) + " "+ getStlDir()).c_str(), (fn+".cpp").c_str(), "");
	os_c << wrapperCode;
	os_c.close();
	// generate meta mpc
	std::ofstream os_m( (fn + "_meta.mpc").c_str());
	memset( wrapperCode, '\0', 5000);
	FileNameDecorator cppDecorator( "\""+ metaDir, ".cpp\"", metas);
//	std::cout << cppDecorator.print() << std::endl;
	sprintf_s( wrapperCode, mpctmpl.c_str(), (fn+"_meta").c_str(), baseProject.c_str(), (fn+"_meta").c_str(), (putPathIntoCmdLine( metaDir) + " "+ getStlDir()).c_str(), cppDecorator.print().c_str(), "");
	os_m << wrapperCode;
	os_m.close();
	// generate order mpc
	std::ofstream os_o( (fn + "_order.mpc").c_str());
	memset( wrapperCode, '\0', 5000);
	sprintf_s( wrapperCode, mpctmpl.c_str(), (fn+"_order").c_str(), baseProject.c_str(), (fn+"_order").c_str(), (putPathIntoCmdLine( metaDir) + " "+ getStlDir()).c_str(), (fn + "_genops.cpp").c_str(), "");
	os_o << wrapperCode;
	os_o.close();
}

void GRCMPCExeUsingNoLibGen::generate( const std::string& fn, const std::string& metaDir, const StrList_t& metas) {
	std::string pn= dropPath( _projname);
	std::string greatTempDir= getGreatTempDir();
	char wrapperCode[ 5000] = {'\0'};
	std::string mpctmpl= loadMpcTemplate( "GRC_exe.MPC.TMPL");
	std::string baseProjects= ": " + fn+ "_codeinc"+ ", "+ fn+ "_codelib";
	FileNameDecorator cppDecorator( "\""+ metaDir, ".cpp\"", metas);
	std::string cppFiles= cppDecorator.print() + "\""+ pn + ".cpp\"\n\"" + pn + "_main.cpp\"\n";
#if _MSC_VER < 1300
	cppFiles+= "\""+ greatTempDir+ fn+ "_genops.cpp\"\n";
#endif
	std::string headerFiles= "\""+ pn + ".h\"\n\""+ greatTempDir+ fn+ "_cmptempl.h\"\n\""+ greatTempDir+ fn + "_userinc.h\"\n";
	// generate code mpc
	std::ofstream os( (_projname + ".mpc").c_str());
	memset( wrapperCode, '\0', 5000);
	sprintf( wrapperCode, mpctmpl.c_str(), pn.c_str(), baseProjects.c_str(), pn.c_str(), (putPathIntoCmdLine( metaDir) + " "+ getStlDir()).c_str(), "", cppFiles.c_str(), headerFiles.c_str());
	os << wrapperCode;
	os.close();
}

void GRCMPCExeUsingStaticLibGen::generate( const std::string& fn, const std::string& metaDir, const StrList_t& metas) {
	std::string pn= dropPath( _projname);
	std::string greatTempDir= getGreatTempDir();
	char wrapperCode[ 5000] = {'\0'};
	std::string mpctmpl= loadMpcTemplate( "GRC_exe.MPC.TMPL");
	std::string baseProjects= ": " + fn+ "_codeinc"+ ", "+ fn+ "_codelib";
	std::string cppFiles=  "\""+ pn + ".cpp\"\n\"" + pn + "_main.cpp\"\n";
	std::string headerFiles= "\""+ pn + ".h\"\n\""+ greatTempDir+ fn+ "_cmptempl.h\"\n\""+ greatTempDir+ fn + "_userinc.h\"\n";
	std::string libFiles= fn+ "_codes"+ " "+ fn+ "_metas";
#if _MSC_VER < 1300
	libFiles+= " "+ fn+ "_orders";
#endif
	// generate code mpc
	std::ofstream os( (_projname + ".mpc").c_str());
	memset( wrapperCode, '\0', 5000);
	sprintf( wrapperCode, mpctmpl.c_str(), pn.c_str(), baseProjects.c_str(), pn.c_str(), (putPathIntoCmdLine( metaDir) + " "+ getStlDir()).c_str(), libFiles.c_str(), cppFiles.c_str(), headerFiles.c_str());
	os << wrapperCode;
	os.close();
}
