#include "GReATParameters.h"

#include "UdmBase.h"
#include "UdmDom.h"
#include "UdmGme.h"

#include <map>
#include "UdmUtil.h"

#include "GReATConfigMeta.h"
#include "InputFileRegistry.h"

#include <direct.h>
#include "GReATUtils.h"
#include "RTTGenerator.h"

static std::string getExtension( const std::string& filename)
{
	std::string ext( filename);
	std::string::size_type findExt= filename.rfind( '.');
	if ( findExt != std::string::npos)
		ext.erase( ext.begin(), ext.begin()+ ++findExt);	// delete '.' too.
	return ext;
}

static string changeExtension(const string& filename, const string& newext)
{
	string newfilename = filename;
	size_t extPos = filename.find(".mga");
	if(extPos != string::npos)
	{
		newfilename.replace(extPos,3,newext);
	}

	return newfilename;
}

static string creatMemFilename(const string& filename)
{
	int memrand = rand();
	char buffer[20];
	_itoa_s(memrand, buffer, 10);
	string randname(buffer);
	randname += ".mem";
	string memFileName = filename;

	size_t extPos = filename.find(".xml");
	if(extPos == string::npos)
	{
		extPos = filename.find(".mga");
		if(extPos == string::npos)
		{
			extPos = filename.find(".mem");
			if(extPos == string::npos) 
				throw udm_exception("Backend NOT CLEAR for "+filename);
		}
	}
	memFileName.replace(extPos,randname.size(),randname);
	return memFileName;
}

//static void convertUdmCopyMap(const UdmUtil::copy_assoc_map& copyMap, map<int, int>& copyIdMap, bool reverse)
//{
//	for(UdmUtil::copy_assoc_map::const_iterator pos=copyMap.begin();pos!=copyMap.end();++pos)
//	{
//		if(!reverse)
//			copyIdMap[(*pos).first.uniqueId()] = (*pos).second.uniqueId();
//		else
//			copyIdMap[(*pos).second.uniqueId()] = (*pos).first.uniqueId();
//	}
//}

void dummy(void) 
{
} // Dummy function for initialization

//============================================
MetaInfo::MetaInfo(const Udm::DataNetwork *meta)
{
	theUmlDiagram = new Uml::Diagram;
	*theUmlDiagram = Uml::Diagram::Cast(meta->GetRootObject());
	
	udmDataDiagram = new Udm::UdmDiagram;
	udmDataDiagram->dgr = theUmlDiagram;
	udmDataDiagram->init = dummy;
}

MetaInfo::~MetaInfo()
{
  if (udmDataDiagram)
    delete udmDataDiagram, udmDataDiagram = 0;
  if (theUmlDiagram)
    delete theUmlDiagram, theUmlDiagram = 0;
}

//========================================
GReATParameters::GReATParameters():
	m_backendMap(0),
	udmDN(Uml::diagram),
	cfg_usageInfo(0),
	cfg_runtimeInfo(0),
	printMsg(""),
	currentWD(""),
	clUmlDgr(0),
	clUdmDataDgr(0),
	clProject(0),
	use_mem(false)
{
	sdnConfig = new Udm::SmartDataNetwork(GReATConfig::diagram);
	sdnTransformation = new Udm::SmartDataNetwork(GR::diagram);
}
//=================================
GReATParameters::~GReATParameters()
{
	closeNoSave();
	for( vector<Uml::Class*>::iterator itr = rootClasses.begin(); itr != rootClasses.end(); ++itr)
		delete *itr;
}

//=================================
void GReATParameters::getUdmObjs(const string &metaName, vector<string> udmObjPath, \
								 set<Udm::Object>& parentObjs, set<Udm::Object>& udmObjSet)
{
	vector<string>::iterator it = udmObjPath.begin();
	string currPath = *it;
	udmObjPath.erase(it);
	
	set<Udm::Object> objSet;

	for(set<Udm::Object>::iterator i=parentObjs.begin();i!=parentObjs.end();++i)
	{
		Udm::Object pObj = *i;
		
		int strPos = currPath.find_first_of("{");
		const string &clsName = currPath.substr(0, strPos);
		const string &condition = currPath.substr(strPos+1, currPath.length()-clsName.length()-2);
		const string &attrName = condition.substr(0, condition.find_first_of("=")); 
		const string &attrVal = condition.substr(condition.find_first_of("=")+1,condition.length()-attrName.length()-1); 

		Uml::Class objClass;
	//	m_backendMap->GetUmlClass(metaName, clsName, objClass);

		set<Udm::Object> childObjs = pObj.GetChildObjects(objClass);
		for(set<Udm::Object>::iterator ii=childObjs.begin();ii!=childObjs.end();++ii)
		{
			Udm::Object currObj = *ii;
			
			bool isTrue = true;
			if(strPos!=-1) isTrue = isDesiredObj(currObj,attrName,attrVal);
				
			if(isTrue)
			{
				if(udmObjPath.empty()) udmObjSet.insert(currObj);
				else objSet.insert(currObj);
			}
		}
	}

	if(udmObjPath.empty()) return;
	else 
	{
		getUdmObjs(metaName, udmObjPath, objSet, udmObjSet);
	}
}

bool GReATParameters::isDesiredObj(Udm::Object obj, const string &attrName, const string &attrVal)
{
	__int64 intVal;
	string strVal;
	double realVal;
	bool boolVal;
	if(obj.GetIntValue(attrName, intVal))
	{
		if( intVal==atoi(attrVal.c_str())) return true;	
		else return false;	
	}
	else if(obj.GetStrValue(attrName, strVal))
	{
		if(strVal==attrVal) return true;
		else return false;
	}
	else if(obj.GetRealValue(attrName, realVal))
	{
		if( realVal==atof(attrVal.c_str())) return true;	
		else return false;
	}
	else if(obj.GetBoolValue(attrName, boolVal))
	{
		if(boolVal)
		{
			if(_stricmp(attrVal.c_str(), "true") == 0) return true;
			else return false;
		}
		else
		{
			if(_stricmp(attrVal.c_str(), "false") == 0) return true;
			else return false;
		}
	}
	else return false;
}

void GReATParameters::buildPassAlongPkts(PassAlongPacket& newArg,InputPort2UdmObjMap& inputPkts,
										 map<GR::InputPort, set<Udm::Object> >::iterator mpos)
{
	mpos++;
	if(mpos == inputPkts.end()) 
	{
		m_startArgs.push_back(newArg);
		return;
	}

	GR::InputPort port = (*mpos).first;
	set<Udm::Object> objs = (*mpos).second;
	for(set<Udm::Object>::iterator objPos=objs.begin();objPos!=objs.end();objPos++)
	{
		Udm::Object	obj=*objPos;
		newArg[port] = obj;
		buildPassAlongPkts(newArg,inputPkts,mpos);
	}
}

//============IGReATParameters
//=================================
void GReATParameters::getCfgDataNetwork(const std::string& fname)
{
	if(_stricmp(getExtension(fname).c_str(),"mga") == 0)
	{
		Udm::SmartDataNetwork openCfg(GReATConfig::diagram);
		try
		{
			openCfg.OpenExisting(fname, "", Udm::CHANGES_PERSIST_ALWAYS);
		}
		catch(udm_exception)
		{
			throw udm_exception("Udm exception: "+fname+" could not open.");
		}

		cfgmemfile = creatMemFilename(fname);
		sdnConfig->CreateNew(cfgmemfile,"",GReATConfig::RootFolder::meta,Udm::CHANGES_PERSIST_ALWAYS);
		*sdnConfig = openCfg;

		openCfg.CloseNoUpdate();
	}
	else sdnConfig->OpenExisting(fname, "", Udm::CHANGES_PERSIST_ALWAYS);

	currentWD = getFullPath(currentWD, fname);
	size_t pos = currentWD.rfind( "\\");
	if ( pos != string::npos)
		currentWD.erase(pos, currentWD.length());

	_chdir(currentWD.c_str());
}

void GReATParameters::open(const std::string& fname)
{
	getCfgDataNetwork(fname);
	open();
}

void GReATParameters::open(const string &fname, const vector<string> &files)
{
	getCfgDataNetwork(fname);
	GReATConfig::RootFolder cfgrf = GReATConfig::RootFolder::Cast(sdnConfig->GetRootObject());

	GReATConfig::RuntimeInformation rt;
	set<GReATConfig::RuntimeInformation> rts = cfgrf.RuntimeInformation_kind_children();
	if(rts.empty())
	{
		rt = GReATConfig::RuntimeInformation::Create(cfgrf);
	}
	else rt = *(rts.begin());

	map<string, GReATConfig::InputFile> tempInputFileMap;
	set<GReATConfig::InputFile> inputfiles = rt.InputFile_kind_children();
	for(set<GReATConfig::InputFile>::iterator j=inputfiles.begin();j!=inputfiles.end();++j)
	{
		tempInputFileMap[(string)((*j).FileID())] = *j;
	}

	InputFileRegistry inputfile;

	GReATConfig::UsageInformation m_usageInfo = cfgrf.UsageInformation_child();
	if(m_usageInfo==Udm::null) throw udm_exception("No file usage infomation found!");
	GReATConfigUsage temp_cfg_usageInfo(m_usageInfo);
	vector<string> fileIDs;
	temp_cfg_usageInfo.getFileUsageFileIDs(fileIDs);
	
	for(vector<string>::const_iterator i=files.begin(); i!=files.end(); ++i)
	{
		std::string fid, filename, copyname, doctypename;
		inputfile.parseFileExpr(*i, fid, filename, copyname, doctypename);
		//Check whether fid is in fileIDs;
		bool existId = false;
		for(vector<string>::const_iterator vi=fileIDs.begin();vi!=fileIDs.end();++vi)
		{
			if((*vi) == fid)
			{
				existId = true;
				break;
			}
		}
		if(!existId) throw udm_exception("FileID \""+fid+"\" could not be found in file usages.");

		GReATConfig::InputFile newcfginput;
		map<string, GReATConfig::InputFile>::iterator pos = tempInputFileMap.find(fid);
		if(pos==tempInputFileMap.end())	newcfginput = GReATConfig::InputFile::Create(rt);
		else newcfginput = (*pos).second;
		newcfginput.FileID() = fid;
		newcfginput.FileName() = filename;
		newcfginput.CopyName() = copyname;
		if(!doctypename.empty())
		{
			GReATConfig::FileUsage fileusage;
			cfg_usageInfo->getFileUsage(fid, fileusage);
			fileusage.DTDFileName() = doctypename;
		}
	}

	open();
	tempInputFileMap.clear();
}

void GReATParameters::open()
{
	GReATConfig::RootFolder cfgRootFolder = GReATConfig::RootFolder::Cast(sdnConfig->GetRootObject());
	
	// Get sdnTransformation
	GReATConfig::MetaInformation m_metaInfo = cfgRootFolder.MetaInformation_child();
	if(m_metaInfo == Udm::null) throw udm_exception("No transformation rule file and meta files specified.");

	//const string &tFileName = m_metaInfo.TransformationFile();
	string tFileName = getFullPath(currentWD, (string)m_metaInfo.TransformationFile());
	// FIXME: specify GREAT_PATH\etc\gr.xsd here
	sdnTransformation->OpenExisting(tFileName, "", Udm::CHANGES_PERSIST_ALWAYS);

	string prjFile = getFullPath(currentWD, (string)m_metaInfo.UdmProjectFile());
	string extname = getExtension(prjFile);
	if(_stricmp(extname.c_str(), "udm")==0)
	{
		udmPrj.OpenExisting(prjFile, Udm::CHANGES_PERSIST_ALWAYS);
		vector<Udm::DataNetwork*> prDns= udmPrj.GetDataNetworks();
		for(vector<Udm::DataNetwork*>::const_iterator ix= prDns.begin(); ix!= prDns.end(); ++ix)
		{
			const string &udmmetaname = Uml::Diagram::Cast( (*ix)->GetRootObject()).name();
			MetaInfo *udmDataDgr = new MetaInfo(*ix);
			udmDiagramMap[udmmetaname] = udmDataDgr;
			umlDiagramMap[udmmetaname] = udmDataDgr->theUmlDiagram;
		}
		//m_umlDiagram = udmDiagramMap.begin()->second->theUmlDiagram;
		if (udmPrj.IsCrossMetaSpecifier())
		{
		  Udm::DataNetwork& crn =  udmPrj.GetCrossMetaNetwork();
		  m_crossMetaName = Uml::Diagram::Cast( crn.GetRootObject()).name();
		}
	}	
	else if(_stricmp(extname.c_str(), "xml")==0)
	{
		udmDN.OpenExisting(prjFile, "", Udm::CHANGES_PERSIST_ALWAYS);
		MetaInfo *udmDataDgr = new MetaInfo(&udmDN);
		const string mname = Uml::Diagram::Cast(udmDN.GetRootObject()).name();
		udmDiagramMap[mname] = udmDataDgr;
		umlDiagramMap[mname] = udmDataDgr->theUmlDiagram;
	}
	else throw udm_exception("Incorrect udm project file.");
		
	m_ttLastWrite = tFileName;
	m_getUdmDirName = prjFile.substr(0,prjFile.find_last_of("\\"));
	
	//Get UsageInformation
	GReATConfig::UsageInformation m_usageInfo = cfgRootFolder.UsageInformation_child();
	if(m_usageInfo==Udm::null) throw udm_exception("No file usage infomation found!");
	cfg_usageInfo = new GReATConfigUsage(m_usageInfo);
	//Get StartRule
	m_rootFolder = GR::RootFolder::Cast( sdnTransformation->GetRootObject());
	cfg_usageInfo->getStartRule(m_rootFolder, m_startRule);
	cfg_usageInfo->getTransformationInputs(mappings);

	//Get RuntimeInformation
	GReATConfig::RuntimeInformation m_runtimeInfo = cfgRootFolder.RuntimeInformation_child();
	if(m_runtimeInfo==Udm::null) throw udm_exception("No runtime infomation found!");
	cfg_runtimeInfo = new GReATConfigRuntime(m_runtimeInfo);
	cfg_runtimeInfo->getInputFiles(inputFiles);

	//Get Temporary packages
	cfg_usageInfo->getTemporaryFileUsages(tempPkgs);

	load();
}

//=================================
void GReATParameters::load()
{
	map<std::string, int> inDNs, outDNs;
	map<int, map<int, int> > udmcopyMaps;

	string saved_dtd_path = UdmDom::DomDataNetwork::DTDPath;
	
	map<string, Udm::DataNetwork *> dataNetworkMap;
	vector<Udm::StaticDataNetworkSpecifier> vsnds;
	for(set<GReATConfig::InputFile>::const_iterator i=inputFiles.begin();i!=inputFiles.end();++i)
	{
		const GReATConfig::InputFile &inputFile = *i;
		const string& fileId = inputFile.FileID();
		string filename = getFullPath(currentWD, (string)inputFile.FileName());

		if(fileId.empty() || filename.empty()) continue;
		
		string metaname;
		if(!cfg_usageInfo->getMetaName(fileId, metaname))
			throw udm_exception("No meta found for "+filename);
		
		map<string, MetaInfo*>::iterator pos = udmDiagramMap.find(metaname);
		if(pos==udmDiagramMap.end()) 	
			throw udm_exception("No meta file found for "+filename);
		MetaInfo *udmDgr = (*pos).second;

		string rootClassName;
		if(!cfg_usageInfo->getRootClassName(fileId, rootClassName))
			throw udm_exception("Error with RootClassName for "+filename);
		
		string filemode;
		if(!cfg_usageInfo->getMode(fileId, filemode))
			throw udm_exception("Error with FileID for "+filename);
	
		string dtdfile; 
		cfg_usageInfo->getDTDFileName(fileId, dtdfile);
		string dtdfilefullpath;
		if(!dtdfile.empty())
		{
			dtdfilefullpath = getFullPath(currentWD,dtdfile);
			string dtdpath(dtdfilefullpath);
			size_t pos = dtdpath.rfind( "\\");
			if ( pos != string::npos) dtdpath.erase(pos, dtdpath.length());
			UdmDom::DomDataNetwork::DTDPath = dtdpath;
		}
		
		string memFileName = creatMemFilename(filename);
//		const Uml::Class & safe_type = Uml::SafeTypeContainer::GetSafeType(Uml::classByName(*(udmDgr->theUmlDiagram), rootClassName));
//		const Uml::Class & safe_type = Uml::SafeTypeContainer::GetSafeType(Uml::classByPath(*(udmDgr->theUmlDiagram), rootClassName));
		const Uml::Class& safe_type = getRootClass(*(udmDgr->theUmlDiagram), rootClassName);

		Udm::SmartDataNetwork* workDN = new Udm::SmartDataNetwork(*(udmDgr->udmDataDiagram));
		
		bool useMem = false;
		cfg_usageInfo->getRunInMemory(fileId, useMem);

		if(useMem)
		{
///////////////////////////////////////////////////////////////////////////////////
			workDN->CreateNew(memFileName,"",safe_type,Udm::CHANGES_PERSIST_ALWAYS);
		
			tempFiles.insert(memFileName);

			if(_stricmp(filemode.c_str(),"w")!=0)
			{
				Udm::SmartDataNetwork sdnFile(*(udmDgr->udmDataDiagram));
				sdnFile.OpenExisting(filename, "", Udm::CHANGES_PERSIST_ALWAYS);
				try
				{
//					*workDN = sdnFile;   //udmcopy
					UdmUtil::copy_assoc_map copyMap;
					UdmUtil::CopyObjectHierarchy(sdnFile.GetRootObject().__impl(), workDN->GetRootObject().__impl(),workDN,copyMap);
					map<int, int> copyIdMap;
					RTTGenerator::convertUdmCopyMap(copyMap, copyIdMap, true);
					udmcopyMaps[workDN->testdn()->uniqueId()]=copyIdMap;
				}
				catch(udm_exception)
				{
					throw udm_exception("Udm exception: UdmCopy fails for "+filename);
				}

				sdnFile.CloseNoUpdate();
			}
//////////////////////////////////////////////////////////////////////////////////////////
		}
		else
		{
			if(_stricmp(filemode.c_str(),"w")!=0) 
			{
				workDN->OpenExisting(filename, "", Udm::CHANGES_PERSIST_ALWAYS);
			}
			else
			{
				if(!dtdfile.empty() && _stricmp(getExtension(filename).c_str(),"xml") == 0)
				{
					UdmDom::DomDataNetwork::DTDPath = "";
					workDN->CreateNew(filename,dtdfilefullpath,safe_type,Udm::CHANGES_PERSIST_ALWAYS);	
				}
				else
					workDN->CreateNew(filename,metaname,safe_type,Udm::CHANGES_PERSIST_ALWAYS);
			}
		}
		
		std::string filename_RTT = filename;
		if(_stricmp(filemode.c_str(),"rwc")==0)		
			filename_RTT = getFullPath(currentWD, (string)inputFile.CopyName());

		if(_stricmp(filemode.c_str(),"w")!=0)
			inDNs[filename_RTT] = workDN->testdn()->uniqueId();
		else
			outDNs[filename_RTT] = workDN->testdn()->uniqueId();

		sfile2workdn[inputFile] = workDN;
		
		dataNetworkMap[fileId] = workDN->testdn();	 //used for general backend

		Udm::StaticDataNetworkSpecifier snds(metaname, workDN);
		vsnds.push_back(snds);

		UdmDom::DomDataNetwork::DTDPath = saved_dtd_path;
	}
	
	//Initiate RTT here:
	std::string rttname("RTT");
	map<std::string, int>::iterator dn_it;
	for(dn_it=inDNs.begin();dn_it!=inDNs.end();++dn_it)
	{
		rttname.append("_");
		std::string inFile = (*dn_it).first;
		rttname.append(dropExtension(dropPath(inFile)));
	}
	for(dn_it=outDNs.begin();dn_it!=outDNs.end();++dn_it)
	{
		rttname.append("_");
		std::string outFile = (*dn_it).first;
		rttname.append(dropExtension(dropPath(outFile)));
	}
	RTTGenerator::Instance(getFullPath(currentWD,rttname).append(".mga"), rttname);

	for(dn_it=inDNs.begin();dn_it!=inDNs.end();++dn_it)
	{
		RTTGenerator::Instance()->generateDataNetwork((*dn_it).second, (*dn_it).first, dropPath((*dn_it).first));
	}
	for(dn_it=outDNs.begin();dn_it!=outDNs.end();++dn_it)
	{
		RTTGenerator::Instance()->generateDataNetwork((*dn_it).second, (*dn_it).first, dropPath((*dn_it).first));
	}
	
	for(map<int, map<int, int> >::iterator map_it=udmcopyMaps.begin(); map_it!=udmcopyMaps.end();++map_it)
		RTTGenerator::Instance()->addUdmCopyMap((*map_it).first, (*map_it).second);

	//Add the initiation of Temporary package here, supported after 11/18/04, GReAT1.3.4
//	vector<GReATConfig::TemporaryFileUsage> tempPkgs;
//	cfg_usageInfo->getTemporaryFileUsages(tempPkgs);

/*	GReATConfig::UsageInformation usgInfo = (GReATConfig::RootFolder::Cast(sdnConfig->GetRootObject())).UsageInformation_child();
	vector<GReATConfig::TemporaryFileUsage> tempPkgs;

	vector< GReATConfig::TemporaryFileUsage> tfus = usgInfo.TemporaryFileUsage_kind_children();
	for( std::vector< GReATConfig::TemporaryFileUsage>::const_iterator ik= tfus.begin(); ik!= tfus.end(); ++ik) {
		const GReATConfig::TemporaryFileUsage& currTFU= *ik;
		if ( currTFU.type() == GReATConfig::TemporaryFileUsage::meta) {
			tempPkgs.push_back( currTFU);
		}
	}
*/
	map<string, Object> globalObjs;
	
	vector< GReATConfig::TemporaryFileUsage>::const_iterator ti=tempPkgs.begin();
	for(ti;ti!=tempPkgs.end();++ti)
	{
		GReATConfig::TemporaryFileUsage tempFile = *ti;
		string metaname = tempFile.MetaName();
		string rootname = tempFile.RootClassName();
		map<string, MetaInfo*>::iterator pos = udmDiagramMap.find(metaname);
		if(pos==udmDiagramMap.end()) 	
			throw udm_exception("No meta file found for Temporary meta: "+metaname);
		MetaInfo *udmDgr = (*pos).second;

		string tempfn = creatMemFilename((string)(metaname+".mem"));
//		const Uml::Class & safe_type = Uml::SafeTypeContainer::GetSafeType(Uml::classByName(*(udmDgr->theUmlDiagram), rootname));
//		const Uml::Class & safe_type = Uml::SafeTypeContainer::GetSafeType(Uml::classByPath(*(udmDgr->theUmlDiagram), rootname));
		const Uml::Class& safe_type = getRootClass(*(udmDgr->theUmlDiagram), rootname);

		Udm::SmartDataNetwork* tempDN = new Udm::SmartDataNetwork(*(udmDgr->udmDataDiagram));
		tempDN->CreateNew(tempfn,"",safe_type,Udm::CHANGES_PERSIST_ALWAYS);
		tempFiles.insert(tempfn);

		Udm::Object rootObj = tempDN->testdn()->GetRootObject();
		globalObjs[(string)(metaname+"::"+rootname)] = rootObj;

		tempDNs.push_back(tempDN);
		Udm::StaticDataNetworkSpecifier snds(metaname, tempDN);
		vsnds.push_back(snds);
	}
	//////////////////////////////////////////////

	m_backendMap = new GeneralBackend(umlDiagramMap,globalObjs);

	//Create UDM static project
	if(udmPrj.IsCrossMetaSpecifier()) 
	{
		clUmlDgr = new Uml::Diagram;
		*clUmlDgr = Uml::Diagram::Cast(udmPrj.GetCrossMetaNetwork().GetRootObject());

		clUdmDataDgr = new Udm::UdmDiagram;
		clUdmDataDgr->dgr = clUmlDgr;
		clUdmDataDgr->init = dummy;

		clProject = new Udm::StaticUdmProject(vsnds, *clUdmDataDgr);
	}

	init(dataNetworkMap);
}
//=================================
void GReATParameters::init(map<string, Udm::DataNetwork *> &dataNetworkMap)
{
	m_startArgs.clear();
	//Build the input packets
	InputPort2UdmObjMap inputPkts;
	for(set<GReATConfig::TransformationInput>::const_iterator mi=mappings.begin();mi!=mappings.end();mi++)
	{
		const GReATConfig::TransformationInput &mapping = *mi;
		const string &fileId = mapping.FileID();
		string meta;
		if(!cfg_usageInfo->getMetaName(fileId, meta))
			throw udm_exception("GReATParameters::init()");
		
		vector<string> udmObjPath = mapping.ObjectPath();
		udmObjPath.erase(udmObjPath.begin());           //First is supposed always be the Root
		
		Udm::Object ro;
		map<string, Udm::DataNetwork *>::iterator pos = dataNetworkMap.find(fileId);
		if(pos!=dataNetworkMap.end())
		{
			ro = ((*pos).second)->GetRootObject();
		}
		else throw udm_exception("GReATParameters::init()");

//		if(!m_backendMap->GetRootObject(fileId, ro)) 
//			throw udm_exception("GReATParameters::init()");

		set<Udm::Object> parentObjs, inputObjs;
		parentObjs.insert(ro);
		if(!udmObjPath.empty()) getUdmObjs(meta, udmObjPath, parentObjs, inputObjs);
		else inputObjs.insert(ro);

		set<GR::InputPort> inputPorts = m_startRule.InputPort_kind_children();
		for(set<GR::InputPort>::iterator ipos=inputPorts.begin();ipos!=inputPorts.end();ipos++)
		{
			GR::InputPort inputPort = *ipos;
			if((string)inputPort.name()==(string)mapping.PortName())
			{
				inputPkts[inputPort] = inputObjs;
				break;
			}
		}
	}
	
	//11/25/03: throw the exception here:
//	ASSERT(!inputPkts.empty());
	if(inputPkts.empty())
		throw udm_exception("The input packets is empty. Please check the input ports.");
	//===================================
	map<GR::InputPort, set<Udm::Object> >::iterator mpos=inputPkts.begin();		
	GR::InputPort port = (*mpos).first;
	set<Udm::Object> objs = (*mpos).second;
	for(set<Udm::Object>::iterator objPos=objs.begin();objPos!=objs.end();objPos++)
	{
		PassAlongPacket newArg;
		Udm::Object	obj=*objPos;
		newArg[port] = obj;
		buildPassAlongPkts(newArg,inputPkts,mpos);
	}
}

//=================================
IGReATParameters::SaveResult GReATParameters::saveProgram()
{
	return saveInternalBackend();
}

IGReATParameters::SaveResult GReATParameters::saveInternalBackend(bool doException)
{
	bool halfsucc = false;
	for(map<GReATConfig::InputFile, Udm::SmartDataNetwork *>::const_iterator i=sfile2workdn.begin();\
		i!=sfile2workdn.end();++i)
	{
		const GReATConfig::InputFile &inputfile = (*i).first;
		Udm::SmartDataNetwork *memFile = (*i).second;

		const string& fileId = inputfile.FileID();

		const string& fname = inputfile.FileName();
		if(fileId.empty() || fname.empty()) continue;
		string filename = getFullPath(currentWD, fname);

		string filemode;
		cfg_usageInfo->getMode(fileId, filemode);
		if(_stricmp(filemode.c_str(),"r")==0) continue;

		string cname;
		if(_stricmp(filemode.c_str(),"rwc")==0)
		{
			cname = inputfile.CopyName();
			if(cname.empty()) 
			{
				throw udm_exception("No copy file name found for "+filename);		
			}
		}
		if(_stricmp(filemode.c_str(),"w")==0 || _stricmp(filemode.c_str(),"rw")==0)
			cname = inputfile.FileName();

		string copyname = getFullPath(currentWD, cname);
		
		if(!printMsg.empty()) printMsg += ", \n";
		printMsg += copyname;
		if(_stricmp(filemode.c_str(),"w")==0 || _stricmp(filemode.c_str(),"rwc")==0)
			printMsg += " is created";
		if(_stricmp(filemode.c_str(),"rw")==0)
			printMsg += " is updated";

///////////////////////////////////////////////////////
		bool useMem = false;
		cfg_usageInfo->getRunInMemory(fileId, useMem);
		if(!useMem)
			if(_stricmp(filemode.c_str(),"rw")==0 || _stricmp(filemode.c_str(),"w")==0) continue;
//////////////////////////////////////////////////////

		remove(copyname.c_str());   //delete the file with the same name

		string metaname;
		cfg_usageInfo->getMetaName(fileId, metaname);

		string rootClassName;
		cfg_usageInfo->getRootClassName(fileId, rootClassName);
			
		map<string, MetaInfo*>::iterator pos = udmDiagramMap.find(metaname);
		if(pos==udmDiagramMap.end()) 	
			throw udm_exception("No meta file found for "+filename);
		MetaInfo *udmDgr = (*pos).second;

//		const Uml::Class & safe_type = Uml::SafeTypeContainer::GetSafeType(Uml::classByName(*(udmDgr->theUmlDiagram), rootClassName));
//		const Uml::Class & safe_type = Uml::SafeTypeContainer::GetSafeType(Uml::classByPath(*(udmDgr->theUmlDiagram), rootClassName));
		const Uml::Class& safe_type = getRootClass(*(udmDgr->theUmlDiagram), rootClassName);


		string dtdfile;
		if(!cfg_usageInfo->getDTDFileName(fileId, dtdfile))
			throw udm_exception("Error with DTD for "+filename);

		Udm::SmartDataNetwork sdnFile_copy(*(udmDgr->udmDataDiagram));
		if(!dtdfile.empty() && _stricmp(getExtension(copyname).c_str(),"xml") == 0)
		{
			dtdfile = getFullPath(currentWD,dtdfile);
			sdnFile_copy.CreateNew(copyname,dtdfile,safe_type,Udm::CHANGES_PERSIST_ALWAYS);				
		}
		else
		{
			sdnFile_copy.CreateNew(copyname,metaname,safe_type,Udm::CHANGES_PERSIST_ALWAYS);
		} 

		try
		{
			//sdnFile_copy = *memFile;
			UdmUtil::copy_assoc_map copyMap;
			UdmUtil::CopyObjectHierarchy(memFile->GetRootObject().__impl(), sdnFile_copy.GetRootObject().__impl(), sdnFile_copy.testdn(),copyMap);
			map<int, int> copyIdMap;
			RTTGenerator::convertUdmCopyMap(copyMap, copyIdMap, false);
			RTTGenerator::Instance()->addUdmCopyMap(memFile->testdn()->uniqueId(), copyIdMap);
			sdnFile_copy.CloseWithUpdate();
		}
		catch(udm_exception)
		{
			if(!doException) throw udm_exception("Udm exception: UdmCopy fails for "+copyname);

			remove(copyname.c_str());
			if(dtdfile.empty()) return FAILED;
			
			string cxmlfile = changeExtension(copyname, "xml");
			Udm::SmartDataNetwork sdnXMLFile(*(udmDgr->udmDataDiagram));
			sdnXMLFile.CreateNew(cxmlfile,dtdfile,safe_type,Udm::CHANGES_PERSIST_ALWAYS);
			sdnXMLFile = *memFile;
			sdnXMLFile.CloseWithUpdate();
			halfsucc = true;
		}
	}
	if(halfsucc) return HALF_SUCCESS;
	return SUCCESS;
}

//=================================
void GReATParameters::reloadProgram()
{
	closeProgram();
	load();

}
//=================================

void GReATParameters::closeProgramNoSave()
{
	if (sfile2workdn.size()) 
	{
		closeDN();
	}
	if(tempDNs.size()) 
	{
		closeTempDN();
	}
	if (m_backendMap)
		delete m_backendMap, m_backendMap =0;

	//delete temporary files
	for(set<string>::const_iterator i=tempFiles.begin();i!=tempFiles.end();++i)
	{
		const string &tempFile = *i;
		if(!tempFile.empty()) remove(tempFile.c_str());	
	}
	tempFiles.clear();
}

//=================================

void GReATParameters::closeProgram()
{
	saveProgram();
	closeProgramNoSave();
}

//=================================
void GReATParameters::closeDN()
{
	for(map<GReATConfig::InputFile, Udm::SmartDataNetwork *>::iterator j=sfile2workdn.begin();j!=sfile2workdn.end();++j)
	{
		Udm::SmartDataNetwork *memDn = (*j).second;
		const GReATConfig::InputFile &inputfile = (*j).first;
		const string& fileId = inputfile.FileID();
		string filemode;
		cfg_usageInfo->getMode(fileId, filemode);

		bool useMem = false;
		cfg_usageInfo->getRunInMemory(fileId, useMem);
		if(!useMem) 
		{
			if(_stricmp(filemode.c_str(),"r")==0 || _stricmp(filemode.c_str(),"rwc")==0) 
				memDn->CloseNoUpdate();
			else
				memDn->CloseWithUpdate();
		}
		else memDn->CloseNoUpdate();
		
		delete memDn;
	}
	sfile2workdn.clear();
}

void GReATParameters::closeTempDN()
{
	vector<Udm::SmartDataNetwork*>::iterator i=tempDNs.begin();
	for(i;i!=tempDNs.end();++i)
	{
		Udm::SmartDataNetwork* tempdn = *i;
		tempdn->CloseNoUpdate();
		delete tempdn;
	}
	tempDNs.clear();
}

//=================================
void GReATParameters::close()
{
	saveProgram();
	closeNoSave();
}
//==============================================
void GReATParameters::closeNoSave()
{
//	deleteUdmProject();
	RTTGenerator::Instance()->close();
	
	if(sdnConfig && sdnConfig->isOpen()) sdnConfig->CloseNoUpdate();
	if(sdnTransformation && sdnTransformation->isOpen()) sdnTransformation->CloseNoUpdate();

	closeDataNetworks();
	deleteHeaps();
	deleteTempFiles();
	this->mappings.clear();
	this->inputFiles.clear();
	this->tempPkgs.clear();

	if(udmDN.isOpen()) udmDN.CloseNoUpdate();	
}
//==============================================
std::string GReATParameters::getTransfLastWrittenFileName()
{
	return m_ttLastWrite;
}
//==============================================
std::string GReATParameters::getUdmDirName()
{
  	return m_getUdmDirName;
}
// //==============================================
// std::set<string> GReATParameters::getMetaNames() 
// {
//   std::set<string> tmp;
//   map<string, MetaInfo*>::const_iterator it = udmDiagramMap.begin();
//   for (; it != udmDiagramMap.end(); ++it)
//     tmp.insert(it->first);
// 
//   if (!m_crossMetaName.empty())
//     tmp.insert(m_crossMetaName);
//   return tmp;
// 
// }
//==============================================
StrUmlDiagrams_t  GReATParameters::getMetas()
{
  StrUmlDiagrams_t tmp;
	// !!VIZA order is important, cross links must come first
  if (udmPrj.IsCrossMetaSpecifier())
  {
	  Udm::DataNetwork& crn =  udmPrj.GetCrossMetaNetwork();
	  std::string crossMetaName = Uml::Diagram::Cast( crn.GetRootObject()).name();
	  tmp.push_back( std::make_pair( crossMetaName, Uml::Diagram::Cast( crn.GetRootObject() )));
  }
  map<string, MetaInfo*>::const_iterator it = udmDiagramMap.begin();
  for (; it != udmDiagramMap.end(); ++it)
	  tmp.push_back( std::make_pair(it->first, *(it->second->udmDataDiagram->dgr)));
  return tmp;
}
//==============================================
/*
void GReATParameters::getUserCodeInfo(StrVec& incFiles,
                                      StrVec& libFiles,
                                      StrVec& incDirs,
                                      StrVec& libDirs)
{
 	GReATConfig::RootFolder cfgrf = GReATConfig::RootFolder::Cast(sdnConfig->GetRootObject());


	const set<GReATConfig::MetaInformation>& mis = cfgrf.MetaInformation_kind_children();
  if (!mis.empty())
  {
    const GReATConfig::MetaInformation& mi = *(mis.begin());
  
    GReATConfigMeta gmi(mi);

    tokenize_csv_lines(gmi.incFiles(), incFiles);
    tokenize_csv_lines(gmi.libFiles(), libFiles);
    tokenize_csv_lines(gmi.incDirs(), incDirs);
    tokenize_csv_lines(gmi.libDirs(), libDirs);
  }
}
*/
//==============================================
void GReATParameters::closeDataNetworks()
{
	for(map<GReATConfig::InputFile, Udm::SmartDataNetwork *>::const_iterator i=sfile2workdn.begin();
				i!=sfile2workdn.end();++i)
	{
		Udm::SmartDataNetwork *odn = (*i).second;
		if(odn && odn->isOpen()) 
		{
			const GReATConfig::InputFile &inputfile = (*i).first;
			const string& fileId = inputfile.FileID();
			string filemode;
			cfg_usageInfo->getMode(fileId, filemode);

			bool useMem = false;
			cfg_usageInfo->getRunInMemory(fileId, useMem);
			if(!useMem) 
			{
				if(_stricmp(filemode.c_str(),"r")==0 || _stricmp(filemode.c_str(),"rwc")==0) 
					odn->CloseNoUpdate();
				else
					odn->CloseWithUpdate();
			}
			else odn->CloseNoUpdate();
		}
	}

	//close the temporary datanetworks for global object	
	for(vector<Udm::SmartDataNetwork*>::iterator j=tempDNs.begin();j!=tempDNs.end();++j)
	{
		Udm::SmartDataNetwork* tempdn = *j;
		if(tempdn && tempdn->isOpen()) tempdn->CloseNoUpdate();
	}
}

void GReATParameters::deleteUdmProject()
{
	if(clProject)
		delete clProject, clProject = 0;

	if(clUdmDataDgr)
		delete clUdmDataDgr, clUdmDataDgr = 0;

	if(clUmlDgr)
		delete clUmlDgr, clUmlDgr = 0;

}

void GReATParameters::deleteHeaps()
{
	for(map<GReATConfig::InputFile, Udm::SmartDataNetwork *>::const_iterator i=sfile2workdn.begin();
				i!=sfile2workdn.end();++i)
	{
		delete (*i).second;
	}
	sfile2workdn.clear();

	///delete temporary datanetworks for global object
	for(vector<Udm::SmartDataNetwork*>::iterator ti=tempDNs.begin();ti!=tempDNs.end();++ti)
	{
		delete (*ti);
	}
	tempDNs.clear();
	//////////////////////////////////////////
	
	if (m_backendMap)
		delete m_backendMap, m_backendMap =0;

	for(map<string, MetaInfo*>::iterator j=udmDiagramMap.begin();j!=udmDiagramMap.end();++j)
	{
		MetaInfo *metainfo = (*j).second;
		delete metainfo;
	}
	udmDiagramMap.clear();

	umlDiagramMap.clear();

	m_startArgs.clear();
	inputFiles.clear();
	mappings.clear();

	////////////////////
	tempPkgs.clear();


  	if (cfg_usageInfo)
		delete cfg_usageInfo, cfg_usageInfo = 0;
	if (cfg_runtimeInfo)
		delete cfg_runtimeInfo, cfg_runtimeInfo = 0;
	if (sdnTransformation)
		delete sdnTransformation, sdnTransformation = 0;
	if (sdnConfig)
		delete sdnConfig, sdnConfig = 0;

	deleteTempFiles();
}

void GReATParameters::deleteTempFiles()
{
	for(set<string>::const_iterator i=tempFiles.begin();i!=tempFiles.end();++i)
	{
		const string &tempFile = *i;
		if(!tempFile.empty()) remove(tempFile.c_str());	
	}
	tempFiles.clear();
	if(!cfgmemfile.empty()) remove(cfgmemfile.c_str());
}


//=================================
void GReATParameters::getRootFolder(GR::RootFolder& rootFolder)
{
  rootFolder = m_rootFolder;
}
//=================================
void GReATParameters::getGeneralBackend(GeneralBackendPtr& generalBackend)
{
  generalBackend = m_backendMap;
}
//=================================
void GReATParameters::getTransformation(UdmDataNetworkPtr& transformation)
{
  transformation = sdnTransformation->testdn();
}
//=================================
void GReATParameters::getStartRule(GR::RewritingRule& rule)
{
  rule = m_startRule; 
}
//=================================
void GReATParameters::getPassAlongPackets(PassAlongPackets& arguments)
{
	arguments =  m_startArgs;
}
//================================================
void GReATParameters::useMemoryBackend()
{
	use_mem = true;
}
