#ifndef GReATParameters_h
#define GReATParameters_h

#include "GReATConfig.h"
#include "GReATConfigUsage.h"
#include "GReATConfigRuntime.h"

#include "IGReATParameters.h"

typedef set<std::string> StrSet;

typedef set<GR::RewritingRule> RewritingRuleSet;
typedef map<GR::InputPort, set<Udm::Object> > InputPort2UdmObjMap;

class MetaInfo
{
public:
	MetaInfo(const Udm::DataNetwork *meta);
  virtual ~MetaInfo();
	Uml::Diagram *theUmlDiagram;
	Udm::UdmDiagram *udmDataDiagram;
};



class GReATParameters : public IGReATParameters
{
public:
	Udm::SmartDataNetwork *sdnConfig;
protected:
	Udm::SmartDataNetwork *sdnTransformation;
	map<GReATConfig::InputFile, Udm::SmartDataNetwork *> sfile2workdn;
	vector<Udm::SmartDataNetwork*> tempDNs;

	/////////////////////////////////////////
	Uml::Diagram *clUmlDgr;
	Udm::UdmDiagram *clUdmDataDgr;
	Udm::StaticUdmProject *clProject;
	/////////////////////////////////////////
	
	string cfgmemfile;
	set<string> tempFiles;  // store of the temporary file name

	map<string, MetaInfo*> udmDiagramMap;
	
	map<string, Uml::Diagram *> umlDiagramMap;
	
	set<GReATConfig::InputFile> inputFiles;
	set<GReATConfig::TransformationInput> mappings;
	GReATConfigUsage *cfg_usageInfo;
	GReATConfigRuntime *cfg_runtimeInfo;

	GeneralBackendPtr m_backendMap; 

	PassAlongPackets      m_startArgs;
	GR::RewritingRule     m_startRule;
	GR::RootFolder        m_rootFolder;
	
	Udm::UdmProject udmPrj;
	Udm::SmartDataNetwork udmDN;

	std::string m_ttLastWrite;
	std::string m_getUdmDirName;
	std::string m_crossMetaName;
	StrUmlDiagrams_t   m_metas;
	string printMsg;
	string currentWD;

	bool use_mem;

	////////////////////////////
	vector<GReATConfig::TemporaryFileUsage> tempPkgs;

	vector<Uml::Class*> rootClasses;
	const Uml::Class& getRootClass(const Uml::Diagram &d, const std::string& rootClassName) {
		Uml::Class* root = new Uml::Class(Uml::classByPath(d, rootClassName));
		// will be deleted in the destructor
		rootClasses.push_back(root);
		return *root;
	}

protected:
	void getUdmObjs(const string &metaName, vector<string> udmObjPath, set<Udm::Object>& parentObjs, set<Udm::Object>& udmObjSet);
	bool isDesiredObj(Udm::Object obj, const string &attrName, const string &attrVal);
	void buildPassAlongPkts(PassAlongPacket& newArg,InputPort2UdmObjMap& inputPkts,
							map<GR::InputPort, set<Udm::Object> >::iterator mpos);
	void getCfgDataNetwork(const std::string& fname);

	void open(); 
	void load();
	void init(map<string, Udm::DataNetwork *> &dataNetworkMap);
	void closeDN();
	void closeTempDN();
	void closeDataNetworks();
	void deleteHeaps();
	void deleteTempFiles();

	void deleteUdmProject();

public: 
	GReATParameters();
	~GReATParameters();
	void open(const std::string& fname);
	void open(const string &fname, const vector<string> &files);
	void close();
	void closeNoSave();
	SaveResult saveProgram();
	SaveResult saveInternalBackend(bool doException = false);
	void closeProgramNoSave();
	void reloadProgram();	
	void closeProgram();
	void getRootFolder(GR::RootFolder& rootFolder);
	void getGeneralBackend(GeneralBackendPtr& generalBackend) ;
	void getTransformation(UdmDataNetworkPtr& transformation);
	void getStartRule(GR::RewritingRule& rule);
	void getPassAlongPackets(PassAlongPackets& agruments);
	void getOutputMsg(string &outputMsg) {outputMsg = printMsg;};
	void setCurrentWorkingPath(const string &path) {currentWD = path;};
	std::string getTransfLastWrittenFileName();
	std::string getUdmDirName();
	StrUmlDiagrams_t getMetas();
//	void getUserCodeInfo(StrVec& incFiles,
//						  StrVec& libFiles,
//						  StrVec& incDirs,
//						  StrVec& libDirs);


	void useMemoryBackend();
};

#endif //GReATParameters_h
