#include "CIntCompiler.h"
#include <algorithm> // for 'for_each'
#include "PEStringTable.h"
#include "GReATUtils.h"
#include "Logging.h"
#include "Uml.h"

using namespace std;

bool CIntCompiler::isGuard() const
{
	const std::string& guardExpression= ( std::string) _rule->guard();
	return ( !guardExpression.empty());
}

bool CIntCompiler::isCode() const
{
	set< GR::Code> codes= _rule->Code_kind_children();
	return ( !codes.empty());
}

/*
const std::string CIntCompiler::createGuard( PEComposite& code)
{
	_currCode= &code;
	buildOWNameMap( true);
	passObjects();
	const std::string guardVarName= createGuardInit();
	const std::string& guardExpression= ( std::string) _rule.guard();
	_currCode->add( *new Expression( guardExpression));
	return guardVarName;
}
*/

std::string CIntCompiler::getBoundVariableName(const GR::ObjectWrapper &grObj)
{
	std::string ret;
	std::map< GR::ObjectWrapper, std::string>::const_iterator it = _boundObjects->find(grObj);
	if ( it!=_boundObjects->end())
	{
		ret = (*it).second;
	}
	return ret;
}

const std::string CIntCompiler::createGuard( PEComposite& code)
{	
	//
	Function* guardFuncDef= new Function;
	guardFuncDef->_name= "isGuardTrue";
	guardFuncDef->_returnValue= BOL;
	//
	FunctionCall* guardFuncCall= new FunctionCall;
	guardFuncCall->_func= guardFuncDef;
	//
	buildOWNameMap( true);
	passParameters( *guardFuncDef, *guardFuncCall);
	//
	_currCode= guardFuncDef;
	const std::string guardVarName= createGuardInit();
	const std::string& guardExpression= ( std::string) _rule->guard();
	_currCode->add( *new Expression( guardExpression));
	_currCode->add( *new Expression( RET+ SEP+ guardVarName));
	//
	guardFuncCall->_retType= BOL;
	guardFuncCall->_retVar= guardVarName;
	_addGuardFuncDefTo->addProtected( *guardFuncDef);
	code.add( *guardFuncCall);
	//
	return guardVarName;
}

// Adds 'code' to '_output'
// This is an adoptable functor.
class AddCodeFunctor : public std::unary_function< GR::Code, void>
{
public:
	AddCodeFunctor( PEComposite* output) : _output( output) {}
	void operator()( const GR::Code& code)
	{
		_output->add( *new Expression( "{\n"+ ( std::string) code.code()+ "\n}"));
	}
private:
	PEComposite* _output;
};

void CIntCompiler::createCode( PEComposite& code)
{
	_currCode= &code;
	buildOWNameMap( false);
	passObjects();
	set< GR::Code> codes= _rule->Code_kind_children();
	for_each( codes.begin(), codes.end(), AddCodeFunctor( _currCode));
}

const std::string CIntCompiler::createGuardInit() 
{
	// print 'bool Gz_guard= true;'
//	Definition* guard= new Definition;
//	guard->_type= BOL;
//	guard->_variable= "Gz_guard";
//	guard->_initValue= FAL;
	_currCode->add( *new Expression( "bool Gz_guard= false"));
	// print 'bool* guardVal= &guard;'
//	Definition* guardVal= new Definition;
//	guardVal->_type= BOL+ PTR;
//	guardVal->_variable= "guardVal";
//	guardVal->_initValue= REF+ guard->_variable;
//	_currCode->add( *guardVal);
	// return with guard variable name.
	return "Gz_guard";
}

void CIntCompiler::buildOWNameMap( bool onlyBoundObjects)
{
	std::set< GR::ObjectWrapper> OWs= _rule->ObjectWrapper_kind_children();
	for( set< GR::ObjectWrapper>::const_iterator itOWs= OWs.begin(); itOWs!= OWs.end(); ++itOWs)
	{
		const GR::ObjectWrapper& currOW= *itOWs;
		if((string)currOW.type().name()=="PatternGroup") continue;
		if ( onlyBoundObjects && ( "CreateNew" == ( std::string) currOW.action()))
			continue;
		OWNameMap_t::iterator itFindName= _owNameMap.find( currOW.name());
		if ( _owNameMap.end() != itFindName)
		{
			printWarning( "Object name: "+ ( std::string) currOW.name()+ " is not unique. Undeterministic AttributeMapping and/or Guard.");
			itFindName->second= std::make_pair( itFindName->second.first, false);	// set to not unique
		}
		else
			_owNameMap[ ( std::string) currOW.name()]= std::make_pair( currOW, true);	// unique by default
	}
}

class CollectNames {
public:
	void operator()( const std::pair< GR::ObjectWrapper, std::string>& bo) {
		_names+= ( std::string) bo.first.name() + "= "+ bo.second+ "\n";
	}
	std::string _names;
};

void CIntCompiler::passObjects()
{
	for( OWNameMap_t::const_iterator itNames= _owNameMap.begin(); itNames!= _owNameMap.end(); ++itNames)
	{
		if ( true == itNames->second.second)	// if pattern object name is unique..
		{
			const GR::ObjectWrapper& currOW= itNames->second.first;
			std::map< GR::ObjectWrapper, std::string>::const_iterator itBoundObject= _boundObjects->find( currOW);
			if ( _boundObjects->end()== itBoundObject)
			{
				CollectNames nameColl;
				for_each_ref( _boundObjects->begin(), _boundObjects->end(), nameColl);
				throw CGEx("FATAL: CG internal error. Pattern object: "+ ( std::string) currOW.name()+ " not bound in rule:" + ( std::string)_rule->name()+ " (CInt AM compiler.)\nBound Objects:\n"+ nameColl._names);
			}
			const std::string& generatedName= (*itBoundObject).second;
			Expression* pseudo= new Expression;	// use expression, because Definition will lowercast the variable name.
			//pseudo->_expression= ( std::string) currOW.paradigmName()+ RES+ ( std::string) currOW.classType()+ REF +SEP+ ( std::string) currOW.name()+ EQ+ SEP+ generatedName;
			pseudo->_expression= PE::getFullType(currOW)+ REF +SEP+ ( std::string) currOW.name()+ EQ+ SEP+ generatedName;
			_currCode->add( *pseudo);
		}
	}
}

void CIntCompiler::passParameters( Function& guardFuncDef, FunctionCall& guardFuncCall)
{
	for( OWNameMap_t::const_iterator itNames= _owNameMap.begin(); itNames!= _owNameMap.end(); ++itNames)
	{
		if ( true == itNames->second.second)	// if pattern object name is unique..
		{
			const GR::ObjectWrapper& currOW= itNames->second.first;
			if((string)currOW.type().name() == "PatternGroup") continue;
			std::map< GR::ObjectWrapper, std::string>::const_iterator itBoundObject= _boundObjects->find( currOW);
			if ( _boundObjects->end()== itBoundObject)
			{
				CollectNames nameColl;
				for_each_ref( _boundObjects->begin(), _boundObjects->end(), nameColl);
				throw CGEx("FATAL: CG internal error. Pattern object: "+ ( std::string) currOW.name()+ " not bound in rule:" + ( std::string)_rule->name()+ " (CInt guard compiler.)\nBound Objects:\n"+ nameColl._names);

//				throw CGEx("FATAL: CG internal error. Pattern object: "+ ( std::string) currOW.name()+ " not bound in rule:" + ( std::string)_rule->name()+ " (CInt compiler.)");
			}
			//std::string paramType= CST+ SEP+ "Udm"+ RES+ "Object"+ REF;
//			std::string paramType= CST+ SEP+ ( std::string)currOW.paradigmName()+ RES+ ( std::string)currOW.classType()+ REF;
			std::string paramType= PE::getFullType(currOW)+ REF;
			const std::string& paramName= currOW.name();
			const std::string& generatedName= (*itBoundObject).second;
			guardFuncDef.insertInputParam( FuncParam_t( make_pair( paramType, paramName)));
			guardFuncCall._params.push_back( generatedName);
		}
	}
}

void CIntCompiler::createGroupRule( PEComposite& code, GR::PatternGroup& group)
{
	_currCode= &code;
	buildOWNameMap(true);

	for( OWNameMap_t::const_iterator itNames= _owNameMap.begin(); itNames!= _owNameMap.end(); ++itNames)
	{
		if ( (*itNames).second.second )	// if pattern object name is unique..
		{
			const GR::ObjectWrapper& currOW= (*itNames).second.first;
			std::map< GR::ObjectWrapper, std::string>::const_iterator itBoundObject= _boundObjects->find( currOW);
			if ( _boundObjects->end()== itBoundObject)
			{
				CollectNames nameColl;
				for_each_ref( _boundObjects->begin(), _boundObjects->end(), nameColl);
				throw CGEx("FATAL: CG internal error. Pattern object: "+ ( std::string) currOW.name()+ " not bound in rule:" + ( std::string)_rule->name()+ " (CInt AM compiler.)\nBound Objects:\n"+ nameColl._names);
			}
			std::string generatedName= (*itBoundObject).second;
			generatedName.erase(0,9);
			Expression* the_var= new Expression;	// use expression, because Definition will lowercast the variable name.
//			the_var->_expression= CST+SEP+( std::string) currOW.paradigmName()+ RES+ ( std::string) currOW.classType()+ REF +SEP+ "the_"+( std::string) currOW.name()+ EQ+ SEP+ "the_match"+generatedName;
			the_var->_expression= CST+SEP+PE::getFullType(currOW)+ REF +SEP+ "the_"+( std::string) currOW.name()+ EQ+ SEP+ "the_match"+generatedName;
			_currCode->add( *the_var);
			Expression* other_var= new Expression;	// use expression, because Definition will lowercast the variable name.
//			other_var->_expression= CST+SEP+( std::string) currOW.paradigmName()+ RES+ ( std::string) currOW.classType()+ REF +SEP+ "other_"+( std::string) currOW.name()+ EQ+ SEP+ "other_match"+generatedName;
			other_var->_expression= CST+SEP+PE::getFullType(currOW)+ REF +SEP+ "other_"+( std::string) currOW.name()+ EQ+ SEP+ "other_match"+generatedName;
			_currCode->add( *other_var);
		}
	}

	//const set<GR::PatternGroup> groups = _rule->PatternGroup_kind_children();
	//GR::PatternGroup group = *(groups.begin());
	Expression* group_rule = new Expression;
	if(!((std::string)group.rule()).empty()) 
		group_rule->_expression = (std::string)group.rule();
	else
		group_rule->_expression = "return false";
	_currCode->add(*group_rule);
}

void CIntCompiler::buildOWNameMap(bool onlyBoundObjects, OWNameMap_t& nameMap)
{
	buildOWNameMap(onlyBoundObjects);
	nameMap = _owNameMap;
}

