#include "InputFilter.h"
#include "PEStringTable.h"
#include "GReATUtils.h"
#include <algorithm>

using namespace std;


void InputFilter::execute( ParamDatas_t& ruleInputs, Class& addTo, PEComposite& addCallTo)
{
	Function* filterFuncDef;
	FunctionCall* filterFuncCall;
	createFunction( &filterFuncDef, &filterFuncCall, &addTo, &addCallTo);
	createUniqueInputMembers( ruleInputs, addTo);
	inputIdentityChecks( ruleInputs, filterFuncDef, filterFuncCall);
}

void InputFilter::createFunction( Function** filterFuncDef, FunctionCall** filterFuncCall, Class* addDefTo, PEComposite* addCallTo)
{
	// func def
	*filterFuncDef= new Function;
	(*filterFuncDef)->_name= "isInputUnique";
	(*filterFuncDef)->_returnValue= "bool";
	// func call
	*filterFuncCall= new FunctionCall;
	(*filterFuncCall)->_func= *filterFuncDef;
	(*filterFuncCall)->_retType= "bool";
	(*filterFuncCall)->_retVar= "isUnique";
	// add func def
	if ( addDefTo)
	{
		addDefTo->addProtected( **filterFuncDef);
	}
	// add func call
	if ( addCallTo)
	{
		addCallTo->add( **filterFuncCall);
		If* isUnique= new If;
		isUnique->createIfExprThenContinue( "!"+ (*filterFuncCall)->_retVar);
		addCallTo->add( *isUnique);
	}
}

void InputFilter::createUniqueInputMembers( ParamDatas_t& ruleInputs, Class& addTo)
{
	for( ParamDatas_t::iterator i= ruleInputs.begin(); i!= ruleInputs.end(); ++i)
	{
		createUniqueInputMember( *i, addTo);
	}
}

void InputFilter::createUniqueInputMember( ParamData& ruleInput, Class& addTo)
{
	// create definition
	ClassDef* inputDef= new ClassDef;
	inputDef->createPackets( "_"+ lowerBegin( PE::createUniqueName( ( std::string) ruleInput._param.name())));
	addTo.addPrivate( *inputDef);
	// register definiton
	ruleInput._memberDef= inputDef;
}

void InputFilter::createInitCode( Function* filterFuncDef, Loop** inputLoop, If** uniqueCondition, If** checkForUnique)
{
	Definition* unique= new Definition;
	unique->_type= BOL;
	unique->_variable= UNQ;
	unique->_initValue= TRU;
	filterFuncDef->add( *unique);
	// match is added output only if the match is unique among outputs.
	(*inputLoop)= new Loop;
	filterFuncDef->add( **inputLoop);
	//
	(*uniqueCondition)= new If;
	(*uniqueCondition)->_expression= UNQ;
	filterFuncDef->add( **uniqueCondition);
	//
	(*checkForUnique)= new If;
	(*inputLoop)->add( **checkForUnique);
}

void InputFilter::createFinalCode( If* checkForUnique, Function* filterFuncDef)
{
	checkForUnique->add( *new Expression(UNQ+ EQ+ SEP+ FAL));
	checkForUnique->add( *new Expression("break"));
	filterFuncDef->add( *new Expression(RET+ SEP+ UNQ));
}

void InputFilter::inputIdentityCheck( 
		const ParamData& ruleInput, 
		const std::string& callerParamName, 
		Function* filterFuncDef, 
		FunctionCall* filterFuncCall,
		Loop* inputLoop,
		If* uniqueCondition,
		If* checkForUnique
		)
{
	// Do multiple things in parallel...
	// 1, Add parameter name to function caller.
	filterFuncCall->_params.push_back( callerParamName);
	// 2, Add parameter definition to function definition.
	FuncParam_t calleeParam;
	createParam( ruleInput, calleeParam);
	const std::string& paramName= calleeParam._param.second;
	filterFuncDef->insertInputParam( calleeParam);
	// 3, Append input loop expression with current input
	std::string iteratorName= PE::createUniqueName( IT+ ( std::string)ruleInput._param.name());
	inputLoop->addInputIterator( iteratorName, lowerBegin( ruleInput._memberDef->_variable));
	// 4, Append unique condition with current input;
	checkForUnique->appendWithAnd( PTR+ iteratorName+ EQEQ+ SEP+ lowerBegin( paramName));
	// 5, Create output appender expression into 'uniqueCondition'.
	Expression* appendInput= new Expression;
	const string& inputMemberName= ruleInput._memberDef->_variable;
	appendInput->_expression= inputMemberName+ "."+ "push_back("+ SEP+ lowerBegin( paramName)+ ")";
	uniqueCondition->add( *appendInput);


/*
	// 3, Create find definiton.
	Definition* findInput= new Definition;
	findInput->_type= PKS+ RES+ CIT;
	findInput->_variable= IT+ "Find"+ paramName;
	const string& inputMemberName= ruleInput._memberDef->_variable;
	findInput->_initValue= "std::find("+ SEP+ inputMemberName+ "."+ BEG+ ","+ SEP+ inputMemberName+ "."+ END+ ","+ SEP+ lowerBegin( paramName)+ ")";
	currHitCheck->add( *findInput);
	// 4, Create hit check condition.
	If* hitCheck= new If;
	hitCheck->createHitCheckCondition( inputMemberName, findInput->_variable);
	currHitCheck->add( *hitCheck);
	// 5, Create output appender expression into 'uniqueCondition'.
	Expression* appendInput= new Expression;
	appendInput->_expression= inputMemberName+ "."+ "push_back("+ SEP+ lowerBegin( paramName)+ ")";
	uniqueCondition->add( *appendInput);
	//
	return hitCheck;
*/
 }

void InputFilter::inputIdentityChecks( const ParamDatas_t& ruleInputs, Function* filterFuncDef, FunctionCall* filterFuncCall)
{
	Loop* inputLoop= 0;
	If* uniqueCondition= 0;
	If* checkForUnique= 0;
	createInitCode( filterFuncDef, &inputLoop, &uniqueCondition, &checkForUnique);
	// For every input...
	for( ParamDatas_t::const_iterator itInputs= ruleInputs.begin(); itInputs!= ruleInputs.end(); ++itInputs)
	{
		std::map< GR::ObjectWrapper, std::string>::const_iterator itBoundObject= _boundObjects->find( itInputs->_param);
		if ( _boundObjects->end()== itBoundObject)
		{
			throw CGEx("FATAL: CG internal error. Input parameter: "+ ( std::string)itInputs->_param.name()+ " not bound in input filter.");
		}
		inputIdentityCheck( *itInputs, (*itBoundObject).second, filterFuncDef, filterFuncCall, inputLoop, uniqueCondition, checkForUnique);
	}
	createFinalCode( checkForUnique, filterFuncDef);
}

void InputFilter::createParam( const GR::ObjectWrapper& input, FuncParam_t& param) const
{
	std::set< GR::ObjectWrapper> OWs= _rule.ObjectWrapper_kind_children();
	std::set< GR::ObjectWrapper>::const_iterator itFindInput= std::find( OWs.begin(), OWs.end(), input);
	if ( OWs.end()== itFindInput)
		throw CGEx("FATAL: Object wrapper belonging to input parameter: "+ ( std::string) input.name()+ " not found in rule: "+ ( std::string)_rule.name()+ ".");
//	const std::string& paradigm= itFindInput->paradigmName();
//	const std::string& type= itFindInput->classType();
	const std::string argType= "Udm::Object";
	param._param= make_pair( CST+ SEP+ argType+ REF, PE::createUniqueName( lowerBegin(( std::string)itFindInput->name())));
}
