#include "OutputSorter.h"
#include "PEStringTable.h"
#include "GReATUtils.h"
#include <algorithm>

///////////////////////////////////////////////////////////////////////////////
void collectOutputsToSort( const ParamDatas_t& outputs, ParamDatas_t& outputsToSort) {
	for( ParamDatas_t::const_iterator itOutputs= outputs.begin(); itOutputs!= outputs.end(); ++itOutputs) {
		const GR::ObjectWrapper& ow= itOutputs->_param;
		if ( ow.type() != GR::PatternObject::meta)
			continue;
		GR::PatternObject po= GR::PatternObject::Cast( ow);
		if ( false ==(( std::string)po.cmpfunc()).empty()) {
			outputsToSort.push_back( *itOutputs);
		}
	}
}

bool OutputSorter::areOutputsToBeSorted( const ParamDatas_t& outputs) {
	ParamDatas_t outputsToSort;
	collectOutputsToSort( outputs, outputsToSort);
	return !outputsToSort.empty();
}

///////////////////////////////////////////////////////////////////////////////
void OutputSorter::generate( Function& sorterFuncDef) {
	ParamDatas_t outputsToSort;
	collectOutputsToSort( _ruleOutputs, outputsToSort);
	if ( 1 < outputsToSort.size()) {
		throw CGEx("More than one (" + num_to_string( outputsToSort.size()) +") compare function specified in rule: " + ( std::string)_rule.name());
	} else if ( outputsToSort.empty()) {
		return; // nothing to sort
	}
	// exactly one output with compare functinon. This is the primary key.
	const ParamData& primaryKey= *outputsToSort.begin();
	generatePrimaryKeySorterCode( primaryKey, sorterFuncDef);
	generateOutputSorterCode( primaryKey, sorterFuncDef);
}

void OutputSorter::generatePrimaryKeySorterCode( const ParamData& primaryKey, PEComposite& addTo) {
//	const std::string TYPE= ( std::string)primaryKey._param.paradigmName() + RES + ( std::string)primaryKey._param.classType();
	const std::string TYPE= PE::getFullType(primaryKey._param);
	const std::string CMP_FUNC= GR::PatternObject::Cast( primaryKey._param).cmpfunc();
	const std::string PRIMARY_KEY= primaryKey._memberDef->_variable;
//	typedef std::multiset< TYPE, std::pointer_to_binary_function< const TYPE&, const TYPE&, bool> > SortedSet_t;
	addTo.add( *new Expression( "typedef std::multiset< "+ TYPE+ ", std::pointer_to_binary_function< const "+ TYPE + "&, const "+ TYPE +"&, bool> > SortedSet_t"));
//	std::pointer_to_binary_function< const TYPE&, const TYPE&, bool> ptr_CMP_FUNC( CMP_FUNC< TYPE>);
	addTo.add( *new Expression( "std::pointer_to_binary_function< const "+ TYPE+ "&, const "+ TYPE+ "&, bool> ptr_"+ CMP_FUNC+ "( "+ CMP_FUNC+ "< "+ TYPE+ ">)"));
//	SortedSet_t sortedSet( ptr_CMP_FUNC);
	addTo.add( *new Expression( "SortedSet_t sortedSet( ptr_"+ CMP_FUNC+ ")"));
//	std::vector< int> permutationVector( PRIMARY_KEY->size());
	addTo.add( *new Expression( "std::vector< int> permutationVector( "+ PRIMARY_KEY+ "->size())"));
//	int currStep= 0;
	addTo.add( *new Expression( "int currStep= 0"));
//	for( Packets_t::const_iterator it= PRIMARY_KEY->begin(); it!= PRIMARY_KEY->end(); ++it, ++currStep) {
	Loop* packetLoop= new Loop;
	packetLoop->createIteratorLoop( IT, PRIMARY_KEY, PKS, true);
	packetLoop->_increment+= ", ++currStep";
	addTo.add( *packetLoop);
//		SortedSet_t::const_iterator inserted= sortedSet.insert( TYPE::Cast(*it));
	packetLoop->add( *new Expression( "SortedSet_t::const_iterator inserted= sortedSet.insert( "+ TYPE+ "::Cast(*"+ IT+ "))"));
//		SortedSet_t::difference_type pos= std::distance( SortedSet_t::const_iterator( sortedSet.begin()), inserted);
	packetLoop->add( *new Expression( "SortedSet_t::difference_type pos= std::distance( SortedSet_t::const_iterator( sortedSet.begin()), inserted)"));
//		permutationVector[ currStep]=  pos;
	packetLoop->add( *new Expression( "permutationVector[ currStep]=  pos"));
//		for( int i= 0; i< currStep; ++i) {
	Loop* permVectorLoop= new Loop;
	permVectorLoop->_from= "int i= 0";
	permVectorLoop->_to= "i< currStep";
	permVectorLoop->_increment= "++i";
	packetLoop->add( *permVectorLoop);
//			if ( permutationVector[ i] >= pos) 
	If* checkPos= new If;
	checkPos->_expression= "permutationVector[ i] >= pos";
	permVectorLoop->add( *checkPos);
//				++permutationVector[ i];
	checkPos->add( *new Expression( "++permutationVector[ i]"));
//		}
//	}
//	// order the primary container
	addTo.add( *new CommentLine("order the primary container"));
//	std::copy( sortedSet.begin(), sortedSet.end(), PRIMARY_KEY->begin());
	addTo.add( *new Expression( "std::copy( sortedSet.begin(), sortedSet.end(), "+ PRIMARY_KEY+ "->begin())"));
}

void OutputSorter::generateOutputSorterCode( const ParamData& primaryKey, PEComposite& addTo) {
//	// order rest of the containers
	addTo.add( *new CommentLine("order rest of the containers"));
	for( ParamDatas_t::const_iterator itOutputs= _ruleOutputs.begin(); itOutputs!= _ruleOutputs.end(); ++itOutputs) {
		const GR::ObjectWrapper& ow= itOutputs->_param;
		if ( ow != primaryKey._param) {
			const std::string& OUTPUT= itOutputs->_memberDef->_variable;
//			permutate( permutationVector, *OUTPUT);
			addTo.add( *new Expression( "permutate( permutationVector, *"+ OUTPUT+ ")"));
		}
	}
}

