/* NAME:	Output sorter
 * AUTHOR:	Attila Vizhanyo - viza@isis-server.isis.vanderbilt.edu
 *
 * Generate code, which sorts the outputs based on a 
 * compare function specified for one output (primary key).
 */
#ifndef __OUTPUTSORTER_H
#define __OUTPUTSORTER_H

#ifdef _MSC_VER
#pragma warning (disable : 4786)
#pragma warning (disable : 4503) 
#endif

#include <map>
#include "RuleTypes.h"
#include "PE.h"
#include "PacketPassing.h"


// This class is not designed to be a base class. 
// Either don't inherit from it, or add virtual destructor.
class OutputSorter
{
public:
	// static methods
	static bool areOutputsToBeSorted( const ParamDatas_t& outputs);
	// ctor; see parameter desc. at the member variables declaration part. 
	OutputSorter( const GRSimpleRule& rule, 
		const ParamDatas_t& ruleOutputs
		) : _rule( rule) , _ruleOutputs( ruleOutputs)
		{}
	void generate( Function& sorterFuncDef);

protected:
	void generatePrimaryKeySorterCode( const ParamData& primaryKey, PEComposite& addTo);
	void generateOutputSorterCode( const ParamData& primaryKey, PEComposite& addTo);

private:
	// Rule, on which output sorter code generation executes.
	GRSimpleRule _rule;
	// 
	const ParamDatas_t& _ruleOutputs;
};

#endif //__OUTPUTSORTER_H