#include "PatternMatcher.h"
#include "PacketPassing.h"
#include "CIntCompiler.h"
#include "PEStringTable.h"
#include "GReATUtils.h"
#include "CGEx.h"
#include "Logging.h"
#include "UmlExt.h"
#include <UdmStatic.h>

using namespace std;

///////////////////////////////////////////////////////////////////////////////
// global functions

// template specialization of less for Uml::AssociationRole-s
template<> struct std::less< Uml::AssociationRole> : public std::binary_function< Uml::AssociationRole, Uml::AssociationRole, bool> {
	bool operator()( const Uml::AssociationRole& lhs, const Uml::AssociationRole& rhs) const {
		return std::less< std::string>()( ( std::string) lhs.name(), ( std::string) rhs.name());
	}
};

bool findAssoc( const std::set< Uml::AssociationRole>& assocRoles, const Uml::AssociationRole& assocToFind, Uml::Class& targetClass, Uml::Association& assocClass, int& maxMultiplicity) 
{
	std::set< Uml::AssociationRole>::const_iterator itFind= assocRoles.find( assocToFind);;
	if ( itFind!= assocRoles.end())
	{
		// Found.
		maxMultiplicity= ( int) itFind->max();
		targetClass= itFind->target();
		assocClass= itFind->parent();
		return true; 
	}
	else
	{	
		// OK this is ugly, but works. set::find assumes the container is ordered by the comparison operator
		// used during the search. Howevever, the set is not ordered by UDM, so we order it first, then try
		// to search again. 
		// This solution will be replaced, once UDM supports COMP functors in its interface.
		std::set< Uml::AssociationRole, std::less< Uml::AssociationRole> > sorted;
		for( itFind= assocRoles.begin(); itFind!= assocRoles.end(); ++itFind)
		{
//			std::cout << ( std::string) itFind->name() << std::endl;
			sorted.insert( *itFind);
		}
//		std::cout << "sorted" << std::endl;
//		for( itFind= sorted.begin(); itFind!= sorted.end(); ++itFind)
//		{
//			std::cout << ( std::string) itFind->name() << std::endl;
//		}
		itFind= sorted.find( assocToFind);;
		if ( itFind!= sorted.end())
		{
			// Found.
			maxMultiplicity= ( int) itFind->max();
			targetClass= itFind->target();
			assocClass= itFind->parent();
			return true; 
		}
	}
	return false;	// Not found.
}

std::set< Uml::AssociationRole> getAncestorAssociationRoles( const Uml::Diagram& diag, const Uml::Class& connClass)
{
	std::set< Uml::AssociationRole> assocRoles;
	std::set< Uml::Class> ancestors= Uml::AncestorClasses( connClass);
	for( std::set< Uml::Class>::const_iterator it= ancestors.begin(); it!= ancestors.end(); ++it) {
		const Uml::Class& currClass= *it;
		Uml::Association assoc = currClass.association();
//		Uml::Association assoc= Uml::associationByName( diag, currClass.name());
		if ( !assoc)
			continue;
		std::set< Uml::AssociationRole> currAssocRoles= assoc.roles();
		assocRoles.insert( currAssocRoles.begin(), currAssocRoles.end());
	}
	return assocRoles;
}

// Get the name of association class, the name of target class, and maximum multiplicity. 
bool getAssociationProperties( 
	const GeneralBackend& backend,
	const std::string& paradigm, 
	const std::string& boundclass, 
	const std::string& association, 
	bool isSourceAssoc,
	Uml::Class& targetClass, 
	Uml::Association& assocClass,
	int& maxMultiplicity,
	bool& isCrossLink
	) 
{
	try
	{
		// create temporary dummy association object for searching of this value
		Uml::AssociationRole assocToFind= Uml::AssociationRole::Cast( UdmStatic::CreateObject( Uml::AssociationRole::meta));
		assocToFind.name()= association;
		if ( association == "dstEnd2") {
			int stophere= 0;
		}
		// association query depends on whether the source is association or class..
		std::set< Uml::AssociationRole> assocRoles;
		//if ( isSourceAssoc)
		{
			// first see if the association is simple (not cross-link)
			// try to find the association (and all its ancestors in case of connection inheritance hierarchies) 
			// in the meta-paradigm of the model
			Uml::Class clBound;
//			clBound= Uml::classByName( backend.getDiagram( paradigm), boundclass);
			clBound= Uml::classByPath( backend.getDiagram( paradigm), boundclass);
			assocRoles= getAncestorAssociationRoles(  backend.getDiagram( paradigm), clBound);
			std::set< Uml::AssociationRole> targetRoles = Uml::AncestorAssociationTargetRoles(clBound);
			assocRoles.insert(targetRoles.begin(), targetRoles.end());
			if ( true== findAssoc( assocRoles, assocToFind, targetClass, assocClass, maxMultiplicity))
			{
				isCrossLink= false;
				return true;	// simple link
			}
			// It is either a cross-link or does not exist
	//		clBound= Uml::classByName( backend.getDiagram( paradigm), boundclass);
			assocRoles= Uml::AncestorCrossAssociationTargetRoles( clBound, backend.getCrossLinkDiagram());
			if ( true== findAssoc( assocRoles, assocToFind, targetClass, assocClass, maxMultiplicity))
			{
				isCrossLink= true;
				return true;	// cross-link
			}
	//		return false; // not found
		}
		//else
		//{
		//	// first see if the association is simple (not cross-link)
		//	Uml::Class clBound;
		//	clBound= Uml::classByName( backend.getDiagram( paradigm), boundclass);
		//	assocRoles= Uml::AncestorAssociationTargetRoles( clBound);
		//	if ( true== findAssoc( assocRoles, assocToFind, targetClass, assocClass, maxMultiplicity))
		//	{
		//		isCrossLink= false;
		//		return true;	// simple link
		//	}
		//	// It is either a cross-link or does not exist
		//	assocRoles= Uml::AncestorCrossAssociationTargetRoles( clBound, backend.getCrossLinkDiagram());
		//	if ( true== findAssoc( assocRoles, assocToFind, targetClass, assocClass, maxMultiplicity))
		//	{
		//		isCrossLink= true;
		//		return true;	// cross-link
		//	}
		//	return false; // not found
		//}
	} catch( udm_exception &ex) 
	{
		std::cout << ex.what() << std::endl;
		// Convert udm_exception to CGEx.
	}
	// Not found, or exception..
	return false;
}


bool getAssociationProperties( 
	const GeneralBackend& backend,
	const std::string& paradigm, 
	const std::string& ns_name,
	const std::string& boundclass, 
	const std::string& association, 
	bool isSourceAssoc,
	Uml::Class& targetClass, 
	Uml::Association& assocClass,
	int& maxMultiplicity,
	bool& isCrossLink
	) 
{
	std::string fullpath = boundclass;
	if(!ns_name.empty())
		fullpath = ns_name + "::" + boundclass;
	return getAssociationProperties(backend,paradigm,fullpath,association,isSourceAssoc,targetClass,assocClass,maxMultiplicity, isCrossLink);
}
///////////////////////////////////////////////////////////////////////////////
// PatternMatcher public interface
const Definition* PatternMatcher::createInputMatch( const GR::ObjectWrapper& input, const std::string& inputName)
{
	// 1, current input definition
//	Definition* currInputDef= new Definition;
//	currInputDef->createCurrentInput( PE::createUniqueName( CURR+ ( std::string)input.name()), iter);
//	_currCode->add( *currInputDef);
	// 2, typecheck
	If* typeCheck= new If;
	typeCheck->createIfTypeMismatchThenContinue( lowerBegin( inputName), ( std::string) input.paradigmName(), (std::string)input.nameSpace(), ( std::string) input.classType());
	_currCode->add( *typeCheck);
	// 3, typecast
	Definition* typeCast= new Definition;
//	typeCast->createTypeCast( ( std::string) input.paradigmName()+ RES+ ( std::string) input.classType(), PE::createUniqueName( ( std::string)input.name()), lowerBegin( inputName));
	typeCast->createTypeCast(PE::getFullType(input), PE::createUniqueName( ( std::string)input.name()), lowerBegin( inputName));
	_currCode->add( *typeCast);
	// this binds the current input definition to its input parameter counterpart.
	bindInput( input, *typeCast);
	return typeCast;
}

void PatternMatcher::findMatch()
{
	// bind all temp roots
	std::set< GR::ObjectWrapper> ows= _rule.ObjectWrapper_kind_children();
	for( std::set< GR::ObjectWrapper>::const_iterator it= ows.begin(); it!= ows.end(); ++it) {
		const GR::ObjectWrapper& ow= *it;
//		std::string paradigmPlusClassName= ( std::string) ow.paradigmName() + RES+ ( std::string) ow.classType();
		std::string paradigmPlusClassName= PE::getFullType(ow);
		Definition* tempRootDef= _backend.getTempRootDef( paradigmPlusClassName);
		if ( tempRootDef) {
			bind( ow);
		}
	}
	// get pattern edge with both vertices having valid bindings
	GR::ObjectLink edge;
	bool isEdge= edgeWithBothVerticesBound( edge);
	while ( true== isEdge)
	{
		translateEdgeWithBothVerticesBound( edge);
		setEdgeTraversed( edge);
		isEdge= edgeWithBothVerticesBound( edge);
	}
	// get pattern edge with one vertex bound
	isEdge= edgeWithOneVertexBound( edge);
	if ( true== isEdge)
	{
		translateEdgeWithOneVertexBound( edge);
		setEdgeTraversed( edge);
		findMatch();
	}
}

class PassAlongsEmptyPredicate {
	bool passAlongsEmpty(const GR::ObjectWrapper& ow) {
		const set<GR::PassAlong> passalongs = ow.srcObjectWrapper();
		return passalongs.empty();
	}

public:
	bool operator() (const GR::ObjectWrapper& ow1, const GR::ObjectWrapper& ow2) {
		bool ow1Empty = passAlongsEmpty(ow1);
		bool ow2Empty = passAlongsEmpty(ow2);
		if (ow1Empty && !ow2Empty) {
			return false;
		}
		if (!ow1Empty && ow2Empty) {
			return true;
		}
		return std::less<GR::ObjectWrapper>()(ow1, ow2);
	}
};

void PatternMatcher::makeMatch( Class& ruleDef, std::map< GR::ObjectWrapper, std::string>& boundObjects)
{
	//Create a function "isValidBound" to check whether the uniqueness of the objects bound to the match
	if(_rule.uniqueIdentity())
	{
		Function* isValidBoundDef = new Function;
		isValidBoundDef->_name = "isValidBound";
		ruleDef.addProtected( *isValidBoundDef);
		isValidBoundDef->_returnValue = BOL;	
		isValidBoundDef->insertInputParam(FuncParam_t( std::make_pair( "set< pair<int, Udm::Object> >&", "boundObjs")));
		isValidBoundDef->insertInputParam(FuncParam_t( std::make_pair( "Udm::Object&", "currObj")));
		isValidBoundDef->insertInputParam(FuncParam_t( std::make_pair( "bool", "isInputObj")));
	
	//if(boundObjs.find(make_pair(currObj.__impl()->__getdn()->uniqueId(), currObj)) != boundObjs.end())
	//{
	//	if(!isInputObj) 
	//		return false;
	//	else 
	//		return true;
	//}
	//else
	//	boundObjs.insert(make_pair(currObj.__impl()->__getdn()->uniqueId(), currObj));

	//return true;
	
		If* issame = new If;
		issame->_expression = "boundObjs.find(make_pair(currObj.__impl()->__getdn()->uniqueId(), currObj)) != boundObjs.end()";
		If* isinput = new If;
		isinput->_expression = "!isInputObj";
		isinput->add(* new Expression("return false"));
		isinput->addElse(* new Expression("return true"));
		issame->add(*isinput);
		issame->addElse(* new Expression("boundObjs.insert(make_pair(currObj.__impl()->__getdn()->uniqueId(), currObj))"));
		isValidBoundDef->add(*issame);
		isValidBoundDef->add(*new Expression("return true"));
	}
	// 1, Generate class definiton for 'Match'.
	Class* matchClassDef= new Class;
	matchClassDef->_name= MATCH;
	ruleDef.addPrivate( *matchClassDef);
	// 2, Generate code for 'Matches'.
	ClassDef* matchesDef= new ClassDef;
	matchesDef->_type= "std::list<"+ SEP+ MATCH+ ">";
	matchesDef->_variable= _MATCHES;
	ruleDef.addPrivate( *matchesDef);
	// 3, Generate definition for local object current match.
	Definition* matchDef= new Definition;
	matchDef->_type= MATCH;
	matchDef->_variable= lowerBegin( CURRM);
	_currCode->add( *matchDef);

	Definition* boundObjsDef = new Definition;
	boundObjsDef->_type = "set< pair<int, Udm::Object> >";
	boundObjsDef->_variable = PE::createUniqueName("boundObjs");
	if(_rule.uniqueIdentity())
	{
		_currCode->add(*boundObjsDef);
	}
	// Filling Match class definition with members and simultaneously adding current matches to the match.
	std::set< GR::ObjectWrapper, PassAlongsEmptyPredicate> OWs= _rule.ObjectWrapper_kind_children_sorted(PassAlongsEmptyPredicate());
	for( std::set< GR::ObjectWrapper, PassAlongsEmptyPredicate>::const_iterator itOWs= OWs.begin(); itOWs!= OWs.end(); ++itOWs)
	{
		const GR::ObjectWrapper& currOW= *itOWs;
		if((string)currOW.type().name() == "PatternGroup") continue;
		if ( ( "Bound" != ( std::string) currOW.action()) && ( "Delete" != ( std::string) currOW.action()))
			continue;	// ...if not a pattern vertex...


		// 'Add member to class Match representing this pattern vertex' code:
		std::string matchedName= getBoundObjectName( currOW);
		std::string matchTagName= PE::createUniqueName( currOW.name());
		ClassDef* matchTagDef= new ClassDef;
//		matchTagDef->_type= ( std::string)currOW.paradigmName()+ RES+ ( std::string)currOW.classType();
		matchTagDef->_type= PE::getFullType(currOW);
		matchTagDef->_variable= lowerBegin(matchTagName);
		matchClassDef->addPublic( *matchTagDef);
		//check the uniqueness of the to-be-bound object
		//Check whether currOW is incoming bound and whether it shares the same with others
		if(_rule.uniqueIdentity())
		{
			set<GR::PassAlong> passalongs = currOW.srcObjectWrapper();
			If *isvalidObj = new If;
			string isInput_arg = passalongs.empty() ? "false" : "true";
			isvalidObj->_expression = "!isValidBound("+boundObjsDef->_variable+", "+lowerBegin( matchedName)+", "+isInput_arg+")";
			isvalidObj->add(*new Expression("continue"));
			_currCode->add(*isvalidObj);
		}
		// 'Assign match variable to the class match member' code:
		Definition* makeMatch= new Definition;
		makeMatch->_variable= matchDef->_variable+ "."+ lowerBegin( matchTagName);
		makeMatch->_initValue= lowerBegin( matchedName);
		_currCode->add( *makeMatch);
		// Bound the name here!
		boundObjects[ currOW]= CURRM+ "."+ lowerBegin( matchTagDef->_variable);
	}

	Expression* makeMatch= new Expression;
	makeMatch->_expression= matchesDef->_variable+ ".push_back("+ SEP+ matchDef->_variable+ ")";
	PEComposite* addMakeMatchTo= _currCode;
	// Code for guards:
	_cc.init( _rule, boundObjects, ruleDef);
	if ( true== _cc.isGuard())
	{
		//const string guardVarName= _cc.createGuardFunction( ruleDef, *_currCode);
		const string guardVarName= _cc.createGuard( *_currCode);
		// There is guard, and the guard variable is 'guardVarName'.
		If* guardCondition= new If;
		guardCondition->_expression=  TRU+ EQEQ+ SEP+ guardVarName;
		_currCode->add( *guardCondition);
		addMakeMatchTo= guardCondition;
	}
	addMakeMatchTo->add( *makeMatch);
	// Exhaustive
	if ( false== _rule.exhaustive())
	{
		// Being a rule not exhaustive means once a match made, don't make more matches.
		// Thus, we can return after match is made.
		addMakeMatchTo->add( *new Expression( RET+ SEP+ TRU));
	}
}

void PatternMatcher::bindInput( const GR::ObjectWrapper& inputOw, const Definition& inputDef)
{
	bind( inputOw);
	_boundObjects.addObjectReference( inputOw, inputDef);
}

void PatternMatcher::bind( const GR::ObjectWrapper& ow) const
{
	ow.objectID()= 1; // not 0
}

bool PatternMatcher::isBound( const GR::ObjectWrapper& ow) const
{
	return ( 0 != ow.objectID());
}

bool PatternMatcher::isEdgeToMatch( const GR::ObjectLink& edge) const
{
	if (( "Bound"== ( std::string) edge.action()) || ( "Delete"== ( std::string) edge.action()))
		return true;
	return false;
}

bool PatternMatcher::edgeWithBothVerticesBound( GR::ObjectLink& edge) const
{
	std::set< GR::ObjectLink> edges= _rule.ObjectLink_kind_children();
	for( std::set< GR::ObjectLink>::const_iterator itEdges= edges.begin(); itEdges!= edges.end(); ++itEdges)
	{
		const GR::ObjectLink& currEdge= *itEdges;
		if ( isEdgeTraversed( currEdge) || !isEdgeToMatch( currEdge))
			continue;
		GR::ObjectWrapper srcVertex= currEdge.src_end();
		GR::ObjectWrapper dstVertex= currEdge.dst_end();
		if ( isBound( srcVertex) && isBound( dstVertex))
		{
			edge= currEdge;
			return true;
		}
	}
	return false;
}

bool PatternMatcher::edgeWithOneVertexBound( GR::ObjectLink& edge) const
{
	std::set< GR::ObjectLink> edges= _rule.ObjectLink_kind_children();
	for( std::set< GR::ObjectLink>::const_iterator itEdges= edges.begin(); itEdges!= edges.end(); ++itEdges)
	{
		const GR::ObjectLink& currEdge= *itEdges;
		if ( isEdgeTraversed( currEdge) || !isEdgeToMatch( currEdge))
			continue;
		GR::ObjectWrapper srcVertex= currEdge.src_end();
		GR::ObjectWrapper dstVertex= currEdge.dst_end();
		if ( ( isBound( srcVertex) && !isBound( dstVertex)) || ( !isBound( srcVertex) && isBound( dstVertex)))
		{
			edge= currEdge;
			return true;
		}
	}
	return false;
}

const Definition& PatternMatcher::parentBoundComposition( const GR::ObjectWrapper& parent, const GR::ObjectWrapper& child)
{
	Definition* setDef= new Definition;
//	if((string)parent.nameSpace()==(string)child.nameSpace())
//		setDef->createContainmentDef( ( std::string)child.paradigmName(), ( std::string)child.classType(), PE::createUniqueName( ( std::string)child.name()+ "s"), lowerBegin( getBoundObjectName( parent)));
//	else
	setDef->createContainmentDef( ( std::string)child.paradigmName(), (std::string)child.nameSpace(),( std::string)child.classType(), PE::createUniqueName( ( std::string)child.name()+ "s"), lowerBegin( getBoundObjectName( parent)), (string)parent.nameSpace()==(string)child.nameSpace());
	_currCode->add( *setDef);
	// negative application condition for composition 
	int childPatternCardinality= atoi( ((std::string)child.cardinality()).c_str());	// child po has pattern cardinality 0?
	if ( 0 == childPatternCardinality) {
		If* checkIfEmpty= new If;
		checkIfEmpty->createIfExprThenContinue( "!" + setDef->_variable+ ".empty()");
		_currCode->add( *checkIfEmpty);
	}
	Loop* childLoop= new Loop;
	childLoop->createIteratorLoop( PE::createUniqueName( IT+ ( std::string)child.name()), lowerBegin( setDef->_variable), setDef->_type);
	_currCode->add( *childLoop);
	_currCode= childLoop;
	Definition* currChild= new Definition;
//	currChild->_type= ( std::string)child.paradigmName()+ RES+ ( std::string)child.classType();
	currChild->_type= PE::getFullType(child);
	currChild->_variable= PE::createUniqueName( CURR+ ( std::string)child.name());
	currChild->_initValue= PTR+ childLoop->_loopVariable;
	_currCode->add( *currChild);
	// Identity check:
	// if there are already same type children bound to this parent, extra If PE-s must be added.
//	const std::list< const Definition*>* children= NULL;
//	if( _rule.uniqueIdentity() && NULL!= ( children= _boundObjects.getAllObjectChildren( parent)))
//		checkChildrenIdentities( *currChild, *children);
	// add this lately created child definition to the parent info,
	// in order to explore same-type occurences later.
	_boundObjects.addObjectChild( parent, *currChild);
	return *currChild;
}

const Definition& PatternMatcher::childBoundComposition( const GR::ObjectWrapper& parent, const GR::ObjectWrapper& child)
{
	Definition* parentDef= new Definition;
	parentDef->_type= "Udm::Object";
	parentDef->_variable= PE::createUniqueName( ( std::string)child.name()+ "Parent");
	// !!VIZA changed this line of code to the commented line after UDM has supported Udm::Object::container()
	//parentDef->_initValue= lowerBegin( _boundObjects.getBoundObjectName( child))+ ".parent()";
	parentDef->_initValue= lowerBegin( getBoundObjectName( child))+ ".container()";
	_currCode->add( *parentDef);
	If* typeCheck= new If;
	typeCheck->createIfTypeMismatchThenContinue( lowerBegin( parentDef->_variable), ( std::string)parent.paradigmName(), (std::string)parent.nameSpace(), ( std::string)parent.classType());
	_currCode->add( *typeCheck);
	Definition* castParent= new Definition;
//	castParent->_type= ( std::string)parent.paradigmName()+RES + ( std::string)parent.classType();
	castParent->_type= PE::getFullType(parent);
	castParent->_variable= PE::createUniqueName( ( std::string)parent.name()+ ( std::string)child.name());
//	castParent->_initValue= ( std::string)parent.paradigmName()+ RES+ ( std::string)parent.classType()+ RES+ "Cast("+ SEP+ lowerBegin( parentDef->_variable)+ ")";
	castParent->_initValue= PE::getFullType(parent) + RES+ "Cast("+ SEP+ lowerBegin( parentDef->_variable)+ ")";
	_currCode->add( *castParent);
	_boundObjects.addObjectChild( parent, *_boundObjects.getBoundObjectDefinition( child));	// child must be bound.
	return *castParent;
}

void PatternMatcher::bothBoundComposition( const GR::ObjectWrapper& parent, const GR::ObjectWrapper& child)
{
	// Have 'parent' to be bound. This way, identity check for children is treated correctly in 'parentBoundComposition()'.
	const std::list< const Definition*>* references= NULL;
	if ( ( NULL== ( references= _boundObjects.getAllObjectReferences( child))) || ( NULL== _boundObjects.getAllObjectReferences( parent)))
		throw CGEx("FATAL: CG internal error. No reference to either parent or child in a both bound composition.");
	const Definition& childObjectCreated= parentBoundComposition( parent, child);
	validateReferences( childObjectCreated, *references);
}

const Definition& PatternMatcher::matchAnyAssociation( const GR::ObjectLink& anyAssocLink) {
	bool isSrcBound= isBound( anyAssocLink.src_end());
	GR::ObjectWrapper bound= isSrcBound ? anyAssocLink.src_end() : anyAssocLink.dst_end();
	GR::ObjectWrapper unbound= isSrcBound ? anyAssocLink.dst_end() : anyAssocLink.src_end();
	return createAnyAssociationCommonCode( bound, unbound, GR::ObjectWrapper());
}

const Definition& PatternMatcher::createAnyAssociationCommonCode( const GR::ObjectWrapper& bound, const GR::ObjectWrapper& unbound, const GR::ObjectWrapper& assocClass) {

//	std::string unboundType= ( std::string)unbound.paradigmName()+ RES+ ( std::string)unbound.classType();
	std::string unboundType= PE::getFullType(unbound);
	std::string unboundName= unbound.name();
	bool isAssocInfo= assocClass;	// !=null?
	Definition* aiDef= 0;
	std::string acType;
	if ( isAssocInfo){
		aiDef= new Definition;
		aiDef->_type= "Udm::Object::AssociationInfo";
		aiDef->_variable= PE::createUniqueName( "ai_"+ (std::string)assocClass.name());
//		acType= (std::string)assocClass.paradigmName() + RES + (std::string)assocClass.classType();
		acType= PE::getFullType(assocClass);
		aiDef->addArg( acType+ RES + "meta");
		_currCode->add( *aiDef);
	}
	// Create container definition which holds objects of type 'unboundType' that are associated with 'bound'.
	Definition* unboundsDef= new Definition;
	unboundsDef->_type= isAssocInfo ? "multiset< pair< Udm::Object, Udm::Object> >" : "set< Udm::Object>";
	unboundsDef->_variable= PE::createUniqueName( unboundName+ "s");
	unboundsDef->_initValue= lowerBegin( getBoundObjectName( bound))+ (isAssocInfo ? ".GetAdjacentObjectsWithAssocClasses(" : ".GetAdjacentUniqueObjects(")+ SEP+ unboundType+ RES+ "meta"+ (isAssocInfo ? (","+ SEP+ aiDef->_variable) : "") +")";
	_currCode->add( *unboundsDef);
	// Create loop which iterates std::set of unbound objects.
	Loop* associationsLoop= new Loop;
	associationsLoop->createIteratorLoop( PE::createUniqueName( IT+ unboundName+ "s"), unboundsDef->_variable, unboundsDef->_type);
	_currCode->add( *associationsLoop);
	_currCode= associationsLoop;
	// type-cast iterator
	Definition* unboundDef= new Definition;
	unboundDef->createTypeCast( unboundType, PE::createUniqueName( unboundName), isAssocInfo ? (associationsLoop->_loopVariable+ "->first"): (PTR+ associationsLoop->_loopVariable));
	_currCode->add( *unboundDef);
	//
	if ( isAssocInfo) {
		Definition* acDef= new Definition;
		acDef->createTypeCast( acType, PE::createUniqueName( assocClass.name()), associationsLoop->_loopVariable+ "->second");
		_currCode->add( *acDef);
		bind( assocClass);
		_boundObjects.addObjectReference( assocClass, *acDef);
	}
	return *unboundDef;
}

// OW1--AC--OW2
void PatternMatcher::associationClassAny( const GR::ObjectLink& assocClassLink) {
	const GR::ObjectWrapper& src= assocClassLink.src_end();
	const GR::ObjectWrapper& dst= assocClassLink.dst_end();
	bool isSrcAssocClass= src.isAssociationClass();
	const GR::ObjectWrapper& AC= isSrcAssocClass ? src : dst;
	const GR::ObjectWrapper& OW1= isSrcAssocClass ? dst : src;
	GR::ObjectWrapper OW2;	// its value is assigned later based on some condition
	GR::ObjectLink otherACLink; // its value is assigned later based on some condition
	if ( isSrcAssocClass) {
		set< GR::ObjectLink> ols= AC.src();
		if ( 1 != ols.size()) {
			throw CGEx("Association class " + ( std::string)src.name()+ "is connected to multiple connectors.");
		}
		otherACLink= *ols.begin();
		OW2= otherACLink.src_end();
	} else {
		set< GR::ObjectLink> ols= AC.dst();
		if ( 1 != ols.size()) {
			throw CGEx("Association class " + ( std::string)src.name()+ "is connected to multiple connectors.");
		}
		otherACLink= *ols.begin();
		OW2= otherACLink.dst_end();
	}
	bool isOW1Bound= isBound( OW1);
	bool isOW2Bound= isBound( OW2);
	bool isACBound= isBound( AC);
	if ( !isOW1Bound && !isOW2Bound) {
		printError("AssociationClassAny relationship where only the AssociationClass is bound, is not supported.");
		return;
	}
	// AnyAssociationCommonCode
	const Definition& unboundDef= createAnyAssociationCommonCode( isOW1Bound? OW1 : OW2, isOW1Bound? OW2 : OW1, AC);
	// both bound association?
	if ( isOW1Bound && isOW2Bound) {
		// since OW1 was the source of object traversal code, it is OW2 that must be checked for identity
		const GR::ObjectWrapper& unbound= OW2;
		If* checkIdentity= new If;
		checkIdentity->createIfExprThenContinue( lowerBegin( unboundDef._variable)+ NEQ+ SEP+ lowerBegin( getBoundObjectName( unbound)));
		_currCode->add( *checkIdentity);
	}
	// bind both links, assoc class object and end objects
	setEdgeTraversed( assocClassLink);
	setEdgeTraversed( otherACLink);
	const GR::ObjectWrapper& unbound= isOW1Bound? OW2 : OW1;
	bind( unbound);
	_boundObjects.addObjectReference( unbound, unboundDef);
}


void PatternMatcher::bothBoundMatchAnyAssociation( const GR::ObjectLink& anyAssocLink) {
	const Definition& unboundDef= matchAnyAssociation( anyAssocLink);
	// check identity
	GR::ObjectWrapper unbound= anyAssocLink.dst_end();
	If* checkIdentity= new If;
	checkIdentity->createIfExprThenContinue( lowerBegin( unboundDef._variable)+ NEQ+ SEP+ lowerBegin( getBoundObjectName( unbound)));
	_currCode->add( *checkIdentity);
}


// This monstrum's gettin' incomprehensible.
// Reason is, there are several different codes created for several different cases:
// 1, simple association specified between two non-association classes
// 2, simple association specified between bound non-association and unbound association classes.
// 3, simple association specified between two non-association classes, but the link is an association class in the host meta.
const Definition& PatternMatcher::simpleAssociation( const GR::ObjectLink& simpleAssocLink)
{
	bool isSrcBound= isBound( simpleAssocLink.src_end());
	std::string dstRoleName= ( isSrcBound ? ( std::string)simpleAssocLink.dstRoleName() : ( std::string)simpleAssocLink.srcRoleName());
	GR::ObjectWrapper bound= isSrcBound ? simpleAssocLink.src_end() : simpleAssocLink.dst_end();
	GR::ObjectWrapper unbound= isSrcBound ? simpleAssocLink.dst_end() : simpleAssocLink.src_end();
// Generated UDM API implementation.
	// Gain meta information about the link.
	int maxMultiplicityAssoc= 0;
	Uml::Class clTarget;
	Uml::Association assoc;
	bool isCrossLink= false;// ( std::string) simpleAssocLink.relationshipType() == "ObjectMapping";
	std::string targetName; // fill this later..
	std::string targetClass;
	std::string targetClass_namespace;
	std::string assocClass;
	std::string unboundType= ( std::string)unbound.paradigmName()+ RES;
	std::string ns_name;
//	if ( true ==getAssociationProperties( _backend, bound.paradigmName(), bound.classType(), dstRoleName, bound.isAssociationClass(), clTarget, assoc, maxMultiplicityAssoc, isCrossLink))
	if ( true ==getAssociationProperties( _backend, bound.paradigmName(), bound.nameSpace(), bound.classType(), dstRoleName, bound.isAssociationClass(), clTarget, assoc, maxMultiplicityAssoc, isCrossLink))
	{
		assocClass= ( std::string) assoc.name();
		targetClass= ( std::string) clTarget.name();
		Uml::Namespace target_ns = clTarget.parent_ns();
		if(target_ns!=&Udm::_null)
		{
			targetClass_namespace = target_ns.name();
			vector<Uml::Namespace> ns_parents = Uml::GetParentNamespaces(target_ns);
			for(vector<Uml::Namespace>::iterator i=ns_parents.begin(); i!=ns_parents.end();++i)
			{
				targetClass_namespace = (std::string)(*i).name()+"::"+targetClass_namespace;
			}
		}

		if ( isCrossLink)
		{
			// remove '_diagram' suffix from class name;
			std::string from= clTarget.from();
			if ( !from.empty())
			{
				string::size_type colonPos= from.find( ":");
				if ( string::npos != colonPos) {										
					// contains namespace
					std::string ns = from.substr(colonPos+1, from.length()-1);
					from= from.erase( colonPos, from.size()-colonPos);

					//get the namespace
					colonPos= ns.find_last_of( ":");
					while( string::npos != colonPos)
					{
						targetClass_namespace = "::"+ns.substr(colonPos+1, ns.length()-1)+targetClass_namespace;
						ns = ns.erase(colonPos, ns.length()-colonPos);
						colonPos= ns.find_last_of( ":");
					}
					targetClass_namespace = ns+targetClass_namespace;
				}
//				from.insert( from.begin(), '_');
//				printLogMsg( from + ", " + targetClass);
				string::size_type fromPos= targetClass.find( from);
				targetClass.erase( fromPos, targetClass.size()-fromPos);
				// remove "_cross_ph" suffix
				std::string temp( Udm::cross_delimiter);
				std::string::size_type crossDelimPos= targetClass.find( Udm::cross_delimiter);
				targetClass.erase( crossDelimPos, targetClass.size()-crossDelimPos);
				assocClass= targetClass;
//				printLogMsg( "->" + targetClass);
				
			}
		}
	}
	else
	{
		printError("Association role: "+ dstRoleName+" not found for"+ ( bound.isAssociationClass() ? " association" : "")+ " class: "+ (std::string)bound.classType()+ ".");
		// lack of better, use the names specified in the rule.
		targetClass= ( std::string) unbound.name();
		assocClass= ( std::string) unbound.name();
		ns_name = (std::string)unbound.name();
	}
	// is source association  association class?
	bool is_sa_ac= assoc && ( Uml::IsAssocClass( assoc)) && !unbound.isAssociationClass(); // association class specified as simple association
	if ( is_sa_ac)
	{
		printLogMsg("INFO: Association class specified in the meta <-> simple association specified in the pattern.");
		unboundType+= targetName= targetClass= assocClass;
	}
	else
	{
		// Feng: consider the scenario that an association class is also an end of simple associaiton
		if ( unbound.isAssociationClass() && !assocClass.empty())
		{
			targetClass= assocClass;
		}
		targetName= ( std::string) unbound.name();
		if(!((std::string)unbound.nameSpace()).empty())
		{
			ns_name = unbound.nameSpace();
			unboundType= unboundType+ns_name+RES+( std::string) unbound.classType();
		}
		else
			unboundType+= ( std::string) unbound.classType();
	}
	// Create 'Get std::set of unbound objects'.
	Definition* unboundDef= new Definition;
	unboundDef->_type= unboundType;
	// Different code generated for different maximum multiplicity. (Because UDM API makes difference) 
	bool isMultipleAssoc= 1< (unsigned int) maxMultiplicityAssoc; // * multiplicity is represented as -1.
	if ( isMultipleAssoc)	
	{
		Definition* unboundsDef= new Definition;		
		unboundsDef->createRoles( ( std::string)unbound.paradigmName(), targetClass_namespace, targetClass, lowerBegin(PE::createUniqueName( targetName+ "s")), lowerBegin( getBoundObjectName( bound)), dstRoleName);
		_currCode->add( *unboundsDef);
		// negative application condition for associations with multiple cardinality
		int linkPatternCardinality= atoi( ((std::string)simpleAssocLink.cardinality()).c_str());	// this is _pattern_ cardinality
		if ( 0 == linkPatternCardinality) {
			If* checkIfEmpty= new If;
			checkIfEmpty->createIfExprThenContinue( "!" + unboundsDef->_variable+ ".empty()");
			_currCode->add( *checkIfEmpty);
		}
		// Create loop which iterates std::set of unbound objects.
		Loop* associationsLoop= new Loop;
		associationsLoop->createIteratorLoop( PE::createUniqueName( IT+ targetName+ "s"), unboundsDef->_variable, unboundsDef->_type);
		_currCode->add( *associationsLoop);
		_currCode= associationsLoop;
		// Create 'current unbound object'.
		if ( is_sa_ac)
		{
			GR::ObjectWrapper owNewAssocClass;
			createPseudoAssocClass( simpleAssocLink, ( std::string) unbound.paradigmName(), assocClass, owNewAssocClass);
			unbound= owNewAssocClass;
		}
		unboundDef->_variable= PE::createUniqueName( "curr"+ ( std::string)unbound.name());
		if ( ( std::string) unbound.classType()== targetClass)
		{
			// No need of type check, and cast. Just simply assign.
			unboundDef->_initValue= PTR+ associationsLoop->_loopVariable;
		}
		else
		{
			// Create type check of unbound object
			If* typeCheck= new If;
			typeCheck->createIfTypeMismatchThenContinue( "("+ PTR+ associationsLoop->_loopVariable+ ")", ( std::string)unbound.paradigmName(), (std::string)unbound.nameSpace(), ( std::string)unbound.classType());
			_currCode->add( *typeCheck);
			unboundDef->_initValue= unboundType+ RES+ "Cast("+ SEP+ PTR+ associationsLoop->_loopVariable+ ")";
		}
	}
	else
	{
		unboundDef->createRole( ( std::string)unbound.paradigmName(), ( std::string)unbound.nameSpace(), targetClass, lowerBegin( PE::createUniqueName( targetName)), lowerBegin( getBoundObjectName( bound)), dstRoleName);
		if ( is_sa_ac)
		{
			GR::ObjectWrapper owNewAssocClass;
			createPseudoAssocClass( simpleAssocLink, ( std::string) unbound.paradigmName(), assocClass, owNewAssocClass);
			unbound= owNewAssocClass;
		}
		// negative application condition for associations with single cardinality
		int linkPatternCardinality= atoi( ((std::string)simpleAssocLink.cardinality()).c_str());	// this is _pattern_ cardinality
		if ( 0 == linkPatternCardinality) {
			If* checkIfNull= new If;
			checkIfNull->createIfExprThenContinue( unboundDef->_variable);
			_currCode->add( *checkIfNull);
		}
		// same types
		if ( ( std::string) unbound.classType()== targetClass)
		{
			// No need of type check, and cast. Just simply use this.
		}
		else
		{
			_currCode->add( *unboundDef);
			std::string nameToCast= unboundDef->_variable;
			// Create type check of unbound object
			If* typeCheck= new If;
			typeCheck->createIfTypeMismatchThenContinue( nameToCast, ( std::string)unbound.paradigmName(), (std::string)unbound.nameSpace(), ( std::string)unbound.classType());
			_currCode->add( *typeCheck);
			unboundDef= new Definition;
			unboundDef->createTypeCast( unboundType, lowerBegin( PE::createUniqueName( ( std::string) unbound.name())), nameToCast);
		}
	}
	_currCode->add( *unboundDef);
	// add code, which checks the result for udm::Null
	if ( false== isMultipleAssoc)
	{
		If* checkIfNull= new If;
		checkIfNull->_expression= "!"+ unboundDef->_variable;
		checkIfNull->add( *new Expression( "continue"));
		_currCode->add( *checkIfNull);
	}
	// Check for simple association link identities..
	// Regardless 'unbound' is association class or not, add as a simple association to 'bound'.
	const std::list< BoundObjectInfo::SimpleAssoc_t>* boundSimpleAssocs= NULL;
	BoundObjectInfo::SimpleAssoc_t currAssoc= make_pair( dstRoleName, ( const Definition*)unboundDef);
	if( NULL!= ( boundSimpleAssocs= _boundObjects.getAllObjectSimpleAssociations( bound)))
		checkSimpleAssocsIdentities( currAssoc, *boundSimpleAssocs);
	_boundObjects.addObjectSimpleAssociation( bound, currAssoc);
	if ( is_sa_ac)
	{
		_boundObjects.addObjectReference( unbound, *unboundDef);
		// The association class link must be created and processed implicitly.
		return createPseudoAssocLink( simpleAssocLink, unbound);
	}
	else
		return *unboundDef;
}

void PatternMatcher::bothBoundSimpleAssociation( const GR::ObjectLink& simpleAssocLink)
{
	const std::list< const Definition*>* references= NULL;
	GR::ObjectWrapper src= simpleAssocLink.src_end();
	GR::ObjectWrapper dst= simpleAssocLink.dst_end();
	if ( ( NULL== ( references= _boundObjects.getAllObjectReferences( dst))) || ( NULL== _boundObjects.getAllObjectReferences( src)))
		throw CGEx("FATAL: CG internal error. No reference to either src or dst in a both bound simple association.");
	// if both bound, 'dst' is considered to be unbound in 'simpleAssociation()'.
	const Definition& dstCreated= simpleAssociation( simpleAssocLink);
	validateReferences( dstCreated, *references);
}

const Definition& PatternMatcher::assocClassUnboundAssociation( const GR::ObjectLink& assocClassLink)
{
	return simpleAssociation( assocClassLink);
}

const Definition& PatternMatcher::assocClassBoundAssociation( const GR::ObjectLink& assocClassLink)
{
	bool isSrcAssocClass= GR::ObjectWrapper::Cast( assocClassLink.src_end()).isAssociationClass();
	bool isAssocClassBound= isBound( isSrcAssocClass ? assocClassLink.src_end() : assocClassLink.dst_end());
	if ( false== isAssocClassBound)
		throw CGEx("FATAL: CG internal error. Unbound association class error.");
	std::string dstRoleName= ( isSrcAssocClass ? ( std::string)assocClassLink.dstRoleName() : ( std::string)assocClassLink.srcRoleName());
	GR::ObjectWrapper bound= isSrcAssocClass ? assocClassLink.src_end() : assocClassLink.dst_end();
	GR::ObjectWrapper unbound= isSrcAssocClass ? assocClassLink.dst_end() : assocClassLink.src_end();
// Generated UDM API implementation.
	// Create 'Get association end: unbound object'.
	Definition* unboundDef= new Definition;
	// Gain meta information about the link.
	int maxMultiplicityAssoc= 0;
	Uml::Class clTarget;
	Uml::Association assoc;
	bool isCrossLink= false;//( std::string) assocClassLink.relationshipType() == "ObjectMapping";
	std::string targetClass;
	std::string assocClass;
//	if ( true == getAssociationProperties( _backend, bound.paradigmName(), bound.classType(), dstRoleName, bound.isAssociationClass(), clTarget, assoc, maxMultiplicityAssoc, isCrossLink))
	if ( true == getAssociationProperties( _backend, bound.paradigmName(), bound.nameSpace(), bound.classType(), dstRoleName, bound.isAssociationClass(), clTarget, assoc, maxMultiplicityAssoc, isCrossLink))
	{
		targetClass= ( std::string) clTarget.name();
		assocClass= ( std::string) assoc.name();
	}
	else
	{
		printError("Association role: "+ dstRoleName+" not found for"+ ( bound.isAssociationClass() ? " association" : "")+ " class: "+ (std::string)bound.classType()+ ".");
		// lack of better, use the names specified in the rule.
		targetClass= ( std::string) unbound.name();
		assocClass= ( std::string) unbound.name();
	}
	unboundDef->createRole( ( std::string)unbound.paradigmName(), ( std::string)unbound.nameSpace(), targetClass, PE::createUniqueName( ( std::string) unbound.name()), lowerBegin( getBoundObjectName( bound)), dstRoleName+ "_end");
	_currCode->add( *unboundDef);
	if ( ( std::string)unbound.classType()== targetClass)
	{
		// do nothing
	} else
	{
		// Create type check of unbound object. (if neccessary)
		If* typeCheck= new If;
		typeCheck->createIfTypeMismatchThenContinue( lowerBegin( unboundDef->_variable), ( std::string)unbound.paradigmName(), (std::string)unbound.nameSpace(), ( std::string)unbound.classType());
		_currCode->add( *typeCheck);
		// Create 'cast unbound object'.
		Definition* castUnboundDef= new Definition;
		//std::string unboundType= ( std::string)unbound.paradigmName()+ RES+ ( std::string)unbound.classType();
		std::string unboundType = PE::getFullType(unbound);
		castUnboundDef->_type= unboundType;
		castUnboundDef->_variable= PE::createUniqueName( "curr"+ ( std::string)unbound.name());
		castUnboundDef->_initValue= unboundType+ RES+ "Cast("+ SEP+ lowerBegin( unboundDef->_variable)+ ")";
		_currCode->add( *castUnboundDef);
		// tricky part: the casted unbound object is the unbound object.
		unboundDef= castUnboundDef;
	}
	// Association class identity check. 
	// Assoc.class triplet must be added to the bound object wrapper end of assoc.class. Find that:
	GR::ObjectWrapper boundOW= otherEndOfAssociationClass( bound, isSrcAssocClass);
	// Now 'boundOW' is the object wrapper, to which assoc.class triplet should be added, 
	// and which should be checked for assoc.class identity. 
	const std::list< BoundObjectInfo::AssocClass_t>* boundClassAssocs= NULL;
	BoundObjectInfo::AssocClass_t currAssoc= make_triplet( dstRoleName, _boundObjects.getBoundObjectDefinition( bound)/*assoc.class def.*/, ( const Definition*)unboundDef);
	if( NULL!= ( boundClassAssocs= _boundObjects.getAllObjectClassAssociations( boundOW)))
		checkClassAssocsIdentities( currAssoc, *boundClassAssocs);
	// Register association class info with the bound non-association class object wrapper.
	_boundObjects.addObjectAssocClass( boundOW, currAssoc);
	return *unboundDef;
}

GR::ObjectWrapper PatternMatcher::otherEndOfAssociationClass( const GR::ObjectWrapper& assocClass, bool getSources)
{
	std::set< GR::ObjectLink> links= getSources ? assocClass.src() : assocClass.dst();
	std::vector< GR::ObjectLink> assocClassLinks;	// "AssociationClass" type links
	for( std::set< GR::ObjectLink>::const_iterator i= links.begin(); i!= links.end(); ++i)
	{
		const GR::ObjectLink& currLink= *i;
		if ( "AssociationClass" == ( std::string) currLink.relationshipType())
			assocClassLinks.push_back( currLink);
	}
	if ( assocClassLinks.empty())
		throw CGEx("FATAL: Association class: "+ ( std::string)assocClass.name()+ " has no "+ (getSources ? "source." : "destination."));
	if ( assocClassLinks.size() > 1)
		printWarning("association class: "+ ( std::string)assocClass.name()+ " is associated with more then one classes. Identity check might not work correctly!");
	return getSources ? assocClassLinks.front().src_end() : assocClassLinks.front().dst_end();
}

void PatternMatcher::bothBoundAssociationClass( const GR::ObjectLink& assocClassLink)
{
	// Have association class to be bound. This way, multiple cardinality is treated correctly in 'assocClassBoundAssociation()'.
	const std::list< const Definition*>* references= NULL;
	GR::ObjectWrapper src= assocClassLink.src_end();
	GR::ObjectWrapper dst= assocClassLink.dst_end();
	if ( ( NULL== _boundObjects.getAllObjectReferences( dst)) || ( NULL== _boundObjects.getAllObjectReferences( src)))
		throw CGEx("FATAL: CG internal error. No reference to either src or dst in a both bound association class.");
	references= src.isAssociationClass() ? _boundObjects.getAllObjectReferences( dst) : _boundObjects.getAllObjectReferences( src);
	const Definition& owEndCreated= assocClassBoundAssociation( assocClassLink);
	validateReferences( owEndCreated, *references);
}

void PatternMatcher::translateEdgeWithBothVerticesBound( const GR::ObjectLink& edge)
{
	std::string type= edge.relationshipType();
	if ("Composition"== type)
	{
		GR::ObjectWrapper parent= edge.dst_end();
		GR::ObjectWrapper child= edge.src_end();
		bothBoundComposition( parent, child);
	} 
	else if ( "SimpleAssociation"== type)
	{
		bothBoundSimpleAssociation( edge);
	} 
	else if ( "AssociationClass"== type)
	{
		bothBoundAssociationClass( edge);
	} 
	else if ( "ObjectMapping"== type)
	{
		bothBoundSimpleAssociation( edge);
	} else if ( "SimpleAssociationAny" == type) {
		bothBoundMatchAnyAssociation( edge);
	} else if ( "ObjectMappingAny"== type) {
		bothBoundMatchAnyAssociation( edge);
	} else if ( "AssociationClassAny" == type) {
		associationClassAny( edge);
	}
	else
	{
		printError( "Unknown link relationship type encountered in pattern matcher: "+ type+ "."); // throw CGEx("Unknown link relationship type encountered in pattern matcher.");
	}
}

void PatternMatcher::translateEdgeWithOneVertexBound( const GR::ObjectLink& edge)
{
	std::string type= edge.relationshipType();
	if ("Composition"== type)
	{
		GR::ObjectWrapper parent= edge.dst_end();
		GR::ObjectWrapper child= edge.src_end();
		if ( isBound( parent))
		{
			const Definition& childObjectCreated= parentBoundComposition( parent, child);
			bind( child);
			_boundObjects.addObjectReference( child, childObjectCreated);
		}
		else
		{
			const Definition& parentObjectCreated= childBoundComposition( parent, child);
			bind( parent);
			_boundObjects.addObjectReference( parent, parentObjectCreated);
		}
	} 
	else if ( "SimpleAssociation"== type)
	{
		const Definition& unboundCreated= simpleAssociation( edge);
		GR::ObjectWrapper unbound= isBound( edge.src_end()) ? edge.dst_end() : edge.src_end();
		bind( unbound);
		_boundObjects.addObjectReference( unbound, unboundCreated);
	} 
	else if ( "AssociationClass"== type)
	{
		GR::ObjectWrapper bound= isBound( edge.src_end()) ? edge.src_end() : edge.dst_end();
		GR::ObjectWrapper unbound= isBound( edge.src_end()) ? edge.dst_end() : edge.src_end();
		bool isAssocClassBound= bound.isAssociationClass();
		if ( true== isAssocClassBound)
		{
			const Definition& owEndCreated= assocClassBoundAssociation( edge);
			bind( unbound);
			_boundObjects.addObjectReference( unbound, owEndCreated);
		}
		else
		{
			const Definition& assocClassCreated= assocClassUnboundAssociation( edge);
			bind( unbound);
			_boundObjects.addObjectReference( unbound, assocClassCreated);
		}
	} 
	else if ( "ObjectMapping" == type)
	{
		const Definition& unboundCreated= simpleAssociation( edge);
		GR::ObjectWrapper unbound= isBound( edge.src_end()) ? edge.dst_end() : edge.src_end();
		bind( unbound);
		_boundObjects.addObjectReference( unbound, unboundCreated);
	} else if ( "SimpleAssociationAny" == type) {
		const Definition& unboundCreated= matchAnyAssociation( edge);
		GR::ObjectWrapper unbound= isBound( edge.src_end()) ? edge.dst_end() : edge.src_end();
		bind( unbound);
		_boundObjects.addObjectReference( unbound, unboundCreated);
	} else if ( "ObjectMappingAny"== type) {
		const Definition& unboundCreated= matchAnyAssociation( edge);
		GR::ObjectWrapper unbound= isBound( edge.src_end()) ? edge.dst_end() : edge.src_end();
		bind( unbound);
		_boundObjects.addObjectReference( unbound, unboundCreated);
	} else if ( "AssociationClassAny" == type) {
			associationClassAny( edge);
	}
	else
	{
		printError( "Unknown link relationship type encountered in pattern matcher: "+ type+ ".");// throw CGEx("Unknown link relationship type encountered in pattern matcher.");
	}
}

// Validation functions.
void PatternMatcher::validateReferences( const Definition& toCompare, const std::list< const Definition*>& references)
{
	std::string nameToCompareWith= toCompare._variable;
	for( std::list< const Definition*>::const_iterator itReferences= references.begin(); itReferences!= references.end(); ++itReferences)
	{
		const Definition& currReference= **itReferences;
		If* equality= new If;
		equality->_expression= lowerBegin( currReference._variable)+ NEQ+ SEP+ lowerBegin( nameToCompareWith);
		_currCode->add( *equality);
		Expression* contExp= new Expression;
		contExp->_expression= "continue";
		equality->add( *contExp);
	}
}

void PatternMatcher::checkChildrenIdentities( const Definition& newChild, const std::list< const Definition*>& adjacentChildren)
{
	If* checkNotSameInstances= new If;
	for( std::list< const Definition*>::const_iterator itChildren= adjacentChildren.begin(); itChildren!= adjacentChildren.end(); ++itChildren)
	{
		const Definition& currChild= **itChildren;
		if ( true== sameTypes( newChild._type, currChild._type))
		{
			// append if expression
			checkNotSameInstances->appendWithAnd( lowerBegin( currChild._variable)+ NEQ+ SEP+ lowerBegin( newChild._variable));
		}
	}
	// So, any expressions? If yes, add to the code.
	if ( 0!= checkNotSameInstances->isExpression())
	{
		CommentLine* commentMulCard= new CommentLine;
		commentMulCard->_comment= "Composition identity check.";
		_currCode->add( *commentMulCard);
		_currCode->add( *checkNotSameInstances);
		_currCode= checkNotSameInstances;
	}
}

void PatternMatcher::checkSimpleAssocsIdentities( const BoundObjectInfo::SimpleAssoc_t& newAssoc, const std::list< BoundObjectInfo::SimpleAssoc_t>& adjacentAssocs)
{
	If* checkNotSameInstances= new If;
	for( std::list< BoundObjectInfo::SimpleAssoc_t>::const_iterator itAssocs= adjacentAssocs.begin(); itAssocs!= adjacentAssocs.end(); ++itAssocs)
	{
		const BoundObjectInfo::SimpleAssoc_t& currAssoc= *itAssocs;
		if ( ( newAssoc.first== currAssoc.first) && ( true== sameTypes( newAssoc.second->_type, currAssoc.second->_type)))
		{
			// append if expression
			checkNotSameInstances->appendWithAnd( lowerBegin( currAssoc.second->_variable)+ NEQ+ SEP+ lowerBegin( newAssoc.second->_variable));
		}
	}
	// So, any expressions? If yes, add to the code.
	if ( 0!= checkNotSameInstances->isExpression())
	{
		CommentLine* commentMulCard= new CommentLine;
		commentMulCard->_comment= "Simple association identity check.";
		_currCode->add( *commentMulCard);
		_currCode->add( *checkNotSameInstances);
		_currCode= checkNotSameInstances;
	}
}

void PatternMatcher::checkClassAssocsIdentities( const BoundObjectInfo::AssocClass_t& newClassAssoc, const std::list< BoundObjectInfo::AssocClass_t>& adjacentAssocs)
{
	If* checkNotSameInstances= new If;
	for( std::list< BoundObjectInfo::AssocClass_t>::const_iterator itAssocs= adjacentAssocs.begin(); itAssocs!= adjacentAssocs.end(); ++itAssocs)
	{
		const BoundObjectInfo::AssocClass_t& currAssoc= *itAssocs;
		if (( newClassAssoc.first== currAssoc.first) &&
			// Only if assoc.classes are not of same type; identity is checked in checkSimpleAssocsIdentities(), otherwise.
			( false== sameTypes( newClassAssoc.second->_type, currAssoc.second->_type)) && 
			( true== sameTypes( newClassAssoc.third->_type, currAssoc.third->_type)))
		{
			// Append if expression with identity check of unbound object wrappers!
			checkNotSameInstances->appendWithAnd( lowerBegin( currAssoc.third->_variable)+ NEQ+ SEP+ lowerBegin( newClassAssoc.third->_variable));
		}
	}
	// So, any expressions? If yes, add to the code.
	if ( 0!= checkNotSameInstances->isExpression())
	{
		CommentLine* commentMulCard= new CommentLine;
		commentMulCard->_comment= "Class association identity check.";
		_currCode->add( *commentMulCard);
		_currCode->add( *checkNotSameInstances);
		_currCode= checkNotSameInstances;
	}
}

/*!!VIZA this is temporary, and should be replace by the corrected one in UmlExt.cpp*/
/*
static set< Uml::AssociationRole> AncestorCrossAssociationTargetRoles(const Uml::Class &c, const Uml::Diagram & cross_dgr)
{
	using namespace Uml;
	set<AssociationRole> ret;
	set<Uml::Class> anc_cs = AncestorClasses(c);
	set<Uml::Class>::iterator i = anc_cs.begin();
	while (i != anc_cs.end())
	{
		Uml::Class a_i = *i;
		Uml::Class ph_a_i = classByName(cross_dgr, a_i.name());
		std::string a_i_name= a_i.name();
		if ( ph_a_i)
		{
			std::string name= ph_a_i.name();
			set<AssociationRole> ret_i = AssociationTargetRoles(ph_a_i);
			int size_ret_i= ret_i.size();
			ret.insert(ret_i.begin(), ret_i.end());
		}
		i++;
	};

	return ret;
}
*/

void PatternMatcher::createPseudoAssocClass( 
	const GR::ObjectLink& simpleLink, 
	const std::string& paradigmName, 
	const string& classType, 
	GR::ObjectWrapper& assocClass
	) const
{
	// Create the association class, and fill its attributes.
	assocClass= GR::ObjectWrapper::Create( simpleLink.parent());
	assocClass.name()= PE::createUniqueName( classType+ "Pseudo");
	assocClass.action()= "Bound";
	assocClass.paradigmName()= paradigmName;
	assocClass.classType()= classType;
	assocClass.isAssociationClass()= true;
	assocClass.dataNetworkID() = -1;
	// Create a simple association type link...
	GR::ObjectLink assocClassLink= GR::ObjectLink::Create( simpleLink.parent());
	assocClassLink.relationshipType()= "SimpleAssociation";
	assocClassLink.srcRoleName()= simpleLink.srcRoleName();
	assocClassLink.dstRoleName()= simpleLink.dstRoleName();
	// ...and assign its ends to the association class just created and the appropriate end of the pattern link.
	bool isSrcBound= isBound( simpleLink.src_end());
	if ( isSrcBound)
	{
		assocClassLink.src_end()= simpleLink.src_end();
		assocClassLink.dst_end()= assocClass;
	}
	else
	{
		assocClassLink.src_end()= assocClass;
		assocClassLink.dst_end()= simpleLink.dst_end();
	}
	// Bind the association class, because it is inserted into the match.
	bind( assocClass);
}

const Definition& PatternMatcher::createPseudoAssocLink( const GR::ObjectLink& simpleAssocLink, const GR::ObjectWrapper& assocClass)
{
	// Create association class link...
	GR::ObjectLink assocClassLink= GR::ObjectLink::Create( simpleAssocLink.parent());
	assocClassLink.relationshipType()= "AssociationClass";
	assocClassLink.srcRoleName()= simpleAssocLink.srcRoleName();
	assocClassLink.dstRoleName()= simpleAssocLink.dstRoleName();
	//
	GR::ObjectLink idCheckAssocClassLink= GR::ObjectLink::Create( simpleAssocLink.parent());
	idCheckAssocClassLink.relationshipType()= "AssociationClass";
	idCheckAssocClassLink.srcRoleName()= simpleAssocLink.srcRoleName();
	idCheckAssocClassLink.dstRoleName()= simpleAssocLink.dstRoleName();
	// ...and assign its ends to the association class 'assocClass' and the appropriate end of the pattern link.
	bool isSrcBound= isBound( simpleAssocLink.src_end());
	if ( isSrcBound)
	{
		assocClassLink.src_end()= assocClass;
		assocClassLink.dst_end()= simpleAssocLink.dst_end();
		// for identiy check
		idCheckAssocClassLink.src_end()= simpleAssocLink.src_end();
		idCheckAssocClassLink.dst_end()= assocClass;
	}
	else
	{
		assocClassLink.src_end()= simpleAssocLink.src_end();
		assocClassLink.dst_end()= assocClass;
		// for identiy check
		idCheckAssocClassLink.src_end()= assocClass;
		idCheckAssocClassLink.dst_end()= simpleAssocLink.dst_end();
	}
	// call association class bound pattern matching on the created link.
	return assocClassBoundAssociation( assocClassLink);
}

bool PatternMatcher::sameTypes( const std::string& type1, const std::string& type2) const
{
	// Extract type, and paradigm name.
	std::string::size_type firstRESpos= type1.find( RES);
	std::string::size_type lastRESpos= type1.rfind( RES);
	std::string type1Paradigm= type1.substr( 0, firstRESpos);
	std::string type1ClassName= type1.substr( lastRESpos+ std::string( RES).size(), /*up to*/type1.size());
	firstRESpos= type2.find( RES);
	lastRESpos= type2.rfind( RES);
	std::string type2Paradigm= type2.substr( 0, firstRESpos);
	std::string type2ClassName= type2.substr( lastRESpos+ std::string( RES).size(), /*up to*/type2.size());
	if ( type1Paradigm != type2Paradigm)
		return false;
	// Get meta information from the backend.
	Uml::Diagram diag= _backend.getDiagram( type1Paradigm);
	if ( diag== &Udm::_null)
		return false;
//	Uml::Class type1Class= Uml::classByName( diag, type1ClassName);
//	Uml::Class type2Class= Uml::classByName( _backend.getDiagram( ( std::string) type2Paradigm), type2ClassName);
	Uml::Class type1Class= Uml::classByPath( diag, type1ClassName);
	Uml::Class type2Class= Uml::classByPath( _backend.getDiagram( ( std::string) type2Paradigm), type2ClassName);

	if ( true== Uml::IsDerivedFrom( type1Class, type2Class))
		return true;
	else 
		return Uml::IsDerivedFrom( type2Class, type1Class);
}

// Edge traverse maintenance functions.
bool PatternMatcher::isEdgeTraversed( const GR::ObjectLink& edge) const
{
	EdgeSet_t::const_iterator itFind= _traversedEdges.find( edge);
	if ( itFind != _traversedEdges.end())
		return true;
	return false;
}

void PatternMatcher::setEdgeTraversed( const GR::ObjectLink& edge)
{
	_traversedEdges.insert( edge);
}

const std::string& PatternMatcher::getBoundObjectName( const GR::ObjectWrapper& ow)
{
	const std::string& name= _boundObjects.getBoundObjectName( ow);
	if ( name.empty()) {
//		std::string paradigmPlusClassName= ( std::string) ow.paradigmName() + RES+ ( std::string) ow.classType();
		std::string paradigmPlusClassName= PE::getFullType(ow);
		Definition* tempRootDef= _backend.getTempRootDef( paradigmPlusClassName);
		if ( tempRootDef) {
			bind( ow);
			_boundObjects.addObjectReference( ow, *tempRootDef);
			return tempRootDef->_variable;
		} else {
			printError(" Object "+ ( std::string)ow.name() + " is not bound error.");
		}
	}
	return name;
}
///////////////////////////////////////////////////////////////////////////////
// Bound Object Info maintenance functions.
void BoundInfo::addObjectReference( const GR::ObjectWrapper& ow, const Definition& referDef)
{
	_boundObjects[ ow]._references.push_back( &referDef);
}

void BoundInfo::addObjectChild( const GR::ObjectWrapper& ow, const Definition& childDef)
{
	_boundObjects[ ow]._children.push_back( &childDef);
}

void BoundInfo::addObjectSimpleAssociation( const GR::ObjectWrapper& ow, const BoundObjectInfo::SimpleAssoc_t& assocDef)
{
	_boundObjects[ ow]._simpleAssociations.push_back( assocDef);
}

void BoundInfo::addObjectAssocClass( const GR::ObjectWrapper& ow, const BoundObjectInfo::AssocClass_t& assocDef)
{
	_boundObjects[ ow]._classAssociations.push_back( assocDef);
}

const BoundObjectInfo* BoundInfo::getObjectInfo( const GR::ObjectWrapper& ow) const
{
	std::map< GR::ObjectWrapper, BoundObjectInfo>::const_iterator itFind= _boundObjects.find( ow);
	if ( itFind == _boundObjects.end())
	{
		return NULL;
	}
	return &( itFind->second);
}

const std::list< const Definition*>* BoundInfo::getAllObjectReferences( const GR::ObjectWrapper& ow) const 
{
	const BoundObjectInfo* objectInfo= NULL;
	return ( NULL== ( objectInfo= getObjectInfo( ow))) ? NULL : &objectInfo->_references;
}

const std::list< const Definition*>* BoundInfo::getAllObjectChildren( const GR::ObjectWrapper& ow) const
{
	const BoundObjectInfo* objectInfo= NULL;
	return ( NULL== ( objectInfo= getObjectInfo( ow))) ? NULL : &objectInfo->_children;
}

const std::list< BoundObjectInfo::SimpleAssoc_t>* BoundInfo::getAllObjectSimpleAssociations( const GR::ObjectWrapper& ow) const
{
	const BoundObjectInfo* objectInfo= NULL;
	return ( NULL== ( objectInfo= getObjectInfo( ow))) ? NULL : &objectInfo->_simpleAssociations;
}

const std::list< BoundObjectInfo::AssocClass_t>* BoundInfo::getAllObjectClassAssociations( const GR::ObjectWrapper& ow) const
{
	const BoundObjectInfo* objectInfo= NULL;
	return ( NULL== ( objectInfo= getObjectInfo( ow))) ? NULL : &objectInfo->_classAssociations;
}

const std::string& BoundInfo::getBoundObjectName( const GR::ObjectWrapper& ow) const 
{
	const Definition* firstReference= getBoundObjectDefinition( ow);
	if ( NULL == firstReference)
	{
		//  static const string literals are not elegant, but must return a reference for efficency reasons
		static const std::string EMPTY_STRING= "";	
		return EMPTY_STRING;
	}
	return firstReference->_variable;
}

const Definition* BoundInfo::getBoundObjectDefinition( const GR::ObjectWrapper& ow) const
{
	const std::list< const Definition*>* references= getAllObjectReferences( ow);
	if ( ( NULL == references) || ( references->empty()))
		return NULL;
	return references->front();
}
