#include <UmlExt.h>	// !!VIZA this #include must be the first, otherwise it won't compile!
#include "Translator.h"
#include "CGEx.h"
#include "Logging.h" // temporaly logger
#include "PEStringTable.h"
#include "GReATUtils.h"
#include "PatternMatcher.h"
#include "Effector.h"
#include "GroupEffector.h"
#include "InputFilter.h"
#include "OutputAppender.h"
#include "OutputSorter.h"
#include "DistinguishedOutputSorter.h"
#include <iterator>

using namespace std;

void Translator::translateBlock( const GRBlock& block)
{
	Class* blockDef= NULL;
	Function* funcOperator= NULL;
	translateRuleDef( block, &blockDef, &funcOperator);
	translateOutputs( block, blockDef, funcOperator);
}

void Translator::translateForBlock( const GRForBlock& forBlock)
{
	Class* blockDef= NULL;
	Function* funcOperator= NULL;
	translateRuleDef( forBlock, &blockDef, &funcOperator);
	translateOutputs( forBlock, blockDef, funcOperator);
	translateForCall( forBlock, blockDef, funcOperator);
}

// decorator functor for bound objects
class DecoratorFunctor
{
public:
	std::pair< GR::ObjectWrapper, std::string> operator()( const std::pair< GR::ObjectWrapper, std::string>& boundObject)
	{
		GR::ObjectWrapper ow= boundObject.first;
//		std::string name= std::string( ow.paradigmName())+ RES+ std::string( ow.classType())+ RES+ "Cast("+ SEP+ PTR+ boundObject.second+ ")";
		std::string name= PE::getFullType(ow) + RES+ "Cast("+ SEP+ PTR+ boundObject.second+ ")";
		return std::make_pair( ow, name);
	}
};
typedef DecoratorFunctor  decorateNamesWithCast;

// printer functor for bound objects
class PrinterFunctor
{
public:
	void operator()( const std::pair< GR::ObjectWrapper, std::string>& boundObject)
	{
		std::cout << ( std::string) boundObject.first.name() << "\t" << boundObject.second << std::endl;
	}
};
typedef PrinterFunctor printObject;

void Translator::translateRule( const GRRule& rule)
{
	if ( isIsolated( rule))
	{
		// if rule is isolated (not connected) do not generate class definition!
		if ( isReferred( rule))
			printError( "Rule"+ ( std::string)rule.name()+ " is isolated, and also referred by a proxy.");
		else
			printWarning( "Rule "+ ( std::string)rule.name()+ " is isolated, no code is generated.");
		return;	
	}
	Class* ruleDef= NULL;
	Function* funcOperator= NULL;
	translateRuleDef( rule, &ruleDef, &funcOperator);
	translateOutputs( rule, ruleDef, funcOperator);
	std::map< GR::ObjectWrapper, std::string> boundObjects;
	Function* pipFunc= 0;
	Loop* inputLoop= 0;
	std::map< GR::ObjectWrapper, std::string> inputBoundObjects;	// 

	translatePatterns( rule, ruleDef, funcOperator, boundObjects, pipFunc, inputLoop, &inputBoundObjects);
	// create code which call effector conditionally, and clear matches.
	If* matchCond= 0;
	inputLoop->add( *(matchCond= new If));
	matchCond->_expression= ISMATCH;
	inputLoop->add( *new Expression(_MATCHES+ ".clear()"));
	//
	PEComposite* effectorCode;
	set<GR::PatternGroup> groups = rule.PatternGroup_kind_children();
	if(groups.empty())
		effectorCode= translateEffector( rule, ruleDef, matchCond, boundObjects); 
	else
		effectorCode= translateGroupEffector( rule, ruleDef, matchCond, boundObjects);
	// If all outputs are inputs, and there is no sorting specified on outputs, then the 
	// inputs can be forwarded irrespective the no. of matches.
	// If this conditon is true, then a separate function is going to be generated for this purpose.
	// 'canForwardInputs' is a flag indicating if this function is generated.
	bool allOutputsAreInputs= PacketPassing::allOutputsAreInputs( *_passPacket.outputParams( rule));
	bool areOutputsToBeSorted= OutputSorter::areOutputsToBeSorted( *_passPacket.outputParams( rule));
	bool canForwardInputs= allOutputsAreInputs && !areOutputsToBeSorted;
	// If all outputs are inputs AND there are pattern objects with action DELETE among outputs,
	// then outputs are appended only if there was no match, to prevent passing null references,
	// and this is NOT done in 'effector()', but in 'processInputPackets()'.
	if ( canForwardInputs)
	{
		bool isDeletedOutput= PacketPassing::isAnyDeletedObjectAmongParams( *_passPacket.outputParams( rule));
		if ( isDeletedOutput) // rule with input forwarding with deleted objects among outputs
		{
			std::map< GR::ObjectWrapper, std::string> decoratedBoundObjects;
			std::transform( inputBoundObjects.begin(), inputBoundObjects.end(), std::inserter( decoratedBoundObjects, decoratedBoundObjects.begin()), decorateNamesWithCast());
// log			std::for_each( decoratedBoundObjects.begin(), decoratedBoundObjects.end(), printObject());
			translateOutputAppender( rule, ruleDef, &matchCond->getElseBranch(), decoratedBoundObjects);
		}
		else	// rule with input forwarding with NO deleted objects among outputs
		{
//			Function* outputAppenderFunc= translateOutputAppender( rule, ruleDef, effectorCode, boundObjects);
			translateForwardInputs( rule, ruleDef, funcOperator);
		}
	}
	else	// simple rule with NO input forwarding
	{
		if(groups.empty()) 
			translateOutputAppender( rule, ruleDef, effectorCode, boundObjects);
		if ( areOutputsToBeSorted) 
			Function* funcDef= translateOutputSorter( rule, ruleDef, funcOperator);
	}
}

void Translator::translateTest( const GRTest& test)
{
	Class* testDef= NULL;
	Function* funcOperator= NULL;
	translateRuleDef( test, &testDef, &funcOperator);
	translateOutputs( test, testDef, funcOperator);
}

void Translator::translateForTest( const GRForTest& forTest)
{
	Class* testDef= NULL;
	Function* funcOperator= NULL;
	translateRuleDef( forTest, &testDef, &funcOperator);
	translateOutputs( forTest, testDef, funcOperator);
	translateForCall( forTest, testDef, funcOperator);
}

void Translator::translateCase( const GRCase& testCase)
{
	if ( isIsolated( testCase))
	{
		// if case is isolated (not connected) do not generate class definition!
		if ( isReferred( testCase))
			printError( "Case"+ ( std::string)testCase.name()+ " is isolated, and also referred by a proxy.");
		else
			printWarning( "Case "+ ( std::string)testCase.name()+ " is isolated, no code is generated.");
		return;	
	}
	Class* caseDef= NULL;
	Function* funcOperator= NULL;
	translateRuleDef( testCase, &caseDef, &funcOperator);
	translateOutputs( testCase, caseDef, funcOperator);
	std::map< GR::ObjectWrapper, std::string> boundObjects;
	Function* pipFunc= 0;
	Loop* inputLoop= 0;
	translatePatterns( testCase, caseDef, funcOperator, boundObjects, pipFunc, inputLoop);
	// No effector function. Match loop is added to patternMatcher().
	Loop* matchLoop= new Loop;
	matchLoop->createMatchLoop();
	pipFunc->add( *matchLoop);
	// output sorting
	bool areOutputsToBeSorted= OutputSorter::areOutputsToBeSorted( *_passPacket.outputParams( testCase));
	if ( areOutputsToBeSorted) {
		translateOutputAppender( testCase, caseDef, matchLoop, boundObjects);
		Function* funcDef= translateOutputSorter( testCase, caseDef, funcOperator);
	} else {
		translateOutputAppender( testCase, caseDef, matchLoop, boundObjects);
	}
	// Set return value to bool for case.
	funcOperator->_returnValue= BOL;
	// Return value is true, if _matches_ is not empty.
	If* isOutput= new If;
	isOutput->_expression= FAL +EQEQ+ SEP+  "_matches.empty()";
	Expression* expTrue= new Expression;
	expTrue->_expression= RET+ SEP+ TRU;
	isOutput->add( *expTrue);
	funcOperator->add( *isOutput);
	//	
	Expression* expFalse= new Expression;
	expFalse->_expression= RET+ SEP+ FAL;
	funcOperator->add( *expFalse);
}

void Translator::translateCallInTest( const GRTestType& test, const GRCase& testCase)
{
	Function* testCallerFunc= getTestCaller( test);
	if ( NULL== testCallerFunc)
		throw CGEx("FATAL: CG internal error. Test class definition not found while translating case call.");
	Class* caseDefClass= getRuleDef( testCase);
	if ( NULL== caseDefClass)
		throw CGEx("FATAL: CG internal error. Case class definition not found while translating case call.");
	ParamDatas_t* caseInputs= _passPacket.inputParams( testCase);
	if ( NULL== caseInputs)
		throw CGEx("FATAL: CG internal error. Inputs of case not found while translating case call.");
	ParamDatas_t* caseOutputs= _passPacket.outputParams( testCase);
	if ( NULL== caseOutputs)
		throw CGEx("FATAL: CG internal error. Outputs of case not found while translating case call.");
	// Create new func.operator call on class belonging to 'testCase'.
	FuncOperator* callCaseOperator= new FuncOperator;
	callCaseOperator->_class= caseDefClass;
	callCaseOperator->_retType= BOL;
	callCaseOperator->_retVar= "isMatch"+ caseDefClass->_name;
	// Fill parameter std::list of func. operator with the source parameter names of 'testCase' inputs.
	//for( ParamDatas_t::const_iterator itInputs= caseInputs->begin(); itInputs!= caseInputs->end(); ++itInputs)
	_passPacket.matchParams( test, testCase, callCaseOperator->_params, true);
	// Create output definition for every output with the same name as in case parameter std::list.
	for( ParamDatas_t::const_iterator itOutputs= caseOutputs->begin(); itOutputs!= caseOutputs->end(); ++itOutputs)
	{
		Definition* outDef= new Definition;
		outDef->createPackets( itOutputs->_name);
		testCallerFunc->add( *outDef);
		callCaseOperator->_params.push_back( itOutputs->_name);
	}
	// Add func.operator call to either 'operator()'/ 'executeOne()' of 'test'.
	testCallerFunc->add( *callCaseOperator);
	// Case output marshaling to test output.
	addAppendOutputCalls( *caseOutputs, test, *testCallerFunc);
	// if 'testCase' has a cut, create the if-match-then-return condition.
	if ( true== testCase.cut())
	{
		If* isMatch= new If;
		isMatch->createIfExprThenReturn( callCaseOperator->_retVar);
		testCallerFunc->add( *isMatch);
	}
}

void Translator::translateCallInBlock( const GR::Sequence& seq)
{
	// Encompassing block is parent of 'seq'.
	GRBlockType currBlock= GRBlockType( GR::RewritingRule::Cast( seq.parent()));
	GR::RuleBase sourceRule= seq.srcRule_end();
	GR::RuleBase ruleToCall= seq.dstRule_end();
//	printLogMsg( "Translating seq: " + ( std::string)sourceRule.name()+ "->"+ ( std::string)ruleToCall.name()+ " in: "+ ( std::string)currBlock.name());
	// Get the function to which the call should be added. 
	Function* sourceCaller= ( sourceRule== currBlock) ? getBlockCaller( currBlock) : getCaller( sourceRule);
	if ( NULL== sourceCaller)
	{
			printError( "Sequence "+  ( std::string) sourceRule.name()+ "->"+ ( std::string) ruleToCall.name()+ " did not compile. Class definition of "+ ( std::string) sourceRule.name()+ " not exist.");
			return;
	}
	if ( ruleToCall == currBlock)
	{
		// Destination is output port. Generate the code which appends member output variables.
		ParamDatas_t* sources= ( sourceRule == currBlock) ? _passPacket.inputParams( getRewritingRule( sourceRule)) : _passPacket.outputParams( getRewritingRule( sourceRule));
		if ( Uml::IsDerivedFrom( sourceRule.type(), GR::RewritingRuleProxy::meta))
		{
//			if ( "AIsParent" == ( std::string) sourceRule.name()){
//				int stop_here= 0;
//			}
			ParamDatas_t proxySources;
			set< GR::ObjectWrapperProxy> owProxies= GR::RewritingRuleProxy::Cast( sourceRule).ObjectWrapperProxy_kind_children();
			for( set< GR::ObjectWrapperProxy>::const_iterator i= owProxies.begin(); i!= owProxies.end(); ++i)
			{
				std::set< GR::PassAlong> dests= i->dstObjectWrapper();
				if ( dests.empty())
					continue;
				const GR::ObjectWrapper& refersTo= i->refersTo();
				ParamDatas_t::iterator itFindRefersTo= find( sources->begin(), sources->end(), ParamData( refersTo, ""));
				if ( itFindRefersTo != sources->end())
				{
					ParamData foundParam= *itFindRefersTo;
					foundParam._param= *i;
					proxySources.push_back( foundParam);
				} else {
					// only for debug purposes
					std::string nn= i->name();
					std::string name= refersTo.name();
				}
			}
			addAppendOutputCalls( proxySources, currBlock, *sourceCaller);
		}
		else
			addAppendOutputCalls( *sources, currBlock, *sourceCaller);
	}
	else
	{
		// Straight calling. Generate the code which calls the rule-caller function.
		Function* callerFunc= getOrCreateCaller( ruleToCall, currBlock);
		if ( NULL == callerFunc)
		{
			printError( "Sequence "+  ( std::string) sourceRule.name()+ "->"+ ( std::string) ruleToCall.name()+ " did not compile. Class definition of "+ ( std::string) ruleToCall.name()+ " not exist.");
			return;
		}
		// 'callerFunc' must not be null. Either previously registered and fetched, or created.
		// Create function call to caller function.
		FunctionCall* callCaller= new FunctionCall;
		callCaller->_func= callerFunc;
		// This fills up the parameter std::list of function call: 'callCaller' with the parameter names of sources of callee inputs.
		_passPacket.matchParams( sourceRule, ruleToCall, callCaller->_params, (sourceRule== currBlock));
		// Build if-output-is-empty condition.
		If* checkIfOutputEmpty= new If;
		for( CallParams_t::const_iterator itParams= callCaller->_params.begin(); itParams!= callCaller->_params.end(); ++itParams)
			checkIfOutputEmpty->appendWithAnd( "!"+ lowerBegin( *itParams)+ ".empty()");
		// Add it to the caller function of the source rule.
		checkIfOutputEmpty->add( *callCaller);
		sourceCaller->add( *checkIfOutputEmpty);
	}
}

void Translator::translateRuleDef( const GR::RewritingRule& rule, Class** ruleDef, Function** funcOperator)
{
	*ruleDef= new Class;
	( *ruleDef)->_name= upperBegin( PE::createUniqueName( convertToCPlusPlusName( rule.name())));
	*funcOperator= new Function();
	( *funcOperator)->_name= "operator()";

	// print info
	Macro* printMacro= new Macro;
	printMacro->_defVar= "PRINT_INFO";
	Expression* logRuleName= new Expression( "std::cout << \""+ ( *ruleDef)->_name+ "\" << std::endl");
	printMacro->add( *logRuleName);
	(*funcOperator)->add( *printMacro);

	if((std::string)(rule.ruleType()) == "RULE")
	{
		char ruleid [33];
		itoa (rule.ruleId(),ruleid,10);
		std::string rttargs=(std::string)ruleid+", \""+(std::string)rule.name()+"\"";
		(*funcOperator)->add( *new Expression("RTTGenerator::Instance()->generateRule("+rttargs+")"));
	}

	_passPacket.createParamList( rule, **funcOperator);
	( *ruleDef)->addPublic( **funcOperator);
	_program.add( **ruleDef);
	registerRuleDef( rule, *ruleDef);
}

void Translator::translateOutputs( const GR::RewritingRule& rule, Class* ruleDef, Function* initFunc)
{	
	ParamDatas_t* outputs= _passPacket.outputParams( rule);
	for( ParamDatas_t::const_iterator itOutputs= outputs->begin(); itOutputs!= outputs->end(); ++itOutputs)
	{
		ClassDef* outputDef= new ClassDef;
		outputDef->createPacketsPtr( "_"+ lowerBegin( PE::createUniqueName( ( std::string) itOutputs->_param.name())));
		ruleDef->addPrivate( *outputDef);
		Expression* initOutput= new Expression;
		initOutput->_expression= outputDef->_variable+ EQ+ SEP+ REF+ lowerBegin( itOutputs->_name);
		initFunc->add( *initOutput);
		registerPort( itOutputs->_param, outputDef, *outputs);
	}
}

void Translator::translateForCall( const GR::RewritingRule& forRule, Class* ruleDef, Function* initFunc)
{
	// Here inputs are torn down into pieces, and passed one by one to a protected function: 'executeOne()'.
	Function* execute= new Function;
	execute->_name= "executeOne";
	ruleDef->addProtected( *execute);
	FunctionCall* executeCall= new FunctionCall;
	executeCall->_func= execute;
	Loop* inputLoop= new Loop;
	ParamDatas_t* inputs= _passPacket.inputParams( forRule);
	// Initialize input filter
	InputFilter inputFilter( forRule);
	Function* filterFuncDef;
	FunctionCall* filterFuncCall;
	inputFilter.createFunction( &filterFuncDef, &filterFuncCall, ruleDef, inputLoop);
	If* uniqueCond= 0;
	If* checkForUnique= 0;
	Loop* inputFilterLoop= 0;
	inputFilter.createInitCode( filterFuncDef, &inputFilterLoop, &uniqueCond, &checkForUnique);
	// do multiple things in parallel for each input..
	for( ParamDatas_t::iterator itInputs= inputs->begin(); itInputs!= inputs->end(); ++itInputs)
	{
		ParamData& currInput= *itInputs;
		inputFilter.createUniqueInputMember( currInput, *ruleDef);	// create member variable for current input
		const std::string& itName= _passPacket.addInputIterator( currInput, *inputLoop); // add input iterator into packet loop
		inputFilter.inputIdentityCheck( 
			currInput, PTR+ itName, filterFuncDef, filterFuncCall, inputFilterLoop, uniqueCond, checkForUnique); // add identiy checker for current input into packet loop
		// create list of packets initialized with current input
		Definition* oneDef= new Definition;
		oneDef->createPackets( PE::createUniqueName( "one"+ ( std::string) currInput._param.name()));
		oneDef->addArg( "1");
		oneDef->addArg( PTR+ itName);
		inputLoop->add( *oneDef);
		executeCall->_params.push_back( oneDef->_variable); // add argument to function call
		execute->insertInputParam( Function::createInputParam( lowerBegin( currInput._name), false)); // add argument to function def
	}
	inputFilter.createFinalCode( checkForUnique, filterFuncDef);
	inputLoop->add( *executeCall);
	initFunc->add( *inputLoop);
}

void Translator::translatePatterns( 
		const GRSimpleRule& rule, 
		Class* ruleDef, 
		PEComposite* patternFuncCaller, 
		std::map< GR::ObjectWrapper, std::string>& boundObjects,
		Function* &pipFuncDef,
		Loop* &inputLoop,
		std::map< GR::ObjectWrapper, std::string>* inputBoundObjects
		)
{
	// create processInputPackets() function
	pipFuncDef= new Function;
	pipFuncDef->_name= "processInputPackets";
	inputLoop= new Loop;
	// create patternMatcher() function
	Function* pmFuncDef= new Function;
	pmFuncDef->_returnValue= BOL;
	pmFuncDef->_name= "patternMatcher";
	FunctionCall* pmFuncCall= new FunctionCall;
	pmFuncCall->_func= pmFuncDef;
	pmFuncCall->_retType= BOL;
	pmFuncCall->_retVar= ISMATCH;
	
	// dummy loop for legitimate use of the keyword continue thoroghout in pattern matcher code.
	Loop* dummyLoop= new Loop;	
	dummyLoop->_from= "int i= 0";
	dummyLoop->_to= "i< 1";
	dummyLoop->_increment="++i";
	pmFuncDef->add( *dummyLoop);
	// init pattern matcher
	PatternMatcher patternMatcher( rule,  _backend, *dummyLoop, _cc);
	// create input iteration
	ParamDatas_t* inputs= _passPacket.inputParams( rule);
	// Initialize input filter
	InputFilter inputFilter( rule);
	Function* filterFuncDef;
	FunctionCall* filterFuncCall;
	inputFilter.createFunction( &filterFuncDef, &filterFuncCall, ruleDef, inputLoop);
	If* uniqueCond= 0;
	If* checkForUnique= 0;
	Loop* inputFilterLoop= 0;
	inputFilter.createInitCode( filterFuncDef, &inputFilterLoop, &uniqueCond, &checkForUnique);
	// do multiple things in parallel for each input..
	for( ParamDatas_t::iterator itInputs= inputs->begin(); itInputs!= inputs->end(); ++itInputs)
	{
		ParamData& currInput= *itInputs;
		inputFilter.createUniqueInputMember( currInput, *ruleDef); 	// create member variable for current input
		const std::string& itName= _passPacket.addInputIterator( currInput, *inputLoop);
		// generate code which passes currInput to the 'patternMatcher()'
		pmFuncCall->_params.push_back( PTR+ itName);
		std::string paramType= CST+ SEP+ "Udm"+ RES+ "Object"+ REF;
		std::string paramName= PE::createUniqueName( currInput._param.name());
		pmFuncDef->insertInputParam( FuncParam_t( std::make_pair( paramType, lowerBegin( paramName))));
		inputFilter.inputIdentityCheck( 
			currInput, PTR+ itName, filterFuncDef, filterFuncCall, inputFilterLoop, uniqueCond, checkForUnique); // add identiy checker for current input into packet loop
		// initial binding: bind inputs of rule
		const Definition* inputDef= patternMatcher.createInputMatch( currInput._param, paramName);
		// side-task: fill the parameter list of processInputPackets() function
		pipFuncDef->insertInputParam( Function::createInputParam( lowerBegin( itInputs->_name), false));
		//
		if ( inputBoundObjects)
			(*inputBoundObjects)[ currInput]= lowerBegin( itName);
	}
	inputFilter.createFinalCode( checkForUnique, filterFuncDef);
	// run pattern matcher on rule.
	patternMatcher.findMatch();
	// run match maker on rule.
	patternMatcher.makeMatch( *ruleDef, boundObjects);
	pmFuncDef->add( *new Expression( RET+ SEP+ "!"+ _MATCHES+ ".empty()" ));
	// add mathc condition and clear matches expression to input loop
	inputLoop->add( *pmFuncCall);
	set<GR::PatternGroup> groups = rule.PatternGroup_kind_children();
	if(!groups.empty())
		inputLoop->add(*new Expression("_groupMatches.clear()"));
	// add input loop to function definition.
	pipFuncDef->add( *inputLoop);
	// create caller to pipFuncDef() function
	FunctionCall* pipFuncCaller= new FunctionCall;
	pipFuncCaller->_func= pipFuncDef;
	pipFuncCaller->copyParamNames();
	patternFuncCaller->add( *pipFuncCaller);

// !!VIZA temporaly log size of matches
//	Macro* printMacro= new Macro;
//	printMacro->_defVar= "PRINT_INFO";
//	Expression* logMatches= new Expression( "std::cout << _matches.size() << std::endl");
//	printMacro->add( *logMatches);
//	patternFuncCaller->add( *printMacro);
// !!VIZA end temporary code

	ruleDef->addProtected( *pipFuncDef);
	ruleDef->addProtected( *pmFuncDef);
	
}

// Used for createing code which assigns unique inputs to the outputs
class OutputAssignerFunctor
{
public:
	OutputAssignerFunctor( const ParamDatas_t& inputs, Function& addTo) : _inputs( inputs), _addTo( addTo) {}
	void operator()( const ParamData& output)
	{
		const ClassDef* inputMemberDef= getPort( output, _inputs);
		if ( 0!= inputMemberDef)
		{
			_addTo.add( *new Expression( PTR+ output._memberDef->_variable+ EQ+ SEP+ inputMemberDef->_variable));
		}
	}
private:
	const ParamDatas_t& _inputs;
	Function& _addTo;
};

Function* Translator::translateForwardInputs( const GRRule& rule, Class* ruleDef, PEComposite* forwardFuncCaller)
{
	const ParamDatas_t* inputs= _passPacket.inputParams( rule);
	const ParamDatas_t* outputs= _passPacket.outputParams( rule);
	if ( ( 0 == outputs) || ( outputs->empty()) || ( 0 == inputs) || ( inputs->empty()))
		return 0;
	// Create forwardInputs() function.
	Function* fiFuncDef= new Function;
	fiFuncDef->_name= "forwardInputs";
	// Create code which assigns unique inputs to the outputs
	for_each( outputs->begin(), outputs->end(), OutputAssignerFunctor( *inputs, *fiFuncDef));
	// Create caller to forwardInputs() function.
	FunctionCall* fiFuncCaller= new FunctionCall;
	fiFuncCaller->_func= fiFuncDef;
	fiFuncCaller->copyParamNames();
	forwardFuncCaller->add( *fiFuncCaller);
	// Add function definition to block class definition.
	ruleDef->addProtected( *fiFuncDef);
	return fiFuncDef;
}

PEComposite* Translator::translateEffector( const GRRule& rule, Class* ruleDef, PEComposite* effectorFuncCaller, std::map< GR::ObjectWrapper, std::string>& boundObjects)
{
	// create effector() function
	Function* effFuncDef= new Function;
	effFuncDef->_name= "effector";

	// Run effector on rule.
	Effector effector( rule, _backend, boundObjects, _cc);
	effector.executeRightSide( *effFuncDef);

	// create caller to effector() function
	FunctionCall* effFuncCaller= new FunctionCall;
	effFuncCaller->_func= effFuncDef;
	effFuncCaller->copyParamNames();
	effectorFuncCaller->add( *effFuncCaller);
	ruleDef->addProtected( *effFuncDef);
	//
	return effector.getCode();
}

PEComposite* Translator::translateGroupEffector( const GRRule& rule, Class* ruleDef, PEComposite* effectorFuncCaller, std::map< GR::ObjectWrapper, std::string>& boundObjects)
{
	//Add a new private member
	ClassDef* groupDef= new ClassDef;
	groupDef->_type= "std::list< std::list<"+ SEP+ MATCH+ "> >";
	groupDef->_variable= "_groupMatches";
	ruleDef->addPrivate( *groupDef);

	// create effector() function
	Function* effFuncDef= new Function;
	effFuncDef->_name= "effector";
	GroupEffector effector(rule, _backend, boundObjects, _cc, *_passPacket.outputParams(rule));
	effector.makeGroupRule(*ruleDef);
	effector.executeRightSide( *effFuncDef);
	effector.executeGroup( *ruleDef);

	// create caller to effector() function
	FunctionCall* effFuncCaller= new FunctionCall;
	effFuncCaller->_func= effFuncDef;
	effFuncCaller->copyParamNames();
	effectorFuncCaller->add( *effFuncCaller);

	ruleDef->addProtected( *effFuncDef);
	//
	return effector.getCode();
}

Function* Translator::translateOutputAppender( const GRSimpleRule& rule, Class* ruleDef, PEComposite* outputAppenderFuncCaller, std::map< GR::ObjectWrapper, std::string>& boundObjects)
{
	// create outputAppender() function
	Function* oaFuncDef= new Function;
	oaFuncDef->_name= "outputAppender";
	// create caller to outputAppender() function
	FunctionCall* oaFuncCaller= new FunctionCall;
	oaFuncCaller->_func= oaFuncDef;
	// Run output filter. Function parameters are filled by OutputAppender::execute().
	OutputAppender outputAppender( rule, boundObjects);
	outputAppender.addOutputs( *_passPacket.outputParams( rule), *oaFuncDef, *oaFuncCaller);
	outputAppenderFuncCaller->add( *oaFuncCaller);
	ruleDef->addProtected( *oaFuncDef);
	return oaFuncDef;
}


Function* Translator::translateOutputSorter( const GRSimpleRule& rule, Class* ruleDef, PEComposite* outputSorterFuncCaller)
{
	bool isDistinguished= rule.distinguished();
	// create outputSorter() function
	Function* osFuncDef= new Function;
	osFuncDef->_name= isDistinguished ? "distinguishedSortOutputs" : "sortOutputs";
	// create caller to sortOutputs() function
	FunctionCall* osFuncCaller= new FunctionCall;
	osFuncCaller->_func= osFuncDef;
	// Run output filter. Function parameters are filled by OutputAppender::execute().
	try {
		if ( isDistinguished) {
			DistinguishedOutputSorter outputSorter( rule, *_passPacket.outputParams( rule));
			outputSorter.generate( *osFuncDef);
		} else {
			OutputSorter outputSorter( rule, *_passPacket.outputParams( rule));
			outputSorter.generate( *osFuncDef);
		}
	} catch ( CGEx& ex) {
		printError( ex.what());
	}
	outputSorterFuncCaller->add( *osFuncCaller);
	ruleDef->addProtected( *osFuncDef);
	return osFuncDef;
}

void Translator::addAppendOutputCalls( const ParamDatas_t& sources, const GR::RewritingRule& outputParent, Function& caller) const
{
	std::list< std::string> outputnames;
	// Iterate sources.
	for( ParamDatas_t::const_iterator itSources= sources.begin(); itSources!= sources.end(); ++itSources)
	{
		const ParamData& currSource= *itSources;
		std::set< GR::PassAlong> dstPAs= currSource._param.dstObjectWrapper();
		if ( dstPAs.empty()) {
			std::string sourceName= currSource._param.name();
			if ( Uml::IsDerivedFrom( currSource._param.type(), GR::ObjectWrapperProxy::meta)) {
				GR::ObjectWrapperProxy proxy= GR::ObjectWrapperProxy::Cast( currSource._param);
				long id= proxy.uniqueId();
				std::string refersToName= GR::ObjectWrapper::Cast( proxy.refersTo()).name();
			}
		}
		// Iterate destinations of 'currSource'.
		for( std::set< GR::PassAlong>::const_iterator itDstPAs= dstPAs.begin(); itDstPAs!= dstPAs.end(); ++itDstPAs)
		{
			GR::ObjectWrapper dstOW= itDstPAs->dstObjectWrapper_end();
			// parent of destination of the pass along can be proxy also!
			std::string parentType= Udm::Object::Cast( dstOW.parent()).type().name();
			GR::RewritingRule parent;
			if ( parentType== "RewritingRuleProxy")
			{
				parent= GR::RewritingRuleProxy::Cast( dstOW.parent()).refersTo();
//				printLogMsg("!!!PROXY!!! to: "+ ( std::string)parent.name() + "::"+ ( std::string)dstOW.name());
//				printLogMsg("from: "+ ( std::string)GR::RewritingRule::Cast( currSource._param.parent()).name() + "::"+ ( std::string)currSource._param.name());
			}
			else
			{
				parent= GR::RewritingRule::Cast( dstOW.parent());
			}
			if ( parent == outputParent)
			{
				if ( true== Uml::IsDerivedFrom( dstOW.type(), GR::OutputPort::meta))
				{
					// dstOw is an 'GR::OutputPort' type child of 'outputParent'. Create the copy expression.
					GR::OutputPort dstOutputPort= GR::OutputPort::Cast( dstOW);
					const ParamDatas_t* outputs= _passPacket.outputParams( outputParent);
					const ClassDef* outputMemberVar= getPort( dstOutputPort, *outputs);
					if ( NULL!= outputMemberVar)
					{
						Expression* copyOutput= new Expression;
						copyOutput->_expression= lowerBegin( outputMemberVar->_variable)+ ACC+ "insert("+ SEP+ lowerBegin( outputMemberVar->_variable)+ ACC+ END+ ","+ SEP+ lowerBegin( currSource._name)+ "."+ BEG+ ","+ SEP+ lowerBegin( currSource._name)+ "."+ END+ ")";
						caller.add( *copyOutput);
						outputnames.push_back( outputMemberVar->_variable);
					}
				}
			}
		}
	}
	/*
	if ( !outputnames.empty()) {
		If* checkSizes= new If;
		for( std::list< std::string>::const_iterator it= outputnames.begin(); it!= outputnames.end();) {
			std::list< std::string>::const_iterator itNext= ++it;
			if ( outputnames.end() != itNext) {
				checkSizes->appendWithOr( *it+ "->size()" + SEP+ NEQ+ SEP+ *itNext+ "->size()" );
			}
		}
		checkSizes->add( *new Expression("throw udm_exception(\"Container size mismatch error.\")"));
		caller.add( *checkSizes);
	}
	*/
}

// Return the func.operator()/ executeOne() function, which
// are the first elements of the block class definition public/ protected blocks.
// Thus, correct operation implicitly relies upon the fact, that these functions are added 
// first time in 'translateBlock()'/ 'translateForBlock()'.
Function* Translator::getBlockCaller( const GRBlockType& block) const
{
	Class* ruleDefClass= getRuleDef( block);
	if ( NULL== ruleDefClass)
		return NULL;
	if ( true== block.forBlock())
		return ( Function*) ruleDefClass->getProtected()->front();
	else
		return ( Function*) ruleDefClass->getPublic()->front();
}

// Counterpart of 'getBlockCaller()' for tests.
// Same implementation currently, but might change in the future.
Function* Translator::getTestCaller( const GRTestType& test) const
{
	Class* ruleDefClass= getRuleDef( test);
	if ( NULL== ruleDefClass)
		return NULL;
	if ( true== test.forBlock())
		return ( Function*) ruleDefClass->getProtected()->front();
	else
		return ( Function*) ruleDefClass->getPublic()->front();
}

Function* Translator::getOrCreateCaller( const GR::RuleBase& toCall, const GRBlockType& block)
{
	// Get existing if already created.
	Function* callerFunc= getCaller( toCall);
	if ( NULL!= callerFunc)
		return callerFunc;	// exists
	// Not existing. -> Create.
	callerFunc= new Function;
	callerFunc->_name= PE::createUniqueName( std::string( "call")+ convertToCPlusPlusName( ( std::string)toCall.name()));
	// Create func.op.-call on rule definition class.
	GR::RewritingRule toCallRule( getRewritingRule( toCall));
	Class* ruleClass= getRuleDef( toCallRule);
	if ( NULL== ruleClass)
		return NULL;//throw CGEx("FATAL: CG internal error. Rule definition class is missing while generating caller: "+ ( std::string)toCallRule.name());
	FuncOperator* callOperator= new FuncOperator;
	callOperator->_class= ruleClass;
	// Handle inputs...
	ParamDatas_t* inputs= _passPacket.inputParams( toCallRule);
	if ( NULL== inputs)
		throw CGEx("FATAL: CG internal error. No inputs found for rule: "+ ( std::string)toCallRule.name()+ " while generating caller.");
	for( ParamDatas_t::const_iterator itInputs= inputs->begin(); itInputs!= inputs->end(); ++itInputs)
	{
		// append parameter std::list of both caller function & func. operator with the same name as in rule parameter std::list. 
		callerFunc->insertInputParam( Function::createInputParam( lowerBegin( itInputs->_name), false));
		callOperator->_params.push_back( itInputs->_name);
	}
	// Handle outputs...
	ParamDatas_t* outputs= _passPacket.outputParams( toCallRule);
	if ( NULL!= outputs)
	{
		// There is output of 'toCallRule'.
		for( ParamDatas_t::const_iterator itOutputs= outputs->begin(); itOutputs!= outputs->end(); ++itOutputs)
		{
			// Create output definition for every output with the same name as in rule parameter std::list.
			Definition* outDef= new Definition;
			outDef->createPackets( itOutputs->_name);
			callerFunc->add( *outDef);
			callOperator->_params.push_back( itOutputs->_name);
		}
	}
	// Add func.op. to caller function.
	callerFunc->add( *callOperator);
	// Add caller definition to composite.
	Class* blockClass= getRuleDef( block);
	if ( NULL== blockClass)
		throw CGEx("FATAL: CG internal error. Definition class for parent block: "+ ( std::string)block.name()+ " is missing while generating caller.");
	blockClass->addProtected( *callerFunc);
	// register the caller.
	registerCaller( toCall, callerFunc);
	return callerFunc;
}

void Translator::registerRuleDef( const GR::RewritingRule& rule, const Class* ruleDef)
{
	_ruleDefMap[ rule]= ruleDef;
}

Class* Translator::getRuleDef( const GR::RewritingRule& rule) const
{
	RuleDefMap_t::const_iterator itFind= _ruleDefMap.find( rule);
	return ( itFind== _ruleDefMap.end()) ? NULL : ( Class*) itFind->second;
}

void Translator::registerCaller( const GR::RuleBase& ruleBase, const Function* caller)
{
	_callerMap[ ruleBase]= caller;
}

Function* Translator::getCaller( const GR::RuleBase& ruleBase) const
{
	CallerMap_t::const_iterator itFind= _callerMap.find( ruleBase);
	return ( itFind== _callerMap.end()) ? NULL : ( Function*) itFind->second;
}

bool Translator::isIsolated( const GR::RuleBase& rule) const
{
	std::string nm= rule.name();
	std::set< GR::Sequence> sources= rule.srcRule();
	return sources.empty();
}

bool Translator::isReferred( const GR::RewritingRule& rule) const
{
	std::string nm= rule.name();
	std::set< GR::RewritingRuleProxy> referedByProxies= rule.referedBy();
	if ( referedByProxies.empty())
		return false;
	for( std::set< GR::RewritingRuleProxy>::const_iterator itProxies= referedByProxies.begin(); itProxies!= referedByProxies.end(); ++itProxies)
	{
		const GR::RewritingRuleProxy& currProxy= *itProxies;
		// if any proxy is non-isolated...
		if ( false == isIsolated( currProxy))
			return true;
	}
	return false;
}
