#define NOMINMAX

#include <limits>
#include "GRDBackEnd.h"
#include "../GRDebugInterface/GRDebugInterface.h"
#include <sstream>

namespace GRDI
{

typedef std::map<IFunctionPtr,IFunctionPtrSet> DepMap;
typedef std::map<IFunctionPtr, int> PosMap;

//=====================================
void GRDBackEnd::orderFunctionCalls(FunctionDescr& funcDescr)
{
  IFunctionPtrSet calls(funcDescr.m_calls.begin(),funcDescr.m_calls.end());
  DepMap depMap;
  PosMap posMap;
  FunctionTable newCallOrder;
  IFunctionPtrSet::const_iterator itT =calls.begin();
 	for (; itT != calls.end(); ++itT)
 	{
    const IFunctionPtr& calledName = *itT;
    FunctionDescrMap::const_iterator itF = m_functions.find(calledName); 
    if (itF == m_functions.end())
    {
      std::stringstream str;
      str << *(calledName);
      std::string a(str.str());
      throw std::invalid_argument("invalid function in ordering");
    }
    const FunctionTable& dep = itF->second.m_dependsOn;
    posMap[ calledName ] = itF->second.m_xPos;
    if (!dep.empty())
      depMap[calledName].insert(dep.begin(), dep.end()); 
  }

  while(!calls.empty())
  {
    IFunctionPtrSet::iterator itT = calls.begin();
   	for (; itT != calls.end(); ++itT)
 	  {
      DepMap::const_iterator itF = depMap.find(*itT); 
      if (itF == depMap.end())
        break;
    }
    
    if (itT == calls.end()) // break the loop
    {
		int min = std::numeric_limits<int>::max();
      PosMap::const_iterator minitP;
      PosMap::const_iterator itP = posMap.begin();
      for(; itP != posMap.end(); ++itP)
        if (min > itP->second)
        {
          min = itP->second;
          minitP = itP;
        }
      itT = calls.find(minitP->first);

      if (itT == calls.end())
        throw std::invalid_argument("Can't break the loop!");
    }
    const IFunctionPtr& calledName = *itT;

    newCallOrder.push_back(calledName);
    DepMap::iterator itE = depMap.begin(); 
    for(;itE != depMap.end(); ++itE)
    {
      IFunctionPtrSet& set = itE->second;
      set.erase(calledName);
    }

    posMap.erase(calledName);
    calls.erase(itT);

    itT = calls.begin();
   	for (; itT != calls.end(); ++itT)
 	  {
      const IFunctionPtr& calledName = *itT;
      DepMap::iterator itF = depMap.find(calledName); 
      if (itF != depMap.end() && 
          itF->second.empty())
        depMap.erase(itF);
    }
 	}

 	funcDescr.m_calls = newCallOrder;
}

//x//=====================================
//xvoid GRDBackEnd::orderFunctionCalls(FunctionDescr& funcDescr)
//x{
//x	GraphType	graph;
//x
//x	VertexNameMap vertexNameMap = get(vertex_name, graph);
//x	NameVertexMap nameVertexMap;
//x
//x  const FunctionTable& calls = funcDescr.m_calls;
//x	//add vertecies
//x//x  std::cout <<  "+++++++++++++++++" << std::endl;
//x//x  std::cout <<  funcDescr.m_func->print() << std::endl;
//x	for (FunctionTable::size_type i = 0; i < calls.size(); ++i)
//x	{
//x		const IFunctionPtr& name = calls[i];
//x	 	const VertexType& v = add_vertex(graph);
//x		vertexNameMap[v] = name;
//x		nameVertexMap[name] = v;
//x//    std::cout <<  name->print() << std::endl;
//x	}
//x 	//add egdes
//x 	for (FunctionTable::size_type j = 0; j < calls.size(); ++j)
//x 	{
//x 		const IFunctionPtr& calledName = calls[j];
//x
//x    NameVertexMap::const_iterator itCalled = nameVertexMap.find(calledName);
//x    if (itCalled == nameVertexMap.end())
//x      throw std::invalid_argument("invalid vertex name");
//x
//x  	const VertexType& calledV = itCalled->second;
//x 
//x    const FunctionDescr& calledFuncDescr = m_functions.find(calledName)->second; 
//x
//x 		for (FunctionTable::size_type i = 0; i < calledFuncDescr.m_dependsOn.size(); ++i)
//x 		{
//x 			const IFunctionPtr& callerName = calledFuncDescr.m_dependsOn[i];
//x
//x      NameVertexMap::const_iterator itCaller = nameVertexMap.find(callerName);
//x
//x      if (itCaller  != nameVertexMap.end())  
//x      {
//x   			const VertexType& callerV = itCaller->second;
//x
//x        if (callerV != calledV)
//x        {
//x 	  		  remove_edge(callerV, calledV, graph); 
//x 		  	  add_edge(callerV, calledV, graph);
//x        }
//x      }
//x 		}
//x 	}
//x 
//x 	// topologocal sort stuff
//x  boost::property_map<GraphType, vertex_index_t>::type vertexIndex = get(vertex_index, graph);
//x  typedef std::list<VertexType> Order;
//x  Order order;
//x
//x 	try
//x 	{
//x
//x  		boost::topological_sort(graph, std::front_inserter(order));
//x
//x 	}
//x 	catch(not_a_dag a)
//x 	{
//x 	 std::stringstream str;
//x  for (Order::iterator io = order.begin();io != order.end(); ++io)
//x 	{
//x 
//x 		int idx  = vertexIndex[*io];
//x    str <<  funcDescr.m_calls[idx]->getIFunctionImpl()->getFullPath() << std::endl;
//x
//x 	}
//x 
//x
//x 		throw std::invalid_argument(str.str());
//x 	}
//x
//x 	
//x 
//x 	FunctionTable newCallOrder;
//x  for (Order::iterator io = order.begin();io != order.end(); ++io)
//x 	{
//x 
//x 		int idx  = vertexIndex[*io];
//x 		newCallOrder.push_back(funcDescr.m_calls[idx]);
//x 	}
//x 
//x 	funcDescr.m_calls = newCallOrder;
//x}

}//namespace GRDI