#pragma warning(disable : 4786) 
#include "../GRDebugInterface/GRDebugInterface.h"
#include "GRDBackEnd.h"
#include <iostream>

#include <sstream>
#include <fstream>

namespace GRDI
{

// =========== DebugInterface
//=========================================
//=========================================
Error GRDBackEnd::getStatus(Status& status)
{
  Lock lk(m_mutexCtrl);

  if (m_suspended)
    status = STATUS_SUSPENDED;
  else
    status = STATUS_RUNNING;
  return GRDI_ERROR_NONE;                 
}
//=========================================
Error GRDBackEnd::getContextStacktSize(int& count)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  count = m_context.size();
  return GRDI_ERROR_NONE;                 
}
//=========================================
Error GRDBackEnd::getCurrentContextID(ContextID& contextID)
{

  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  if (m_context.size() <= 0)
    return GRDI_ERROR_NO_MORE_CONTEXTES;

  contextID = m_context.back().m_id;

  return GRDI_ERROR_NONE;                 
}
//=========================================
Error GRDBackEnd::getContextDescr(const ContextID& contextID, ContextDescr& contextDescr)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  if ((contextID < 0) || 
      (contextID >= m_context.size())
      )
    return GRDI_ERROR_INVALID_CONTEXT;
 
 contextDescr = m_context[contextID];
 return GRDI_ERROR_NONE;                 
}
//=========================================
Error GRDBackEnd::getCallerContext(const  ContextID& called,
        								                  ContextID& caller)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  if ((called >= m_context.size()) || (called <0 ))
    return GRDI_ERROR_INVALID_CONTEXT;

  caller = m_context[called].m_callerCtx;

  if (caller == invalidContextID)
    return  GRDI_ERROR_NO_MORE_CONTEXTES;

  return GRDI_ERROR_NONE;                 
}

//=========================================
Error GRDBackEnd::getCurrentLocation( const  ContextID& contextID,
 									                   LocationID& loc)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  if ((contextID >= m_context.size()) ||
       (contextID < 0)) 
    return GRDI_ERROR_INVALID_CONTEXT;

  loc = m_context[contextID].m_currLocationID;
  
  return GRDI_ERROR_NONE;                 
}
//=========================================
Error GRDBackEnd::getLastException( std::string& str)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  str = m_exceptionDescr;  
  return GRDI_ERROR_NONE;                 
}
//=========================================
Error GRDBackEnd::notifyContextPop(const ContextID& context)
{
  boost::recursive_mutex::scoped_lock lkWatch(m_mutexWatch);
  if ((context >= m_context.size()) ||
      (context < 0)) 
  return GRDI_ERROR_INVALID_CONTEXT;

  m_popEventCtxID = context;
  return GRDI_ERROR_NONE;
}
//=========================================
Error GRDBackEnd::suspend()
{
  Lock lk(m_mutexSuspend);
  m_suspended  = true;
  return GRDI_ERROR_NONE;
}

//=========================================
Error GRDBackEnd::resume()
{
  Lock lk(m_mutexSuspend);
  m_suspended  = false;
  m_resume.notify_all();

  return GRDI_ERROR_NONE;
}
//=========================================
Error GRDBackEnd::execute()
{
  Lock lk(m_mutexCtrl);
  m_debugInfo = false;
  m_synchStatus |= RUN;
  return GRDI_ERROR_NONE;
}

//=========================================
Error GRDBackEnd::run()
{
  Lock lk(m_mutexCtrl);
  m_debugInfo = true;
  m_synchStatus |= RUN;
  return GRDI_ERROR_NONE;
}
//=========================================

Error  GRDBackEnd::pause()
{
  Lock lk(m_mutexCtrl);
  m_tempBreakpoint = true;
  return GRDI_ERROR_NONE;
}
//=========================================
Error  GRDBackEnd::stop()
{
  Lock lk(m_mutexCtrl);
  m_synchStatus |= STOP;
  return GRDI_ERROR_NONE;
}


//=========================================  
Error  GRDBackEnd::quit()
{
  Lock lk(m_mutexCtrl);
  m_synchStatus = QUIT;
  return GRDI_ERROR_NONE;
}
//=========================================  
Error  GRDBackEnd::save()
{
  Lock lk(m_mutexCtrl);
  m_synchStatus |= SAVE;
  return GRDI_ERROR_NONE;
}

//=========================================  
Error  GRDBackEnd::load(const std::string& programFileName)
{
  Lock lk(m_mutexCtrl);
  m_programFileName = programFileName;
  if (m_synchStatus & RUN)
      m_synchStatus |= STOP;
  m_synchStatus |= LOAD;
 
  return GRDI_ERROR_NONE;
}
//=========================================  
Error  GRDBackEnd::close()
{
  Lock lk(m_mutexCtrl);
    
  m_synchStatus |= CLOSE;

  return GRDI_ERROR_NONE;
}

//=========================================
Error GRDBackEnd::setBreakpoint(const LocationID& locID)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);

  LocIDLocMap::const_iterator it =  m_id2loc.find(locID);
 
  if (it == m_id2loc.end())
    return GRDI_ERROR_INVALID_LOCATION; 

  m_breakpoints.insert(locID);

  return GRDI_ERROR_NONE;
}
//=========================================
Error GRDBackEnd::clearBreakpoint(const LocationID& loc)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  m_breakpoints.erase(loc);
  return GRDI_ERROR_NONE;
}
//=========================================
Error GRDBackEnd::getBreakPoints(LocationIDSet& breakpoints)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  breakpoints = m_breakpoints;
  return GRDI_ERROR_NONE;
}
//=========================================
Error GRDBackEnd::clearAllBreakpoints()
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  m_breakpoints.clear();
  return GRDI_ERROR_NONE;
}
//=========================================
//xError GRDBackEnd::getStartFunctio(IFunctionPtr& startFunc)
//x{
//x  Lock lk(m_mutexWatch);
//x  startFunc = m_startFunc;
//x  return GRDI_ERROR_NONE;
//x}
//=========================================
Error GRDBackEnd::getRootLocations(LocationIDVec& rootLocs)
{
  boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  rootLocs = m_rootLocs;
  return GRDI_ERROR_NONE;
}
//=========================================

 Error GRDBackEnd::getFunctionDescr(const IFunctionPtr& func,
                             FunctionDescr& descr)
 {
    boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
  
    FunctionDescrMap::const_iterator it = m_functions.find(func);
  
    if (it == m_functions.end())
      return GRDI_ERROR_INVALID_FUNCTION;
  
    descr = it->second;
  
    return GRDI_ERROR_NONE;
 }
 //=========================================
 Error GRDBackEnd::getLocationFromID(const LocationID& locationID,
                                     Location& location)
 {
   boost::recursive_mutex::scoped_lock lk(m_mutexWatch);
   LocIDLocMap::const_iterator it = m_id2loc.find(locationID);
 
   if (it == m_id2loc.end())
     return GRDI_ERROR_INVALID_LOCATION;
 
   location = it->second;
 
   return GRDI_ERROR_NONE;
 }

//=========================================
Error GRDBackEnd::setEventHook(EventHook* hook)
{
 boost::recursive_mutex::scoped_lock lkWatch(m_mutexWatch);
 if (hook == NULL)
   throw std::invalid_argument("EventHook null pointer");

 m_eventHook = hook;

 return GRDI_ERROR_NONE;
}

};//namespace GRDI