#if !defined(IGRDBackEnd_h)
#define IGRDBackEnd_h
#include "../GRDebugInterface/GRDebugInterface.h"

namespace GRDI
{

//===============================
class IGRDBackEnd
{
  public:
    // the values are power of 2 
    typedef enum {STOP = 1, QUIT = 2, LOAD = 4, EXCEPTION = 8,
                  SAVE = 16, CLOSE = 32, RUN = 64} SynchStatus;
    typedef unsigned int SynchType;

  public:

  virtual ~IGRDBackEnd() {}
  virtual void stepIn(const FunctionCallDescr& funcValDescr) = 0;
  virtual void step(const FunctionCallDescr& funcValDescr) = 0;
  virtual void stepOut(const FunctionCallDescr& funcValDescr) = 0;


  virtual DebugInterface* getDebugInterface() = 0;

  virtual void setStartFunction(const IFunctionPtr&) = 0;
  virtual void addRootFunction(const IFunctionPtr&) = 0;
  
  virtual void registerFunctionCall(const FunctionDescr& funcDescr) = 0;
  virtual void clearSource() = 0;

  virtual bool synchronize() = 0;
  virtual SynchType  getSynchStatus() = 0;
  virtual bool  debugInfo() = 0;
  virtual const std::string& getProgramFileName() = 0;
  virtual const std::string& getWD() = 0;
  virtual void up() = 0;
  virtual void down() = 0;
  virtual void saved() = 0;
  virtual void loaded() = 0;
  virtual void closed() = 0;
  virtual void stopped() = 0;
  virtual void exception(const std::string& str) = 0;
  virtual void start()  = 0;
  virtual void finish()   = 0;
  virtual void reportProgress(double prg) = 0;
};
}; //namespace GRDI
#endif