#if !defined(GRDebugTypes_h)
#define GRDebugTypes_h

#include "GRDebugVariable.h"


#include <vector>
#include <set>
#include <map>


namespace GRDI
{

  
/// Status of the execution  
typedef enum {
    STATUS_UNKNOWN, 
    STATUS_ZOMBIE, //Completed execution
    STATUS_RUNNING,
    STATUS_SUSPENDED,
    STATUS_BREAK }  Status;
  

///
///    Typdefs



typedef int Error;


//===============================
typedef unsigned long LocationID;

extern  LocationID invalidLocationID; 

enum  LocationType {FunctionBegin, FunctionEnd, FunctionCall};

typedef std::vector<LocationID> LocationIDVec; 
typedef std::set<LocationID> LocationIDSet;

typedef std::pair<FunctionID, LocationType> FunctionIDLocationTypePair;
typedef unsigned long uniqueId_type;
class Location : public FunctionIDLocationTypePair
{
 public :
     
     LocationID m_locID;
     LocationID m_parID;
     LocationID m_refID;
     IFunctionPtr     m_caller;
     IFunctionPtr     m_called;
		uniqueId_type		m_ID;		///Added by Akash ///Change for "Rule viewing" in GME

 public:

  const FunctionID&			funcID() const;
  const IFunctionPtr&      caller() const;
//const CallPosition&      pos() const;
  const IFunctionPtr&      called() const;
  const uniqueId_type		id() const;
  const LocationType&      type() const;
 
 public:
  Location();
  Location(const IFunctionPtr& caller, 
           const IFunctionPtr& called, 
           const LocationType& type);
 };

 typedef std::map <LocationID,Location> LocIDLocMap;
 typedef std::map <FunctionIDLocationTypePair,LocationID> LocLocIDMap;




//===============================
class  FunctionDescr
{
public:
  IFunctionPtr        m_func;
  FunctionTable       m_calls;
  FunctionTable       m_dependsOn;
  VariableValueMap    m_varTable;
  LocationIDVec       m_callLocs;
  int                 m_xPos;
  IFunctionPtr        m_ref;
  uniqueId_type		 m_id;	///Added by Akash ///Change for "Rule viewing" in GME
} ;

//===============================
typedef struct 
{
  IFunctionPtr            m_caller;
  IFunctionPtr            m_func;
  VariableValueMap        m_varValMap;
} FunctionCallDescr;

typedef std::map<IFunctionPtr, FunctionDescr> FunctionDescrMap;

//typedef const FunctionDescrMap* FunctionDescrMapPtr;


//===============================

typedef unsigned long ContextID;

extern  ContextID invalidContextID; 

class ContextDescr: public FunctionCallDescr
{
public:
  ContextDescr();
  ContextDescr(
               const ContextID& id,            
               const FunctionCallDescr& fcd, 
               const ContextID& caller,
               const LocationID& currLocationID);
  ContextID            m_id;
  ContextID            m_callerCtx;
  LocationID           m_currLocationID;
};

typedef std::vector<ContextDescr> Context;

};


#endif //GRDebugTypes_h