#if !defined(GRDebugVariable_h)
#define GRDebugVariable_h

#include <boost/shared_ptr.hpp>
#include <vector>
#include <iostream>
#include <map>
#include <set>

namespace GRDI
{
class IFunctionImpl;

class IFunction;
typedef boost::shared_ptr<IFunction> IFunctionPtr;

extern  IFunctionPtr nullFunction; 

typedef std::vector <IFunctionPtr> FunctionTable;
typedef std::set<IFunctionPtr> IFunctionPtrSet;
typedef unsigned long uniqueId_type;

typedef unsigned int FunctionID;
extern  FunctionID nullFunctionID; 

class IFunction
{
  public:
  virtual ~IFunction(){}

  virtual IFunctionImpl* getIFunctionImpl() = 0;
  virtual const FunctionID getID() const = 0;
  virtual uniqueId_type getRuleId() const = 0;
  virtual void print(std::ostream& stream, int indent = 0) const = 0;
};




//===========================================
class IVariableImpl;

class IVariable  
{
  public:
  virtual ~IVariable(){}
  virtual IVariableImpl* getIVariableImpl() = 0;
  virtual void print(std::ostream& stream, int indent = 0) const = 0;
  virtual std::string name() const = 0;
};
//===========================================

typedef boost::shared_ptr<IVariable>  IVariablePtr;

class IVariableValueImpl;

class IVariableValue  
{
   public:
   virtual ~IVariableValue() {}
   virtual IVariableValueImpl* getIVariableValueImpl() = 0;
   virtual void print(std::ostream& stream, int indent = 0) const = 0;
};
//===========================================
typedef boost::shared_ptr<IVariableValue>  IVariableValuePtr;


typedef std::map<IVariablePtr, IVariableValuePtr> VariableValueMap;


};

std::ostream& operator<<(std::ostream& stream, const GRDI::IFunction&);
std::ostream& operator<<(std::ostream& stream, const GRDI::IVariable&);
std::ostream& operator<<(std::ostream& stream, const GRDI::IVariableValue&);

#endif //GRDebugVariable_h