#if !defined(GRDebugger_h)
#define GRDebugger_h
#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>

#include "../GRLauncher/GRDebuggerAPI.h"
#include "../GRDFrontEnd/IGRDFrontEnd.h"
#include "../GRDebugInterface/GRDebugInterface.h"
#include "IGRDebuggerCtrl.h"
#include "IGRDebuggerView.h"

namespace GRDI
{



typedef boost::mutex::scoped_lock Lock;
class DebugInterface;
class IGRDebuggerView;

class GRDebugger :  public GRDebuggerAPI,
                    public  GRDebugNotification,
                    public IGRDebuggerCtrl
                  
{
  protected:

    IGRDFrontEnd* m_frontEnd;

    IGRDebuggerView*  m_view;

    DebugInterface*   m_debugInterface; 

    boost::mutex      m_mutexNotification;
    boost::mutex      m_mutexCtrl;
    
    bool              m_quit; 

    std::string       m_notifyMsg;

    bool              m_eventHandlig; 
    bool              m_isUp;
    bool              m_saved;
    bool              m_loaded;
    bool              m_exception;
    bool              m_logging;
    boost::condition    m_events;

//    FunctionDescrMap  m_funcDescrMap;

    std::string       m_helpTxt;
    std::string       m_dataFName;
    double            m_progress;

  protected:


    void readHelp(std::istream& out);

    void listCmd();
    void loadCmd(const std::string& pn);
    void closeCmd();
    void saveCmd();
    void clearCmd(const std::string& cmd);
    void whereCmd();
    void localsCmd();
    void evalCommand(const std::string& cmd);
    void logCmd(const std::string& cmd);
    void progressCmd();
    void setCmd(const std::string& cmd);

    void buildFunctionSourceRec(const LocationID& locID,
                                std::stringstream& str);

    void buildFunctionSource(const LocationID& locID,
                             std::stringstream& str);



    void buildSource();
    void markFuncLocation(std::stringstream& str, 
                          const  GRDI::LocationIDSet& loc,
                          int   mark);
    void printContext(const  ContextID& ctxID, 
                     std::ostream& str,
                     bool exception = false);
    void printContextStack(bool exception = false);

    void logLocation(std::stringstream& str, 
                     const  LocationID& locID);

    void showAllBreakPoints();

  public:
    //========== GRDebugNotification

    void eventHandligStart();
    void eventHandligFinish();

    void breakpoint(const  LocationID& loc, bool pause);
    void stepIn(const  LocationID& loc);
    void step(const  LocationID& loc);
    void stepOut(const  LocationID& loc);
    void ctxPop(const  LocationID& loc);
    void stopped();
    void exception(const  LocationID& loc);
    void start();
    void finish();
    void engineIsUp();
    void engineIsDown();
    void saved();
    void loaded();
    void closed();
    void progress(double);
    //========== GRDebuggerAPI
    void setFrontEnd( IGRDFrontEnd* frontEnd);
    GRDebugNotification* getNotificationInterface();
    IGRDebuggerCtrl* getIGRDebuggerCtrl();
    void setIGRDebuggerView(IGRDebuggerView* );
    void init(const std::string&);

    //========== IGRDebuggerCtrl
    void startCtrl();
    bool stepCtrl(const std::string& cmd);
    void stopCtrl();

  public:
  
  GRDebugger();
  ~GRDebugger(){}

    
};
};//namespace GRDI


#endif