#if !defined(GRDebuggerConsol_h)
#define GRDebuggerConsol_h

#include <boost/thread/mutex.hpp>

#include "../GRDebugger/IGRDebuggerCtrl.h"
#include "../GRDebugger/IGRDebuggerView.h"
#include "../GRLauncher/GRDebuggerAPI.h"


typedef boost::mutex::scoped_lock Lock;

class GRDebuggerConsol : public GRDI::IGRDebuggerView,
                         public GRDebuggerConsolAPI
{
protected:

  GRDI::IGRDebuggerCtrl* m_debuggerCtrl;

  boost::mutex      m_mutexPrint;

  bool              m_init;

  std::string       m_sourceTxt;

  void printLoc(const GRDI::LocationID& locID,
                const std::string& name, 
                std::stringstream& str,
                const GRDI::LocationID& markLocID,
                int mark);
  void buildFunctionSourceConsolImpl(
                                     const GRDI::Location& loc,
                                       const GRDI::FunctionDescr& descr, 
                                       std::stringstream& str,
                                       const GRDI::LocationID& markLocID = GRDI::invalidLocationID,
                                       int mark = 0);

public :
  
  GRDebuggerConsol();

  void setIGRDebuggerCtrl(GRDI::IGRDebuggerCtrl* debuggerCtrl);
  GRDI::IGRDebuggerView* getIGRDebuggerView();

  void operator()();
  void run();
  //======IGRDebuggerView
  void print(const std::string&);
  void buildFunctionSourceImpl( const GRDI::Location& loc,
                                const GRDI::FunctionDescr& descr, 
                                std::stringstream& str);
  void markFuncLocationImpl(  std::stringstream& str,
                              GRDI::IGRDebuggerView::FuncLocationMarkDescrVec& locs);

  void buildSourceImplBegin(std::stringstream& str);
  void buildSourceImplEnd(std::stringstream& str);
  void printSource();

  void printContextImpl(const GRDI::ContextDescr& contextDescr, 
                                  std::ostream& str,
                                  bool exception);
  void printContextStackImplBegin(std::ostream& str, bool exception);
  void printContextStackImplEnd(std::ostream& str, bool exception);

  void stopped();
  void exception();
  void finished();
  void loaded();
  void saved();
  void closed();
  void hitBreakpoint();
  void step();
  void progressReport(double);
  void printCurrentProgress(double);

};


#endif //#GRDebuggerConsol_h
