#pragma warning( disable : 4503)
#include <sstream>
#include "GReATDebugger.h"
#include "GReATDebuggerVars.h"
#include "../GRDBackEnd/IGRDBackend.h"
#include "GReATParametersFactory.h"
#include "GReATFactory.h"
#include "GRCC.h"
#include "GRCNames.h"
#include "GRCLoader.h"
#include "GReATConfig.h"
#include "GReATConfigMeta.h"
#include "GReATParameters.h"
#include "GReATUtils.h"
#include <algorithm>

#include <fstream>
#include <iterator>

//================================
bool GReATDebugger::execute()
{
std::stringstream str;
bool exc = false;
try
{

  GR::RewritingRule startRule;
  m_grParams->getStartRule(startRule);
  PassAlongPackets agruments;
  m_grParams->getPassAlongPackets(agruments);

  m_great->Execute(startRule, agruments);
}
catch( const std::invalid_argument& stdEx )
{
  str << "Debugger internal exception. Reason: " << stdEx.what() << std::endl;
  exc  = true;
}
catch( const udm_exception& udmEx )
{
  str << "GReAT Udm exception " << udmEx.what() << std::endl;
  exc  = true;
}
catch(...)
{
  str << "GReAT execution unknown exception." << std::endl;
  exc  = true;
}
 if (exc)
 {
   if (m_back)
      m_back->exception(str.str());
   else
     std::cout << str.str() << std::endl;
 }
 return !exc;
}
//================================
void GReATDebugger::checkInterActions()
{
  int status = m_back->getSynchStatus();
  if (status & GRDI::IGRDBackEnd::SAVE)
  {
    if (saveProgram())
    {
      m_back->saved();
    }
  }
}

//================================
void GReATDebugger::runDebugMode()
{
  m_back->up();

  m_back->synchronize();

  while(true) 
  {
    std::stringstream whileExcStr;
    bool whileExc = false;
    try
    {

     int status = m_back->getSynchStatus();
     if (status == 0)
     {
       m_back->synchronize();
       status = m_back->getSynchStatus();
     }

     if (status & GRDI::IGRDBackEnd::QUIT)
       break;
     else
     if (status & GRDI::IGRDBackEnd::STOP)
     {
       if (closeProgram()) 
        m_back->stopped();
     }
     else
     if (status & GRDI::IGRDBackEnd::LOAD)
     {
        if(loadGRParams(m_back->getProgramFileName(),
                        m_workingDirectory /*m_back->getWD() not implemented*/))
        {
          buildSource();    
          m_back->loaded();
        }
     }
     else
     if (status & GRDI::IGRDBackEnd::CLOSE)
     {
        clearSource();
        if (closeGRParams())
        {
          m_back->closed();
        }
     }
     else
     if (status & GRDI::IGRDBackEnd::RUN)
     {
       if (reloadProgram())
       {
         m_back->start();
         if (execute() && !quit())
           saveProgram();
         m_back->finish();
       }
       closeProgram();
     }
     checkInterActions();
   } //try
   catch( const std::invalid_argument& stdEx )
   {
      whileExcStr << "Debugger internal exception. Reason: " << stdEx.what() << std::endl;
      whileExc  = true;
   }
   catch( const udm_exception& udmEx )
   {
      whileExcStr << "Udm exception. Reason: " << udmEx.what() << std::endl;
      whileExc  = true;
   }
   catch(...)
   {

      whileExcStr << "Debugger internal unknown exception." << std::endl;
      whileExc  = true;
   }

   if (whileExc)
      m_back->exception(whileExcStr.str());

  }//while

  closeGRParams();
  

  m_back->down();
  m_back->synchronize();
}
//=========================================
void GReATDebugger::operator()()
{
  run();
}

//=========================================
void GReATDebugger::handleGRParamsException(const std::string& str)
{
 try
 {
    if (m_back)
    m_back->exception(str);
   else
     std::cout << str;
   m_grParams->closeNoSave();
   if(m_grParams)
     delete m_grParams, m_grParams = 0;
 }
 catch(...)
 {
   //already caugth an exception
 }
}
//=========================================
bool GReATDebugger::closeProgram()
{
   try
   {
    if (m_loaded)
    {
      m_grParams->closeProgramNoSave();
    }
   }
   catch( const udm_exception& udmEx )
   {
     std::stringstream str;
     str << "Backend close exception: " << udmEx.what() << std::endl;
     handleGRParamsException(str.str());
     return false;
   }
   catch(...)
   {
     std::stringstream str;
     str << "Backend close unkown exception" << std::endl;
     handleGRParamsException(str.str());
     return false;
   }
  return true;
}

//=========================================
bool GReATDebugger::saveProgram()
{
   try
   {
    if (m_loaded)
    {
      if (m_grParams->saveProgram() == IGReATParameters::FAILED)
        handleGRParamsException("Save exception: can not save the MGA backend" );
      else
        if (m_grParams->saveProgram() == IGReATParameters::HALF_SUCCESS)
          handleGRParamsException("Save exception: can not save the MGA backend, although the corresponding XML file has been saved." );
    }
   }
   catch( const udm_exception& udmEx )
   {
     std::stringstream str;
     str << "Backend save exception: " << udmEx.what() << std::endl;
     handleGRParamsException(str.str());
 
     return false;
   }
   catch(...)
   {
     std::stringstream str;
     str << "Backend save unkown exception" << std::endl;
     handleGRParamsException(str.str());
     return false;
   }
  return true;
}
//=========================================
bool GReATDebugger::reloadProgram()
{
   try
   {
     m_grParams->reloadProgram();
 
     m_grParams->getGeneralBackend(m_generalBackEnd);

    UdmDataNetworkPtr transformation;
    m_grParams->getTransformation(transformation);

    m_great->init(m_generalBackEnd, transformation);
   }
   catch( const udm_exception& udmEx )
   {
     std::stringstream str;
     str << "Backend reload exception: " << udmEx.what() << std::endl;
     handleGRParamsException(str.str());
     m_loaded = false;
     return false;
   }
   catch(...)
   {
     std::stringstream str;
     str << "Backend reload unkown exception" << std::endl;
     handleGRParamsException(str.str());
     std::cout << str .str();
     m_loaded = false;
     return false;
   }
  return true;
}

//================================
bool GReATDebugger::closeGRParams()
{
 try
 {
  if (m_loaded)
  {
    m_grcLoader->closeDll();
    if (m_grcLoader)
     delete m_grcLoader, m_grcLoader = 0;

    if(m_grParams)
     delete m_grParams, m_grParams = 0;
    m_loaded = false;
  }
 }
 catch( const udm_exception& udmEx )
 {
   std::stringstream str;
   str << "GRParameters close exception: " << udmEx.what() << std::endl;
   handleGRParamsException(str.str());
   return false;
 }
 catch(...)
 {
   std::stringstream str;
   str << "GRParameters close unkown exception" << std::endl;
   handleGRParamsException(str.str());
   return false;
 }

 return true;
}


//================================
bool GReATDebugger::loadGRParams(const std::string& pfn,
                                 const std::string& wd)
{
   std::stringstream str;
   bool ex = false;
   try
   {
    if (!closeGRParams())
      return false;


    m_grParams = GReATParametersFactory::create();

		m_grParams->setCurrentWorkingPath(wd);

    m_grParams->open(pfn);

    GeneralBackendPtr dataNetworkMap;
    m_grParams->getGeneralBackend(dataNetworkMap);

    UdmDataNetworkPtr transformation;
    m_grParams->getTransformation(transformation);

    m_great->init(dataNetworkMap, transformation);


///////////////////////////////////////////////////////////////////////////////
/*
    m_cintDll = CINTCodeGathererFactory::create();

    UserCodeGenConfig cfg;
    cfg.m_depFileName = m_grParams->getTransfLastWrittenFileName();
    cfg.m_udmDirName = m_grParams->getUdmDirName();
    cfg.m_transf = transformation;
    cfg.m_metas = m_grParams->getMetas();

//    StrVec incFiles, libFiles, incDirs, libDirs;
//    m_grParams->getUserCodeInfo(incFiles, libFiles, incDirs, libDirs);

//    cfg.m_incFiles = incFiles;
//    cfg.m_libFiles = libFiles;
//    cfg.m_incDirs = incDirs;
//    cfg.m_libDirs = libDirs;

    #ifdef _DEBUG
      cfg.m_debug = true;
    #else
      cfg.m_debug = false;
    #endif

    m_cintDll->generateDll(cfg);
    HINSTANCE hDll  = m_cintDll->loadDll();
    m_great->setCINTDll(hDll);
*/
///////////////////////////////////////////////////////////////////////////////
	GReATConfig::RootFolder cfgRoot= GReATConfig::RootFolder::Cast( ((GReATParameters*)m_grParams)->sdnConfig->GetRootObject());
	std::set< GReATConfig::MetaInformation> mis= cfgRoot.MetaInformation_kind_children();	
	GReATConfigMeta metaInfo( *mis.begin());
	std::string configPath= getPath( pfn);

	CCodeGenConfig cc_cfg;
	cc_cfg._transFileName= getFullPath( configPath, metaInfo.getTransformationFileName());
	cc_cfg._depFileName= cc_cfg._transFileName;
	cc_cfg._udmDirName= getPath( getFullPath( configPath, metaInfo.getUdmProjectFileName()), false);
	//
	typedef std::map< std::string, const Udm::DataNetwork*> BackendMap_t;
	BackendMap_t dns;
	metaInfo.getGeneralBackend( dns, configPath);
	StrList_t metanames;
	getFirst( std::make_pair( std::string(), (const Udm::DataNetwork*) 0));	// template instantiation to satisfy the linker
	std::transform( dns.begin(), dns.end(), std::back_inserter( metanames), getFirst< std::string, const Udm::DataNetwork*>);
	if ( metaInfo.isUdmProjectFile()) {
		metanames.push_back( metaInfo.getCrossLinkDiagramName());
	}
	cc_cfg._metas= metanames;
	//
    #ifdef _DEBUG
      cc_cfg._debug = true;
    #else
      cc_cfg._debug = false;
    #endif
	cc_cfg._dynamic= true;
	//
	GRCC grcc( cc_cfg);
	grcc.generateLibrary();
	//
	std::string dll_name= getGreatTempDir() + GRCNames::generateDllName( dropPath( dropExtension( cc_cfg._transFileName)),cc_cfg._debug);
	m_grcLoader= new GRCLoader( dll_name, m_grParams->getMetas());
    HINSTANCE hDll  = m_grcLoader->loadDll();
    m_great->setCINTDll(hDll);

   }
   catch( const udm_exception& udmEx )
   {
     std::stringstream str;
     str << "Datanetwork open exception: " << udmEx.what() << std::endl;
     handleGRParamsException(str.str());
     return false;
   }
   catch(...)
   {
     std::stringstream str;
     str << "Datanetwork open: unknown exception" << std::endl;
     handleGRParamsException(str.str());
     return false;
   }
  m_loaded = true;
  return true;
}

//=========================================
void GReATDebugger::runNonDebugMode()
{
  
  m_quitNonDebug = false;
  if (m_paramFileName.empty())
    throw std::invalid_argument("no parameter file");

  if (!quit() && loadGRParams(m_paramFileName, m_workingDirectory))
  {
//    std::cout << "GReAT has been launched in non-debug mode." << std::endl;
    std::cout << "The program has been loaded." << std::endl;

    if (!quit())
    {
      std::cout << "GReAT is running ..." << std::endl;
      if (execute() && !quit())
      {
        if (saveProgram() && closeGRParams())
           std::cout << "Successfully terminated." << std::endl;
      }
    }

  }
  if (quit())
  {
    std::cout << "The execution has been exited." << std::endl;
    m_grParams->closeNoSave();
  }
   closeGRParams();
  
}
