#include "GReATDebuggerVars.h"
#include <sstream>
#include <UdmUtil.h>

//=========================================
                 
void getObjPath(const Udm::Object& obj, std::string& str)
{
  Udm::Object par = obj;
  Udm::Object::uniqueId_type uid = obj.uniqueId();
  if (!par.GetParent()) {
	  str = "RootFolder";
  } else
  while( par )
  {
     // CHANGED "par.type().name()" to "par.type()" to fix empty string display in call stack.
		//if ((static_cast<const std::string>(par.type()) == "RootFolder")) {
	  if (static_cast<const std::string>(par.type().name()) == "RootFolder") {
			//str = "RootFolder"; ///Added
			break;
		}

    string name = UdmUtil::ExtractName(par);

    if (str.size())
      name += ".";
	//FIX FOR .NET
#if _MSC_VER >= 1310
	str.insert(0, name);
#else
	str.insert(0, name.begin());
#endif

    par = par.GetParent();
  }
}
//=========================================
namespace GRDI
{
IFunctionImpl::IFunctionImpl(const GR::RuleBase& rule):
 m_rule(rule),
 m_match(-1),
// m_id((GR::RewritingRule::Cast(rule)).originalRuleId())		///Added by Akash ///Change for "Rule viewing" in GME
m_id(0)
{
	if((const std::string&)rule.type().name() == "RewritingRuleProxy"){
		const GR::RuleBase& ref = GR::RewritingRuleProxy::Cast(rule).refersTo();
		m_id = (GR::RewritingRule::Cast(ref)).originalRuleId();
	}
	else
		m_id = (GR::RewritingRule::Cast(rule)).originalRuleId();

}
//=========================================
std::string IFunctionImpl::getFullPath() const
{
  std::string str;
  getObjPath(m_rule, str);
  return str;
}
};//namespace GRDI


//==========================================

GRFunction::GRFunction(const GR::RuleBase& rule):
IFunctionImpl(rule)
{
}
//=========================================
const GRDI::FunctionID GRFunction::getID() const
{
  return m_rule.uniqueId();
}
GRDI::uniqueId_type GRFunction::getRuleId() const {
	//return (GR::RewritingRule::Cast(m_rule)).originalRuleId();
	return m_id;
}
//=========================================
inline
GRDI::IFunctionImpl* GRFunction::getIFunctionImpl()
{
   return static_cast<GRDI::IFunctionImpl*>(this);
}
//=========================================
void GRFunction::print(std::ostream& str, int indent) const
{
  if (indent)
  {
    std::string tab;
    tab.resize(indent);
    str << tab;
  }

  str << m_rule.name();
  if (m_match >= 0)
  {
    if (m_match)
      str <<  " - case matched (" << m_match << " )";
    else
      str <<  " - case failed";
  }

}
//=========================================
//=========================================
GRVariable::GRVariable(const GR::ObjectWrapper& var) : 
m_var(var),
m_id(static_cast<GR::RewritingRule>(var.RewritingRule_parent()).originalRuleId())
{
}

//=========================================
void GRVariable::print(std::ostream& stream, int indent) const
{
  if (indent)
  {
    std::string tab;
    tab.resize(indent);
    stream << tab;
  }

//x  std::string str;
//x  getObjPath(*this, str);
  

  std::string typ;
  switch (m_type)
  {
    case IO : typ = "IO";break;     
    case INPUT : typ = "I";break;       
    case OUTPUT : typ = "O";break;   
    case PATTERN : typ = "P";break;
    default : typ = "lofaz";
  }
  stream << m_var.name() << ":" << typ;  
}
//=========================================
std::string GRVariable::name() const
{
  return (const std::string&) m_var.name();
}
//=========================================
inline
GRDI::IVariableImpl* GRVariable::getIVariableImpl()
{
   return static_cast<GRDI::IVariableImpl*>(this);
}

//=========================================
//=========================================
void GRVariableValue::print(std::ostream& str, int indent) const
{
  if (m_objList.size())
  {
    str << std::endl;
    UdmObjectList ::const_iterator it = m_objList.begin();
    for (; it != m_objList.end(); ++it)
    {
    
      //string name = UdmUtil::ExtractName(itm->second);

      std::string name;
      getObjPath(*it, name);

      if (indent)
      {
        std::string tab;
        tab.resize(indent);
        str << tab;
      }
      str << name << std::endl;
    }
  }
  else
    str << "has no value";
} 
//=========================================

GRDI::IVariableValueImpl* GRVariableValue::getIVariableValueImpl()
{
  return static_cast<GRDI::IVariableValueImpl*>(this);
}

