#include "StdAfx.h"
#include "UMLModelTransformer2GR.h"
#include "ErrorLog.h"
#include "umt_exception.h"

using namespace std;

IMPLEMENT_CUSTOMMODELBASE(UMT_ExpressionModel, CBuilderModel)
IMPLEMENT_CUSTOMMODELBASE(UMT_CompoundRuleModel, UMT_ExpressionModel)
IMPLEMENT_CUSTOMMODELBASE(UMT_PrimitiveRuleModel, UMT_ExpressionModel)

IMPLEMENT_CUSTOMMODEL(UMT_ForBlockModel, UMT_CompoundRuleModel,	"ForBlock")
IMPLEMENT_CUSTOMMODEL(UMT_BlockModel, UMT_CompoundRuleModel,	"Block")
IMPLEMENT_CUSTOMMODEL(UMT_TestModel, UMT_CompoundRuleModel,		"Test")
IMPLEMENT_CUSTOMMODEL(UMT_CaseModel, UMT_PrimitiveRuleModel,	"Case")
IMPLEMENT_CUSTOMMODEL(UMT_RuleModel, UMT_PrimitiveRuleModel,	"Rule")

IMPLEMENT_CUSTOMREF(UMT_ExpressionRefReference, CBuilderModelReference, "ExpressionRef")

IMPLEMENT_CUSTOMATOM(UMT_PortAtom, CBuilderAtom,	"In,Out")

IMPLEMENT_CUSTOMREF(UMT_PatternObjectReference, CBuilderAtomReference, "PatternClass")
IMPLEMENT_CUSTOMCONNECTION(UMT_PatternAssociation, CBuilderConnection,	"PatternAssociation")

using namespace GR;
using namespace std;

// Implementation of common functions in the base class "Expression"
bool UMT_ExpressionModel::Phase1_CreateObjects()
{	UdmPeer = RewritingRule::Create(UdmParent);
	UdmPeer.name() = string(GetName());
	UdmPeer.forBlock() = false;
	UdmPeer.exhaustive() = true;  
	UdmPeer.cut() = false;
	UdmPeer.guard() = string();
	UdmPeer.ruleId() = UdmPeer.uniqueId();
	UdmPeer.originalRuleId() = GetObjId();		///Change for "Rule viewing" in GME
	
	CString aspect = "Transform";
	CRect loc;
	CString posn;
	GetLocation(aspect, loc);
	posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
	UdmPeer.position() = string(posn);
	
	SetSpecialPeerAttr();
	CreateInterface();
	CreatePatternObjects();
	
	const CBuilderModelList *submodels = GetModels();
	POSITION pos = submodels->GetHeadPosition();
	while(pos)
	{
		UMT_ExpressionModel *expression = BUILDER_CAST(UMT_ExpressionModel, submodels->GetNext(pos));
		if(!expression)
			continue;
		expression->SetUdmParent(UdmPeer);
		expression->Phase1_CreateObjects();
	}
	
	return true;	
}

bool UMT_ExpressionModel::Phase2_CreateReferencesAndConnections()
{	CreateExpressionRef();
	CreatePassAlong();
	CreateSequence();
	const CBuilderModelList *submodels = GetModels();
	POSITION pos = submodels->GetHeadPosition();
	while(pos)
	{
		UMT_ExpressionModel *expression = BUILDER_CAST(UMT_ExpressionModel, submodels->GetNext(pos));
		if(!expression)
			continue;
		expression->SetUdmParent(UdmPeer);
		expression->Phase2_CreateReferencesAndConnections();
	}
	return true;
}

GR::ObjectWrapper UMT_ExpressionModel::GetPatternObjectConnectedTo(UMT_PortAtom *port)
{	UMT_PatternObjectReference *patRef; 
	CBuilderObjectList lst;
	GR::ObjectWrapper udmO = port->GetUdmPeer();
	if(udmO==Udm::null)
	{	if(port->GetKindName() == "In")
			port->GetOutConnectedObjects("Binding",lst);	
		else	
			port->GetInConnectedObjects("Binding",lst);
		if(lst.GetCount()!=1)
		{	
			printUMTMsg("In Rule:"+(string)GetName()+"-->SrcPort:"+(string)port->GetName()+" doesnt connect to 1 pattern");
//			AfxMessageBox("In Rule:"+GetName()+"-->SrcPort:"+port->GetName()+" doesnt connect to 1 pattern");
			return udmO;
		}
		patRef = BUILDER_CAST(UMT_PatternObjectReference, lst.GetHead());
		if(!patRef)
		{	
			printUMTMsg("In Rule:"+(string)GetName()+"-->Port:"+(string)port->GetName()+" connected to something other than pattern");
//			AfxMessageBox("In Rule:"+GetName()+"-->Port:"+port->GetName()+" connected to something other than pattern");
			return udmO;
		}
		return(patRef->GetUdmPeer());
	}
	return udmO;
}

bool UMT_ExpressionModel::IsAssociationClass(CBuilderObject *obj)
{
	const CBuilderConnectionList *isassoc = obj->GetInConnections("AssociationClass");
	if(isassoc && isassoc->GetCount() >0)
		return true;
	CBuilderObjectList refer;
	
	obj->GetReferencedBy(refer);
	if(refer.GetCount() > 0)
	{
		POSITION pos = refer.GetHeadPosition();
		while(pos)
		{
			CBuilderObject *ref = refer.GetNext(pos);
			const CBuilderConnectionList *refassoc = ref->GetInConnections("AssociationClass");
			if(refassoc  && refassoc->GetCount() >0)
				return true;
		}

	}
	
	
	CBuilderObjectList inherits;
	obj->GetInConnectedObjects("Sub", inherits);
	if(inherits.GetCount() == 0)
		return false;
	POSITION pos = inherits.GetHeadPosition();
	while(pos)
	{	CBuilderObject *inherit = inherits.GetNext(pos);
		CBuilderObjectList bases;
		inherit->GetInConnectedObjects("Base", bases);
		if(bases.GetCount() !=1)
			continue;
		CBuilderObject *base = bases.GetHead();
		if(IsAssociationClass(base))
			return true;
	}
	return false;
}

bool UMT_ExpressionModel::SortAtoms(const CBuilderAtomList *origionalList, CBuilderAtomList *sortedList, CString aspect)
{	POSITION pos = origionalList->GetHeadPosition();
	if(origionalList->GetCount() < 1 )
	{	return true;
	}
	
	int flag;
	int count = origionalList->GetCount();
	CRect mr, smr;
	while(pos)
	{	CBuilderAtom *atom = origionalList->GetNext(pos);
		POSITION spos = sortedList->GetHeadPosition();
		POSITION prev = spos;
		atom->GetLocation(aspect, mr);
		CString mname = atom->GetName();
		flag=false;
		while(spos)
		{	prev = spos;
			CBuilderAtom *sortedAtom = sortedList->GetNext(spos);
			sortedAtom->GetLocation(aspect, smr);
			if(mr.TopLeft().y < smr.TopLeft().y || (mr.TopLeft().y == smr.TopLeft().y && mr.TopLeft().x < smr.TopLeft().x))
			{	sortedList->InsertBefore(prev, atom);
				flag=true;
				break;
			}
		}
		if(!flag)
		{	sortedList->AddTail(atom);
		}
		
	}

	if(sortedList->GetCount() != count)
	{	return false;
	}
	return true;
}

bool UMT_ExpressionModel::SortReference(const CBuilderReferenceList *origionalList, CBuilderReferenceList *sortedList, CString aspect)
{	POSITION pos = origionalList->GetHeadPosition();
	if(origionalList->GetCount() < 1 )
	{	return true;
	}
	
	int flag;
	int count = origionalList->GetCount();
	CRect mr, smr;
	while(pos)
	{	CBuilderReference *reference = origionalList->GetNext(pos);
		POSITION spos = sortedList->GetHeadPosition();
		POSITION prev = spos;
		reference->GetLocation(aspect, mr);
		CString mname = reference->GetName();
		flag=false;
		while(spos)
		{	prev = spos;
			CBuilderReference *sortedReference = sortedList->GetNext(spos);
			sortedReference->GetLocation(aspect, smr);
			if(mr.TopLeft().y < smr.TopLeft().y || (mr.TopLeft().y == smr.TopLeft().y && mr.TopLeft().x < smr.TopLeft().x))
			{	sortedList->InsertBefore(prev, reference);
				flag=true;
				break;
			}
		}
		if(!flag)
		{	sortedList->AddTail(reference);
		}
		
	}

	if(sortedList->GetCount() != count)
	{	return false;
	}
	return true;
}

// Implementation of common functions in the base class "CompoundRule"
bool UMT_CompoundRuleModel::CreateInterface()
{	
	CString aspect = "Transform";
	{	const CBuilderAtomList *uports = GetAtoms("In");
		CBuilderAtomList *ports = new CBuilderAtomList;
		SortAtoms(uports, ports, "Transform");
		POSITION pos = ports->GetHeadPosition();
		CRect loc;
		CString posn;
		while(pos)
		{	UMT_PortAtom *port = BUILDER_CAST(UMT_PortAtom, ports->GetNext(pos));
			if(!port)
				continue;
			InputPort UdmPort = InputPort::Create(UdmPeer);
			UdmPort.name() = string(port->GetName());
/*			port->GetLocation(aspect, loc);
			posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
			UdmPort.position() = string(posn);
*/	
			UdmPort.cardinality() = "1";
			UdmPort.objectID() = 0;
			UdmPort.action() = "Bound";
			UdmPort.classType() = string();
			UdmPort.paradigmName() = string();
			UdmPort.nameSpace() = string();
			

			port->SetUdmPeer(UdmPort);
			port->SetUdmParent(UdmPeer);
		}
	}
	{	const CBuilderAtomList *uports = GetAtoms("Out");
		CBuilderAtomList *ports = new CBuilderAtomList;
		SortAtoms(uports, ports, "Transform");
		POSITION pos = ports->GetHeadPosition();
		CRect loc;
		CString posn;
		while(pos)
		{	UMT_PortAtom *port = BUILDER_CAST(UMT_PortAtom, ports->GetNext(pos));
			if(!port)
				continue;
			OutputPort UdmPort = OutputPort::Create(UdmPeer);
			UdmPort.name() = string(port->GetName());
			/*port->GetLocation(aspect, loc);
			posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
			UdmPort.position() = string(posn);
*/
			UdmPort.cardinality() = "1";
			UdmPort.objectID() = 0;
			UdmPort.action() = "Bound";
			UdmPort.classType() = string();
			UdmPort.paradigmName() = string();
			UdmPort.nameSpace() = string();

			port->SetUdmPeer(UdmPort);
			port->SetUdmParent(UdmPeer);
		}
	}
	
	return true;
}

bool UMT_CompoundRuleModel::CreateExpressionRef()
{	const CBuilderModelReferenceList *submodels = GetModelReferences("ExpressionRef");
	if(!submodels)
		return true;
	POSITION pos = submodels->GetHeadPosition();
	while(pos)
	{
		const CBuilderObject *ref = submodels->GetNext(pos);
		UMT_ExpressionRefReference *expression = BUILDER_CAST(UMT_ExpressionRefReference, ref);
		if(!expression)
		{	
			printUMTMsg("In Rule:"+(string)GetName()+"--> Is not a Expression Reference ");
//			AfxMessageBox("In Rule:"+GetName()+"--> Is not a Expression Reference ");
			continue;
		}
		CBuilderReference *rref = BUILDER_CAST(CBuilderReference, ref);
		while(rref)
		{	ref = rref->GetReferred();
			if(!ref)
			{	break;
			}
			rref = BUILDER_CAST(CBuilderReference, ref);
		}
		
		if(!ref)
		{	
			printUMTMsg("In Rule:"+(string)GetName()+"--> NULL reference ");
//			AfxMessageBox("In Rule:"+GetName()+"--> NULL reference ");
			continue;
		}

		UMT_ExpressionModel *mdl = BUILDER_CAST(UMT_ExpressionModel, ref);

		if(!mdl)
		{	
			printUMTMsg("In Rule:"+(string)GetName()+"--> NULL reference ");
//			AfxMessageBox("In Rule:"+GetName()+"--> NULL reference ");
			continue;
		}
		if(UdmPeer == Udm::null)
			continue;
		GR::RewritingRuleProxy proxy = GR::RewritingRuleProxy::Create(UdmPeer);
		proxy.name() = string(expression->GetName());
		CString aspect = "Transform";
		CRect loc;
		CString posn;
		expression->GetLocation(aspect, loc);
		posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
		proxy.position() = string(posn);

		proxy.refersTo() = GR::RewritingRule::Cast(mdl->GetUdmPeer());

		expression->SetUdmPeer(proxy);

		const CBuilderReferencePortList &rports = expression->GetRefereePorts();
		POSITION ppos = rports.GetHeadPosition();
		while(ppos)
		{	CBuilderReferencePort *rport = rports.GetNext(ppos);
			UMT_PortAtom *report = BUILDER_CAST(UMT_PortAtom, rport->GetAtom());
			if(!report)
				continue;
			GR::ObjectWrapper ow = GetPatternObjectConnectedTo(report);
			if(ow==Udm::null)
			{	
				printUMTMsg("In Rule:"+(string)GetName()+"--> NULL reference ");
//				AfxMessageBox("In Rule:"+GetName()+"--> NULL reference ");
				continue;
			}
			GR::ObjectWrapperProxy owp = GR::ObjectWrapperProxy::Create(proxy);
			owp.name() = string(report->GetName());
/*			report->GetLocation(aspect, loc);
			posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
			owp.position() = string(posn);
*/
			owp.cardinality() = "1";
			owp.objectID() = 0;
			owp.action() = "Bound";
			owp.classType() = string();
			owp.paradigmName() = string();
			owp.nameSpace() = string();
			owp.refersTo() = ow;
			expression->childportmap[rport] = owp;
		}
		
		//create the refernces here
	}
	return true;
}

bool UMT_CompoundRuleModel::CreatePassAlong()
{	
	{	const CBuilderConnectionList *objects = GetConnections("Sequence");
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	CBuilderConnection *object = objects->GetNext(pos);
				
			ObjectWrapper srcUdm, dstUdm; 
				
			UMT_PortAtom *srcPat = BUILDER_CAST(UMT_PortAtom, object->GetSource());
			if(srcPat)
			{	if((srcUdm = GetPatternObjectConnectedTo(srcPat)) == Udm::null) 
					continue;
			}
			else
			{	CBuilderReferencePort *srcRefPort = BUILDER_CAST(CBuilderReferencePort, object->GetSource());
				if(!srcRefPort)
					continue;
				UMT_ExpressionRefReference *srcRefPortParent = BUILDER_CAST(UMT_ExpressionRefReference, srcRefPort->GetOwner());
				if(!srcRefPortParent)
					continue;
				srcUdm = srcRefPortParent->childportmap[srcRefPort];
				if(srcUdm == Udm::null)
					continue;
			}

			UMT_PortAtom *dstPat = BUILDER_CAST(UMT_PortAtom, object->GetDestination());
			if(dstPat)
			{	if((dstUdm = GetPatternObjectConnectedTo(dstPat)) == Udm::null) 
					continue;
			}
			else
			{	CBuilderReferencePort *dstRefPort = BUILDER_CAST(CBuilderReferencePort, object->GetDestination());
				if(!dstRefPort)
					continue;
				UMT_ExpressionRefReference *dstRefPortParent = BUILDER_CAST(UMT_ExpressionRefReference, dstRefPort->GetOwner());
				if(!dstRefPortParent)
					continue;
				dstUdm = dstRefPortParent->childportmap[dstRefPort];
				if(dstUdm == Udm::null)
					continue;
			}

			PassAlong UdmObject = PassAlong::Create(UdmPeer);
			UdmObject.name() = string(object->GetName());
			
			UdmObject.srcObjectWrapper_end() = srcUdm;
			UdmObject.dstObjectWrapper_end() = dstUdm;
		}
	}
	return true;
}



bool UMT_CompoundRuleModel::CreateSequence()
{	
	{	const CBuilderConnectionList *objects = GetConnections("Sequence");
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	CBuilderConnection *object = objects->GetNext(pos);
						
			CBuilderObject *srcPat, *dstPat;
			srcPat = object->GetSource();
			dstPat = object->GetDestination();

			if(!srcPat || !dstPat)
				continue;
			//if(srcPat->GetParent() == this || dstPat->GetParent() == this)
			//	continue;
			//if(srcPat->GetParent()->GetParent() != dstPat->GetParent()->GetParent())
			//	continue;

			RuleBase srcUdm, dstUdm;

			if(BUILDER_CAST(UMT_PortAtom, srcPat) != NULL )
			{	UMT_ExpressionModel *srcPatPar;
				srcPatPar = BUILDER_CAST(UMT_ExpressionModel,srcPat->GetParent());
				srcUdm = srcPatPar->GetUdmPeer();
				
			}
			else
			{	UMT_ExpressionRefReference *srcPatPar;
				srcPatPar = BUILDER_CAST(UMT_ExpressionRefReference, BUILDER_CAST(CBuilderReferencePort,srcPat)->GetOwner());
				srcUdm = srcPatPar->GetUdmPeer();
			}

			if(BUILDER_CAST(UMT_PortAtom, dstPat) != NULL )
			{	UMT_ExpressionModel *dstPatPar;
				dstPatPar = BUILDER_CAST(UMT_ExpressionModel,dstPat->GetParent());
				dstUdm = dstPatPar->GetUdmPeer();
				
			}
			else
			{	UMT_ExpressionRefReference *dstPatPar;
				dstPatPar = BUILDER_CAST(UMT_ExpressionRefReference, BUILDER_CAST(CBuilderReferencePort,dstPat)->GetOwner());
				dstUdm = dstPatPar->GetUdmPeer();
			}

			
			bool flag = false;
			
			set<Sequence> dsts = srcUdm.dstRule();
			for(set<Sequence>::iterator i=dsts.begin(); i!=dsts.end(); i++)
			{	if(RuleBase::Cast((*i).dstRule_end())	== dstUdm)	
				{	flag=true;
					continue;
				}
			}
			
			if(!flag)
			{
				Sequence seq = Sequence::Create(GetUdmPeer());
				seq.name() = string(object->GetName());
				seq.recursive() = false;
				seq.srcRule_end() = srcUdm;
				seq.dstRule_end() = dstUdm;
			}
			
		}
	}
	return true;
}

// Implementation of common functions in the base class "PrimitiveRule"
bool UMT_PrimitiveRuleModel::CreatePatternObjects()
{	
	CString aspect = "Transform";
	
	// Create the pattern objects
	{	const CBuilderReferenceList *uobjects = GetReferences("PatternClass");
		CBuilderReferenceList *objects = new CBuilderReferenceList;
		SortReference(uobjects, objects, "Transform");
		
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	UMT_PatternObjectReference *object = BUILDER_CAST(UMT_PatternObjectReference, objects->GetNext(pos));
			if(!object)
			{	
				printUMTMsg("In Rule:"+(string)GetName()+" There is a PatternClass that doesnt refer to an atom");
//				AfxMessageBox("In Rule:"+GetName()+" There is a PatternClass that doesnt refer to an atom");
				continue;
			}
			object->GetAttribute("Action",attr);
			if(attr == "CreateNew")
			{	const CBuilderConnectionList * cl = object->GetOutConnections("PatternComposition");
				if(!cl || cl->GetCount() != 1)
				{	
					printUMTMsg("In Rule:"+(string)GetName()+"-->Pattern:"+(string)object->GetName()+" A new object specified but no parent mentioned");
//					AfxMessageBox("In Rule:"+GetName()+"-->Pattern:"+object->GetName()+" A new object specified but no parent mentioned");
					continue;
				}
				
			}
			const CBuilderObject *refered = object->GetReferred();

			////
			CBuilderReference *rref = BUILDER_CAST(CBuilderReference, refered);
			const CBuilderObject *ref = BUILDER_CAST(CBuilderAtom, refered); 
			while(rref)
			{	ref = rref->GetReferred();
				if(!ref)
				{	break;
				}
				rref = BUILDER_CAST(CBuilderReference, ref);
			}
										
			PatternObject UdmObject = PatternObject::Create(UdmPeer);
			UdmObject.name() = string(object->GetName());
			// copy cmpFunction attribute. added on 03/31/05 by Viza
			CBuilderObjectList lst;
			object->GetOutConnectedObjects( "Binding", lst);
			if( lst.GetCount() ==1) {
				UMT_PortAtom* port= BUILDER_CAST( UMT_PortAtom, lst.GetHead());
				port->GetAttribute("CmpFunc",attr);
				UdmObject.cmpfunc() = string(attr);
			}

			object->GetAttribute("Action",attr);
			UdmObject.action() = string(attr);
			
			object->GetAttribute("PatternCardinality",attr);
			UdmObject.cardinality() = string(attr);

	/*		object->GetLocation(aspect, loc);
			posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
			UdmObject.position() = string(posn);
*/
			UdmObject.isAssociationClass() = false;

			
			
			if(!ref)
			{	
				printUMTMsg("In Rule:"+(string)GetName()+"-->Pattern:"+(string)object->GetName()+" has a null reference, assumed to be templatized");
				UdmObject.classType() = string();
				UdmObject.paradigmName() = string();
				UdmObject.nameSpace() = string();
			}
			else
			{	CBuilderAtom *refered = BUILDER_CAST(CBuilderAtom, ref);
				if(!refered)
				{	
					printUMTMsg("In Rule:"+(string)GetName()+"-->Pattern:"+(string)object->GetName()+" is not of type Class");
					continue;
				}
				UdmObject.classType() = string(refered->GetName());
//				CString folderName = refered->GetParent()->GetParent()->GetName();
			// VIZA changed on 11/22/05 parent of parent of refered is either a UML namespace, or a UML package, or UMT RootFolder
				const CBuilderModel* container= refered->GetParent()->GetParent();
				string ns_name;
				while ( container->GetKindName() == "Namespace")
				{
					string ns_ns = string(container->GetName());
					ns_name =(ns_name.empty()) ?  ns_ns : ns_ns+ "::"+ns_name;
					container= container->GetParent();					
				}
				
				UdmObject.nameSpace() = ns_name;

				CString folderName = container->GetName();
				if ( container->GetKindName() == "Package") {
					CString alias;
					container->GetAttribute( "alias", alias);
					if ( !alias.IsEmpty()) {
						folderName= alias;
					}
				}
				/*int end = folderName.ReverseFind(".");
				if(end==-1)
					end = folderName.GetLength();
				int start = folderName.ReverseFind('\\');
				CString mid = folderName.Mid(start+1, end-(start+1));*/
				UdmObject.paradigmName() = (string)folderName;
				
//				This was the origional code for finding out if a connection is a connection or not.
//				const CBuilderConnectionList *isassoc = refered->GetInConnections("AssociationClass");
//				if(isassoc && isassoc->GetCount() >0)
//					UdmObject.isAssociationClass() = true;

				if(IsAssociationClass(refered))
					UdmObject.isAssociationClass() = true;
			}
			
			UdmObject.objectID() = 0;
						
			object->SetUdmParent(UdmPeer);
			object->SetUdmPeer(UdmObject);
		}
	}
	// Create the composition links
	{	const CBuilderConnectionList *objects = GetConnections("PatternComposition");
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	CBuilderConnection *object = objects->GetNext(pos);
						
			ObjectLink UdmObject = ObjectLink::Create(UdmPeer);
			UdmObject.name() = string(object->GetName());
			
			object->GetAttribute("Action",attr);
			UdmObject.action() = string(attr);

			UdmObject.cardinality() = "1";
			object->GetAttribute("Cardinality",attr);
			if(string(attr)!="0..*") UdmObject.cardinality() = string(attr);
		 
	/*		object->GetLocation(aspect, loc);
			posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
			UdmObject.position() = string(posn);
	*/		
			UdmObject.relationshipType() = "Composition";
			
			object->GetAttribute("ParentRole",attr);
			UdmObject.dstRoleName() = string(attr);

			object->GetAttribute("ChildRole",attr);
			UdmObject.srcRoleName() = string(attr);
  
			UdmObject.srcObjectID() = 0;
			UdmObject.dstObjectID() = 0;

			UMT_PatternObjectReference *srcPat = BUILDER_CAST(UMT_PatternObjectReference, object->GetSource());
			UMT_PatternObjectReference *dstPat = BUILDER_CAST(UMT_PatternObjectReference, object->GetDestination());
			if(!srcPat || !dstPat)
				continue;

			UdmObject.src_end() = srcPat->GetUdmPeer();
			UdmObject.dst_end() = dstPat->GetUdmPeer();
		}
	}
	
	// Create the simple association links
	{	const CBuilderConnectionList *objects = GetConnections("PatternAssociation");
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	UMT_PatternAssociation *object = BUILDER_CAST(UMT_PatternAssociation, objects->GetNext(pos));
//			CBuilderConnection *object = objects->GetNext(pos);
						
			ObjectLink UdmObject = ObjectLink::Create(UdmPeer);
			UdmObject.name() = string(object->GetName());
			
			object->GetAttribute("Action",attr);
			UdmObject.action() = string(attr);

			UdmObject.cardinality() = "1";
			
			object->GetAttribute("PatternCardinality",attr);
			UdmObject.cardinality() = string(attr);

	/*		object->GetLocation(aspect, loc);
			posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
			UdmObject.position() = string(posn);
	*/		
			UMT_PatternObjectReference *objectsrc = BUILDER_CAST(UMT_PatternObjectReference , object->GetSource());
			UMT_PatternObjectReference *objectdst = BUILDER_CAST(UMT_PatternObjectReference , object->GetDestination());
			if (objectsrc->GetReferred() == 0)
				throw umt_exception("In Rule:" + (string)GetName() + "-->Pattern:" + (string)objectsrc->GetName() + " is a null reference.");
			CString objsrcfoln = objectsrc->GetReferred()->GetParent()->GetParent()->GetName();
			if (objectdst->GetReferred() == 0)
				throw umt_exception("In Rule:" + (string)GetName() + "-->Pattern:" + (string)objectdst->GetName() + " is a null reference.");
			CString objdstfoln = objectdst->GetReferred()->GetParent()->GetParent()->GetName();
			if(objsrcfoln == objdstfoln)
				UdmObject.relationshipType() = "SimpleAssociation";
			else
				UdmObject.relationshipType() = "ObjectMapping";
			
			object->GetAttribute("dstRolename",attr);
			UdmObject.dstRoleName() = string(attr);

			object->GetAttribute("srcRolename",attr);
			UdmObject.srcRoleName() = string(attr);

			UdmObject.srcObjectID() = 0;
			UdmObject.dstObjectID() = 0;

			UMT_PatternObjectReference *srcPat = BUILDER_CAST(UMT_PatternObjectReference, object->GetSource());
			UMT_PatternObjectReference *dstPat = BUILDER_CAST(UMT_PatternObjectReference, object->GetDestination());
			if(!srcPat || !dstPat)
				continue;

			UdmObject.src_end() = srcPat->GetUdmPeer();
			UdmObject.dst_end() = dstPat->GetUdmPeer();
			object->SetUdmParent(UdmPeer);
			object->SetUdmPeer(UdmObject);
		}
	}

	// Create the TracesTo links
	{	const CBuilderConnectionList *objects = GetConnections("TracesTo");
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	CBuilderConnection *object = objects->GetNext(pos);
						
			ObjectLink UdmObject = ObjectLink::Create(UdmPeer);
			UdmObject.name() = string(object->GetName());
			
			UdmObject.relationshipType() = "TracesTo";
			
			UMT_PatternObjectReference *srcPat = BUILDER_CAST(UMT_PatternObjectReference, object->GetSource());
			UMT_PatternObjectReference *dstPat = BUILDER_CAST(UMT_PatternObjectReference, object->GetDestination());
			if(!srcPat || !dstPat)
				continue;

			UdmObject.src_end() = srcPat->GetUdmPeer();
			UdmObject.dst_end() = dstPat->GetUdmPeer();
		}
	}
	

	// Create the associations with association classes using connector
	{	const CBuilderAtomList *objects = GetAtoms("Connector");
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	CBuilderAtom *object = objects->GetNext(pos);
			
			const CBuilderConnectionList *srcs, *dsts, *assocs;
			srcs = object->GetInConnections("Src");
			dsts = object->GetOutConnections("Dst");
			assocs = object->GetOutConnections("AssociationClass");
			if(srcs->GetCount() !=1 || dsts->GetCount() != 1 || assocs->GetCount() !=1)
			{	
				printUMTMsg("In Rule:"+(string)GetName()+"-->Connector doesn't have required links");
//				AfxMessageBox("In Rule:"+GetName()+"-->Connector doesnt have required links");
				continue;
			}
			CBuilderConnection *srcc, *dstc, *assocc;
			srcc = srcs->GetHead();
			dstc = dsts->GetHead();
			assocc = assocs->GetHead();

			UMT_PatternObjectReference *src, *dst, *assoc;
			src = BUILDER_CAST(UMT_PatternObjectReference , srcc->GetSource());
			dst = BUILDER_CAST(UMT_PatternObjectReference , dstc->GetDestination());
			assoc = BUILDER_CAST(UMT_PatternObjectReference , assocc->GetDestination());

			if(!src || !dst || !assoc)
			{	
				printUMTMsg("In Rule:"+(string)GetName()+"-->Connector connected to illegal object");
//				AfxMessageBox("In Rule:"+GetName()+"-->Connector connected to illegal object");
				continue;
			}

			ObjectLink UdmSrcObject = ObjectLink::Create(UdmPeer);
			UdmSrcObject.name() = string(srcc->GetName());
			
			assoc->GetAttribute("Action",attr);
			UdmSrcObject.action() = string(attr);
			
			assoc->GetAttribute("PatternCardinality",attr);
			UdmSrcObject.cardinality() = string(attr);

			UdmSrcObject.relationshipType() = "AssociationClass";

			ObjectLink UdmDstObject = ObjectLink::Create(UdmPeer);
			UdmDstObject.name() = string(dstc->GetName());
			
			assoc->GetAttribute("Action",attr);
			UdmDstObject.action() = string(attr);
			
			assoc->GetAttribute("PatternCardinality",attr);
			UdmDstObject.cardinality() = string(attr);

			UdmDstObject.relationshipType() = "AssociationClass";
			
			CString srcrole,dstrole; 
			
			srcc->GetAttribute("srcRolename",srcrole);
			dstc->GetAttribute("dstRolename",dstrole);

			UdmSrcObject.srcRoleName() = string(srcrole);
			UdmSrcObject.dstRoleName() = string(dstrole);

			
			UdmDstObject.dstRoleName() = string(dstrole);
			UdmDstObject.srcRoleName() = string(srcrole);

			UdmSrcObject.src_end() = src->GetUdmPeer();
			UdmSrcObject.dst_end() = assoc->GetUdmPeer();

			UdmSrcObject.srcObjectID() = 0;
			UdmSrcObject.dstObjectID() = 0;
 
			UdmDstObject.dst_end() = dst->GetUdmPeer();
			UdmDstObject.src_end() = assoc->GetUdmPeer();

			UdmDstObject.srcObjectID() = 0;
			UdmDstObject.dstObjectID() = 0;
		}
	}
	// Create the 'match any association' links
	{	const CBuilderConnectionList *objects = GetConnections("MatchAnyAssociation");
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	CBuilderConnection *object = objects->GetNext(pos);
						
			ObjectLink UdmObject = ObjectLink::Create(UdmPeer);
			UdmObject.name() = string(object->GetName());
			UdmObject.action() = "Bound";	// always bound
			UdmObject.cardinality() = "1";

	/*		object->GetLocation(aspect, loc);
			posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
			UdmObject.position() = string(posn);
	*/		
			UMT_PatternObjectReference *objectsrc = BUILDER_CAST(UMT_PatternObjectReference , object->GetSource());
			UMT_PatternObjectReference *objectdst = BUILDER_CAST(UMT_PatternObjectReference , object->GetDestination());
			CString objsrcfoln = objectsrc->GetReferred()->GetParent()->GetParent()->GetName();
			CString objdstfoln = objectdst->GetReferred()->GetParent()->GetParent()->GetName();
			if(objsrcfoln == objdstfoln)
				UdmObject.relationshipType() = "SimpleAssociationAny";
			else
				UdmObject.relationshipType() = "ObjectMappingAny";
			
			UdmObject.dstRoleName() = "";
			UdmObject.srcRoleName() = "";
			UdmObject.srcObjectID() = 0;
			UdmObject.dstObjectID() = 0;
			UMT_PatternObjectReference *srcPat = BUILDER_CAST(UMT_PatternObjectReference, object->GetSource());
			UMT_PatternObjectReference *dstPat = BUILDER_CAST(UMT_PatternObjectReference, object->GetDestination());
			if(!srcPat || !dstPat)
				continue;
			UdmObject.src_end() = srcPat->GetUdmPeer();
			UdmObject.dst_end() = dstPat->GetUdmPeer();
		}
	}
	// Create the associations with association classes using 'connector any'
	{	const CBuilderAtomList *objects = GetAtoms("ConnectorAny");
		POSITION pos = objects->GetHeadPosition();
		CRect loc;
		CString posn, attr;
		
		while(pos)
		{	CBuilderAtom *object = objects->GetNext(pos);
			
			const CBuilderConnectionList *srcs, *dsts, *assocs;
			srcs = object->GetInConnections("SrcAny");
			dsts = object->GetOutConnections("DstAny");
			assocs = object->GetOutConnections("AssociationClass");
			if(srcs->GetCount() !=1 || dsts->GetCount() != 1 || assocs->GetCount() !=1)
			{	
				printUMTMsg("In Rule:"+(string)GetName()+"-->ConnectorAny doesnt have required links");
//				AfxMessageBox("In Rule:"+GetName()+"-->ConnectorAny doesnt have required links");
				continue;
			}
			CBuilderConnection *srcc, *dstc, *assocc;
			srcc = srcs->GetHead();
			dstc = dsts->GetHead();
			assocc = assocs->GetHead();

			UMT_PatternObjectReference *src, *dst, *assoc;
			src = BUILDER_CAST(UMT_PatternObjectReference , srcc->GetSource());
			dst = BUILDER_CAST(UMT_PatternObjectReference , dstc->GetDestination());
			assoc = BUILDER_CAST(UMT_PatternObjectReference , assocc->GetDestination());

			if(!src || !dst || !assoc)
			{	
				printUMTMsg("In Rule:"+(string)GetName()+"-->Connector connected to illegal object");
//				AfxMessageBox("In Rule:"+GetName()+"-->Connector connected to illegal object");
				continue;
			}

			ObjectLink UdmSrcObject = ObjectLink::Create(UdmPeer);
			UdmSrcObject.name() = string(srcc->GetName());
			UdmSrcObject.action() = "Bound";	// always bound
			assoc->GetAttribute("PatternCardinality",attr);
			UdmSrcObject.cardinality() = string(attr);
			UdmSrcObject.relationshipType() = "AssociationClassAny";
			UdmSrcObject.srcRoleName() = "";
			UdmSrcObject.dstRoleName() = "";
			UdmSrcObject.src_end() = src->GetUdmPeer();
			UdmSrcObject.dst_end() = assoc->GetUdmPeer();
			UdmSrcObject.srcObjectID() = 0;
			UdmSrcObject.dstObjectID() = 0;

			ObjectLink UdmDstObject = ObjectLink::Create(UdmPeer);
			UdmDstObject.name() = string(dstc->GetName());
			UdmDstObject.action() = "Bound";	// always bound
			assoc->GetAttribute("PatternCardinality",attr);
			UdmDstObject.cardinality() = string(attr);
			UdmDstObject.relationshipType() = "AssociationClassAny";
			UdmDstObject.dstRoleName() = "";
			UdmDstObject.srcRoleName() = "";
			UdmDstObject.dst_end() = dst->GetUdmPeer();
			UdmDstObject.src_end() = assoc->GetUdmPeer();
			UdmDstObject.srcObjectID() = 0;
			UdmDstObject.dstObjectID() = 0;
		}
	}
	if((string)GetKindName()=="Case") return true;
	// Create the Group - added by Feng 7/27/06
	{
		const CBuilderSetList *objects = GetSets("Group");
		
		POSITION pos = objects->GetHeadPosition();
		CString attr;
		
		while(pos)
		{	
			CBuilderSet *object = objects->GetNext(pos);
			PatternGroup UdmObject = PatternGroup::Create(UdmPeer);
			UdmObject.name() = string(object->GetName());
			object->GetAttribute("Criterial",attr);
			UdmObject.rule() = (string)attr;
	
			object->GetAttribute("GroupAction",attr);
			UdmObject.action() = (string)attr;
			if(attr == "Move" || attr == "Copy")
			{
				const CBuilderConnectionList *dsts;
				dsts = object->GetOutConnections("MoveCopy");
				if(!dsts->IsEmpty())
				{	
					POSITION dpos = dsts->GetHeadPosition();
					while(dpos)
					{
						CBuilderConnection *dstc = dsts->GetNext(dpos);
						UMT_PatternObjectReference *dst = BUILDER_CAST(UMT_PatternObjectReference , dstc->GetDestination());
						ObjectLink movecopy = ObjectLink::Create(UdmPeer);
						movecopy.name() = "MoveCopyGroup";
						movecopy.action() = string(attr);

						movecopy.relationshipType() = "MoveCopy";
						movecopy.src_end() = UdmObject;
						movecopy.dst_end() = dst->GetUdmPeer();
					}
				}
				else
					printUMTMsg("In Rule:"+(string)GetName()+"-->Group doesn't have required link MoveCopy");
			}

			const CBuilderObjectList *members = object->GetMembers();
			POSITION mpos = members->GetHeadPosition();
			while(mpos)
			{
				CBuilderObject *member = members->GetNext(mpos);
				const CString kname = member->GetKindName();
				if(kname == "PatternClass")
				{
					UMT_PatternObjectReference *pmember = BUILDER_CAST(UMT_PatternObjectReference, member);
					UdmObject.members() += pmember->GetUdmPeer();
				}
				else if(kname == "PatternAssociation")
				{
					UMT_PatternAssociation *pmember = BUILDER_CAST(UMT_PatternAssociation, member);
					UdmObject.members() += pmember->GetUdmPeer();
				}
			}
		}
	}
	return true;
}


// Implementation for "ForBlock"
bool UMT_ForBlockModel::SetSpecialPeerAttr()
{	UdmPeer.forBlock() = true;
	UdmPeer.ruleType() = "BLOCK";
	return true;	
}


// Implementation for "Block"
bool UMT_BlockModel::SetSpecialPeerAttr()
{	UdmPeer.forBlock() = false;
	UdmPeer.ruleType() = "BLOCK";
	return true;	
}


// Implementation for "Test"
bool UMT_TestModel::SetSpecialPeerAttr()
{	UdmPeer.forBlock() = true;
	UdmPeer.ruleType() = "TEST";
	bool foreach;
	if(GetAttribute("ForEach", foreach))
		if(foreach)
			UdmPeer.forBlock() = true;
		else
			UdmPeer.forBlock() = false;
	return true;	
}

bool UMT_TestModel::CreateSequence()
{	CString aspect = "Transform";
	const CBuilderModelList *submodels = GetModels();
	CBuilderModelList sortedmodels;
	POSITION pos = submodels->GetHeadPosition();
	if(submodels->GetCount() <1 )
	{	CString ename;
		GetExtendedName(ename,"->",true);
		printUMTMsg("Test "+(string)ename+" Doesnt Contain any cases");
//		AfxMessageBox("ERROR: Test "+ename+" Doesnt Contain any cases");
		return true;
	}

	int flag;
	int count = submodels->GetCount();
	CRect mr, smr;
	while(pos)
	{	CBuilderModel *model = submodels->GetNext(pos);
		POSITION spos = sortedmodels.GetHeadPosition();
		POSITION prev = spos;
		model->GetLocation(aspect, mr);
		CString mname = model->GetName();
		flag=false;
		while(spos)
		{	prev = spos;
			CBuilderModel *sortedmodel = sortedmodels.GetNext(spos);
			sortedmodel->GetLocation(aspect, smr);
			if(mr.TopLeft().y < smr.TopLeft().y || (mr.TopLeft().y == smr.TopLeft().y && mr.TopLeft().x < smr.TopLeft().x))
			{	sortedmodels.InsertBefore(prev, model);
				flag=true;
				break;
			}
		}
		if(!flag)
		{	sortedmodels.AddTail(model);
		}
		
	}

	if(sortedmodels.GetCount() != count)
	{	CString ename;
		GetExtendedName(ename,"->",true);
		printUMTMsg("Test "+(string)ename+": Sorting of cases failed");
//		AfxMessageBox("ERROR: Test "+ename+": Sorting of cases failed");
		return true;
	}

	POSITION spos = sortedmodels.GetHeadPosition();
	UMT_ExpressionModel *curr = BUILDER_CAST(UMT_ExpressionModel, sortedmodels.GetNext(spos));
	Sequence seq = Sequence::Create(GetUdmPeer());
	seq.name() = string();
	seq.recursive() = false;
	seq.srcRule_end() = GetUdmPeer();
	seq.dstRule_end() = curr->GetUdmPeer();
	while(spos)
	{	UMT_ExpressionModel *next = BUILDER_CAST(UMT_ExpressionModel, sortedmodels.GetNext(spos));
		if(!curr || !next)
			continue;
		seq = Sequence::Create(GetUdmPeer());
		seq.name() = string();
		seq.recursive() = false;
		seq.srcRule_end() = curr->GetUdmPeer();
		seq.dstRule_end() = next->GetUdmPeer();
		curr = next;
	}
	seq = Sequence::Create(GetUdmPeer());
	seq.name() = string();
	seq.recursive() = false;
	seq.srcRule_end() = curr->GetUdmPeer();
	seq.dstRule_end() = GetUdmPeer();
	
	return true;
}

// Implementation for "Case"
bool UMT_CaseModel::SetSpecialPeerAttr()
{	UdmPeer.forBlock() = false;
	UdmPeer.ruleType() = "CASE";
	if(GetAtoms("Cut")->GetCount() >=1)
		UdmPeer.cut() = true;
	bool battr;
	GetAttribute("ForAll",battr);
	UdmPeer.exhaustive() = battr;

	bool buniqueIdentity = true;
	GetAttribute("UniqueIdentity", buniqueIdentity);
	UdmPeer.uniqueIdentity() = buniqueIdentity;

	CString attr;
	const CBuilderAtomList *objects = GetAtoms("Guard");
	if(objects->GetCount()>1)
		printUMTMsg("Rule:"+(string)GetName()+" More than one guard found using only one ");
//		AfxMessageBox("Rule:"+GetName()+" More than one guard found using only one ");
	if(objects->GetCount()>=1)
	{	objects->GetHead()->GetAttribute("ExpressionString",attr);
		UdmPeer.guard() = string(attr);
	}

	return true;	
}


// Implementation for "Rule"
bool UMT_RuleModel::SetSpecialPeerAttr()
{	UdmPeer.forBlock() = false;
	UdmPeer.ruleType() = "RULE";
	
	CString attr;
	const CBuilderAtomList *objects = GetAtoms("Guard");
	if(objects->GetCount()>1)
		printUMTMsg("Rule:"+(string)GetName()+" More than one guard found using only one ");
//		AfxMessageBox("Rule:"+GetName()+" More than one guard found using only one ");
	if(objects->GetCount()>=1)
	{	objects->GetHead()->GetAttribute("ExpressionString",attr);
		UdmPeer.guard() = string(attr);
	}
	
	bool battr;
	GetAttribute("ForAll",battr);
	UdmPeer.exhaustive() = battr;
	
	bool buniqueIdentity = true;
	GetAttribute("UniqueIdentity", buniqueIdentity);
	UdmPeer.uniqueIdentity() = buniqueIdentity;

	bool bdistinguished= false;
	GetAttribute("Distinguished", bdistinguished);
	UdmPeer.distinguished()= bdistinguished;

	objects = GetAtoms("AttributeMapping");
	POSITION pos = objects->GetHeadPosition();
	CString aspect = "Transform";
	CRect loc;
	CString posn;
	while(pos)
	{	CBuilderAtom *object = objects->GetNext(pos);
		object->GetAttribute("ExpressionString",attr);
		if(!attr.IsEmpty())
		{
			Code c = Code::Create(UdmPeer);
			c.code() = string(attr);
		}
		/*c.name() = object->GetName();
		
		object->GetLocation(aspect, loc);
		posn.Format("(%d,%d)",loc.TopLeft().x,loc.TopLeft().y);
		c.position() = string(posn);*/
		
	}

	return true;	
}

// Implementation for "ExpressionRef"
bool UMT_ExpressionRefReference::SetSpecialPeerAttr()
{	
	return true;	
}