#ifndef UMLMODELTRANSFORMER2GR_H
#define UMLMODELTRANSFORMER2GR_H

#include "Builder.h"
#include "GR.h"

class UMT_ExpressionModel;
class UMT_CompoundRuleModel;
class UMT_PrimitiveRuleModel;

class UMT_ForBlockModel;
class UMT_BlockModel;
class UMT_TestModel;
class UMT_CaseModel;
class UMT_RuleModel;
class UMT_ExpressionRefReference;
class UMT_PortAtom;
class UMT_PatternAssociation;

// Custom base class for "Expression"
class UMT_ExpressionModel : public CBuilderModel
{
	DECLARE_CUSTOMMODELBASE(UMT_ExpressionModel, CBuilderModel)
public:
	bool Phase1_CreateObjects();
	bool Phase2_CreateReferencesAndConnections();
	
	virtual bool SetSpecialPeerAttr()=0;
	virtual bool CreateInterface()=0;
	virtual bool CreatePatternObjects()=0;
	virtual bool CreatePassAlong()=0;
	virtual bool CreateSequence()=0;
	virtual bool CreateExpressionRef()=0;
	
protected:
	GR::RewritingRule UdmPeer;
	Udm::Object UdmParent;
public:
	virtual GR::RuleBase GetUdmPeer() { return UdmPeer; }
	virtual bool SetUdmPeer(GR::RewritingRule Peer) { UdmPeer = Peer; return true;}

	virtual Udm::Object GetUdmParent() { return UdmParent; }
	virtual bool SetUdmParent(Udm::Object Parent) {UdmParent = Parent; return true;}
	
	GR::ObjectWrapper GetPatternObjectConnectedTo(UMT_PortAtom *port);
	bool IsAssociationClass(CBuilderObject *);
	bool SortAtoms(const CBuilderAtomList *origionalList, CBuilderAtomList *sortedList, CString aspect);
	bool SortReference(const CBuilderReferenceList *origionalList, CBuilderReferenceList *sortedList, CString aspect);

};

// Custom base class for "CompundRule"
class UMT_CompoundRuleModel : public UMT_ExpressionModel
{
	DECLARE_CUSTOMMODELBASE(UMT_CompoundRuleModel, UMT_ExpressionModel)
public:
	
	virtual bool SetSpecialPeerAttr()=0;
	virtual bool CreateInterface();
	virtual bool CreatePatternObjects() {return true;}
	virtual bool CreatePassAlong();
	virtual bool CreateSequence();
	virtual bool CreateExpressionRef();
};

// Custom base class for "PrimitiveRule"
class UMT_PrimitiveRuleModel : public UMT_ExpressionModel
{
	DECLARE_CUSTOMMODELBASE(UMT_PrimitiveRuleModel, UMT_ExpressionModel)
public:
	virtual bool SetSpecialPeerAttr()=0;
	virtual bool CreateInterface() {return true;}
	virtual bool CreatePatternObjects();
	virtual bool CreatePassAlong() {return true;}
	virtual bool CreateSequence() { return true;}
	virtual bool CreateExpressionRef() {return true;}
};

// Custom class for "ForBlock"
class UMT_ForBlockModel : public UMT_CompoundRuleModel
{
	DECLARE_CUSTOMMODEL(UMT_ForBlockModel, UMT_CompoundRuleModel)
public:
	virtual bool SetSpecialPeerAttr();

private:

};

// Custom class for "Block"
class UMT_BlockModel : public UMT_CompoundRuleModel
{
	DECLARE_CUSTOMMODEL(UMT_BlockModel, UMT_CompoundRuleModel)
public:
	virtual bool SetSpecialPeerAttr();
private:
};

// Custom class for "Test"
class UMT_TestModel : public UMT_CompoundRuleModel
{
	DECLARE_CUSTOMMODEL(UMT_TestModel, UMT_ExpressionModel)
public:
	virtual bool SetSpecialPeerAttr();
	virtual bool CreateSequence();
private:
};

// Custom class for "Case"
class UMT_CaseModel : public UMT_PrimitiveRuleModel
{
	DECLARE_CUSTOMMODEL(UMT_CaseModel, UMT_PrimitiveRuleModel)
public:
	virtual bool SetSpecialPeerAttr();

private:
};

// Custom class for "Rule"
class UMT_RuleModel : public UMT_PrimitiveRuleModel
{
	DECLARE_CUSTOMMODEL(UMT_RuleModel, UMT_PrimitiveRuleModel)
public:
	virtual bool SetSpecialPeerAttr();

private:
};

// Custom class for "ExpressionRef"
class UMT_ExpressionRefReference : public CBuilderModelReference
{
	DECLARE_CUSTOMREF(UMT_ExpressionRefReference, CBuilderModelReference)
protected:
	GR::RewritingRuleProxy UdmPeer;
	Udm::Object UdmParent;
public:
	virtual GR::RewritingRuleProxy GetUdmPeer() { return UdmPeer; }
	virtual bool SetUdmPeer(GR::RewritingRuleProxy Peer) { UdmPeer = Peer; return true;}

	virtual Udm::Object GetUdmParent() { return UdmParent; }
	virtual bool SetUdmParent(Udm::Object Parent) {UdmParent = Parent; return true;}

	std::map<CBuilderReferencePort *, GR::ObjectWrapperProxy> childportmap;

public:
	virtual bool SetSpecialPeerAttr();

private:
};

// Custom class for "Ports"
class UMT_PortAtom: public CBuilderAtom
{
	DECLARE_CUSTOMATOM(UMT_PortAtom, CBuilderAtom)

protected:
	GR::ObjectWrapper UdmPeer;
	Udm::Object UdmParent;
public:
	virtual GR::ObjectWrapper GetUdmPeer() { return UdmPeer; }
	virtual bool SetUdmPeer(GR::ObjectWrapper Peer) { UdmPeer = Peer; return true;}

	virtual Udm::Object GetUdmParent() { return UdmParent; }
	virtual bool SetUdmParent(Udm::Object Parent) {UdmParent = Parent; return true;}

};
 
class UMT_PatternAssociation: public CBuilderConnection
{
	DECLARE_CUSTOMCONNECTION(UMT_PatternAssociation, CBuilderConnection)

protected:
	GR::ObjectLink UdmPeer;
	Udm::Object UdmParent;
public:
	virtual GR::ObjectLink GetUdmPeer() { return UdmPeer; }
	virtual bool SetUdmPeer(GR::ObjectLink Peer) { UdmPeer = Peer; return true;}

	virtual Udm::Object GetUdmParent() { return UdmParent; }
	virtual bool SetUdmParent(Udm::Object Parent) {UdmParent = Parent; return true;}
};

// Custom class for "PatternObject"
class UMT_PatternObjectReference: public CBuilderAtomReference
{
	DECLARE_CUSTOMREF(UMT_PatternObjectReference, CBuilderAtomReference)

protected:
	GR::ObjectWrapper UdmPeer;
	Udm::Object UdmParent;
public:
	virtual GR::ObjectWrapper GetUdmPeer() { return UdmPeer; }
	virtual bool SetUdmPeer(GR::ObjectWrapper Peer) { UdmPeer = Peer; return true;}

	virtual Udm::Object GetUdmParent() { return UdmParent; }
	virtual bool SetUdmParent(Udm::Object Parent) {UdmParent = Parent; return true;}

};

#endif


