/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/


// UdmApp.cpp: implementation of the CUdmApp class.
// This file was automatically generated as UdmApp.cpp
// by UDM Interpreter Wizard on Monday, May 13, 2002 13:45:42

// Tihamr Levendovszky 05-13-02

#include "stdafx.h"
#include <afxdlgs.h> // For CFileDialog
#include "resource.h"
#include "OldNewDialog.h"

//#include <iostream.h>
#include <atlbase.h>
#include <comdef.h>

#include <time.h>
#include "UdmStatic.h"
#include "UdmDom.h"
#include "UdmGME.h"
#include "UmlExt.h"
#include "UdmUtil.h"

#include "UdmApp.h"
#include "UdmConfig.h"

#include "UMLModelTransformer.h"
#include "LibraryUpdate.h"


extern void dummy(void); // Dummy function for UDM meta initialization

/*********************************************************************************/
/* Initialization function. The framework calls it before preparing the backend. */
/* Initialize here the settings in the config static object.					 */
/* Return 0 if successful.														 */
/*********************************************************************************/
int CUdmApp::Initialize()
{
	// TODO: Your initialization code comes here...
	return 0;
}



/* 
Remarks to CUdmApp::UdmMain(...):
0.	The p_backend points to an already open backend, and the framework 
	closes it automatically. DO NOT OPEN OR CLOSE IT!
	To commit changes use p_backend->CommitEditSequence().
	To abort changes use p_backend->AbortEditSequence().
	To save changes to a different file use p_backend->SaveAs() or p_backend->CloseAs().

1.	Focus is the currently open model.

2.	The possible values for param (from GME DecoratorLib.h component_startmode_enum):
	GME_MAIN_START			=   0,
	GME_BROWSER_START		=   1,
	GME_CONTEXT_START		=   2,
	GME_EMBEDDED_START		=   3,
	GME_MENU_START			=  16,
	GME_BGCONTEXT_START		=  18,
	GME_ICON_START			=  32,
	METAMODEL_CHECK_SYNTAX	= 101

 3. The framework catches all the exceptions and reports the error in a message box,
	clean up and close the transactions aborting the changes. You can override this 
	behavior by catching udm_exception. Use udm_exception::what() to form an error 
	message.
*/

/***********************************************/
/* Main entry point for Udm-based Interpreter  */
/***********************************************/

void CUdmApp::UdmMain(
					 Udm::DataNetwork* p_backend,		// Backend pointer(already open!)
					 Udm::Object focusObject,			// Focus object
					 set<Udm::Object> selectedObjects,	// Selected objects
					 long param)						// Parameters
{	
	// TODO: Your Code Comes Here...
	try
	{	

		COldNewDialog oldnewDlg;
		if(oldnewDlg.DoModal() != IDOK)
			return;
		
		// Get old and new libraries.
		string oldlibname, newlibname;

		oldlibname = oldnewDlg.m_oldlib;
		newlibname = oldnewDlg.m_newlib;
		
		UMLModelTransformer::RootFolder oldLibrary, newLibrary;
		vector<UMLModelTransformer::Package> oldlibs, newlibs;
		
		UMLModelTransformer::RootFolder rf = UMLModelTransformer::RootFolder::Cast(p_backend->GetRootObject());
		vector<UMLModelTransformer::RootFolder> libs = rf.RootFolder_kind_children();
		
		// Adding the root folder to the list of folders
		libs.push_back(rf);

		for(vector<UMLModelTransformer::RootFolder>::iterator i=libs.begin(); i!=libs.end(); ++i)
		{	UMLModelTransformer::RootFolder lib = *i;
			vector<UMLModelTransformer::Package> packs = lib.Package_kind_children();
			for(vector<UMLModelTransformer::Package>::iterator j=packs.begin(); j!=packs.end(); ++j)
			{	UMLModelTransformer::Package libpac = *j;
				string libname = libpac.name();
				if(libname == oldlibname)
				{	if(oldlibs.size() > 0 )
					{	throw udm_exception("Error: Two libraries exist with the same name: "+libname);
					}
					oldlibs.push_back(libpac);
				}
				else if(libname == newlibname)
				{	if(newlibs.size() > 0 )
					{	throw udm_exception("Error: Two libraries exist with the same name: "+libname);
					}
					newlibs.push_back(libpac);
				}
			}
		}
		
		if(oldlibs.size() != newlibs.size())
		{	throw udm_exception("Error: Could not find either old or new library");
		}

		if(oldlibs.size() == 0 && newlibs.size() == 0)
		{	// Trying to find the library based on lib names
			for(vector<UMLModelTransformer::RootFolder>::iterator i=libs.begin(); i!=libs.end(); ++i)
			{	UMLModelTransformer::RootFolder lib = *i;
				string libname = lib.name();
				if(libname == oldlibname)
				{	oldLibrary = lib;
				}
				else if(libname == newlibname)
				{	newLibrary = lib;
				}
			}

			if(oldLibrary != Udm::null && newLibrary != Udm::null)
			{	vector<UMLModelTransformer::Package> oldpacks = oldLibrary.Package_kind_children();
				for(vector<UMLModelTransformer::Package>::iterator j=oldpacks.begin(); j!=oldpacks.end(); ++j)
				{	UMLModelTransformer::Package oldpac = *j;
					string oldpacname = oldpac.name();
					vector<UMLModelTransformer::Package> newpacks = newLibrary.Package_kind_children();
					for(vector<UMLModelTransformer::Package>::iterator k=newpacks.begin(); k!=newpacks.end(); ++k)
					{	UMLModelTransformer::Package newpac = *k;
						string newpacname = newpac.name();
						if(oldpacname == newpacname)
						{	oldlibs.push_back(oldpac);
							newlibs.push_back(newpac);
						}
					}
				}

			}
		}
		
		if(oldlibs.size() == 0 || newlibs.size() == 0 || oldlibs.size() != newlibs.size())
		{	throw udm_exception("Error: Could not find either old or new library");
		}
		
		vector<UMLModelTransformer::Package>::iterator o,n;
		for(o=oldlibs.begin(), n=newlibs.begin(); o!=oldlibs.end() && n!=newlibs.end(); ++o, ++n)
		{	UMLModelTransformer::Package oldlib = *o;
			UMLModelTransformer::Package newlib = *n;

			LibraryUpdate updater(*p_backend, oldlib, newlib);
			updater.update();
			char id[30];
			_itoa_s(oldlib.uniqueId(),id,10);
			string oldname = "Old"+oldlibname+id;
			
			//newlib may be libraries and the set may fail.
			if (!newlib.isLibObject())
				newlib.name();

			//oldlib may be libraries and the set may fail.
			if (!oldlib.isLibObject())
				oldlib.name() = oldname;
		}

		AfxMessageBox("Update Complete");
		//////////////////////////////////////////////////////////////////////////////
		
	}
	catch(udm_exception &e)
	{
		AfxMessageBox(e.what());
	}
		
												
}

