#include "ModelDiff.h"

#include "UdmUtil.h"

using namespace std;

set<Uml::Class> assoClss;
set<Udm::Object> src_assoObjs, dst_assoObjs;

map<string,string> nameMap;
set<string> unnamedRefs;
//typedef namepair pair<Uml::Class,string> name_pair;

bool CompareName(Udm::Object srcObj, Udm::Object dstObj)
{
	map<string,string>::const_iterator nameIter = nameMap.find(srcObj.type().name());
	set<string>::const_iterator refIter = unnamedRefs.find(srcObj.type().name());
	if(nameIter == nameMap.end() && refIter == unnamedRefs.end())//special name not in the map, do original code
	{                            //just compare names
		if(UdmUtil::ExtractName(srcObj)== UdmUtil::ExtractName(dstObj))	return true;
		else return false;
	}
	else if(nameIter != nameMap.end()) //for those models like SFC that don't have names on everything, use 
	{    //another attribute as a surrogate for the name
		string surrogateName = nameIter->second;

		//could be any type, don't restrict ourselves to strings
		string srcString, dstString;
		__int64 srcInt, dstInt;
		double srcDouble, dstDouble;
		bool srcBool, dstBool;

		if(srcObj.GetStrValue(surrogateName, srcString))
		{
			if(dstObj.GetStrValue(surrogateName, dstString))// && srcString == dstString) 
			if(srcString == dstString) //dstString wasn't getting set before second check above
			{	
				return true;
			}
			else
			{
				return false;
			}
		}
		if(srcObj.GetIntValue(surrogateName, srcInt))
		{
			if(dstObj.GetIntValue(surrogateName, dstInt))// && srcInt == dstInt)
			if(srcInt == dstInt)
			{	
				return true;
			}
			else
			{
				return false;
			}
		}
		if(srcObj.GetRealValue(surrogateName, srcDouble))
		{
			if(dstObj.GetRealValue(surrogateName, dstDouble))// && srcDouble == dstDouble)
			if(srcDouble == dstDouble)
			{	
				return true;
			}
			else
			{
				return false;
			}
		}
		if(srcObj.GetBoolValue(surrogateName, srcBool))
		{
			if(dstObj.GetBoolValue(surrogateName, dstBool))// && srcBool == dstBool)
			if(srcBool == dstBool)
			{	
				return true;
			}
			else
			{
				return false;
			}
		}
	}
	else if(refIter != unnamedRefs.end())
	{
		//references should have exactly one referred object
		set<Udm::Object> srcReffed = srcObj.GetAdjacentUniqueObjects();
		set<Udm::Object> dstReffed = dstObj.GetAdjacentUniqueObjects();
		if(srcReffed.empty() || dstReffed.empty())
		{
			return false;
		}
		Udm::Object realSrc = *(srcReffed.begin());
		Udm::Object realDst = *(dstReffed.begin());
		
		return CompareName(realSrc,realDst);

	}
	return false;
}

bool CompareType(Udm::Object srcObj, Udm::Object dstObj)
{
	if(srcObj.type() != dstObj.type())
	{
		cout<<(string)srcObj.type().name()<<"\t"<<(string)dstObj.type().name()<<endl;
		return false;
	}
	else return true;
}

bool CompareAttributes(Udm::Object srcObj, Udm::Object dstObj)
{
	Uml::Class cls= srcObj.type();

	set< Uml::Attribute> attrs=cls.attributes();	
	
	// Adding parent attributes
	set< Uml::Attribute> aattrs=Uml::AncestorAttributes(cls);
	attrs.insert(aattrs.begin(),aattrs.end());

	for(set<Uml::Attribute>::iterator ai = attrs.begin();ai != attrs.end(); ai++) 
	{
		if(string(ai->type())=="String") 
		{
			if((string)ai->name()=="name" || (string)ai->name()=="position") continue;
			string svalue=srcObj.getStringAttr(*ai);
			string dvalue=dstObj.getStringAttr(*ai);
			if(svalue != dvalue) 
			{
				cout<<svalue<<"\t"<<dvalue<<endl;			
				return false;
			}
		}
		else if(string(ai->type())=="Integer")
		{
			if((long)srcObj.getIntegerAttr(*ai)!= (long)dstObj.getIntegerAttr(*ai)) 
			{
				cout<<(long)srcObj.getIntegerAttr(*ai)<<"\t"<<(long)dstObj.getIntegerAttr(*ai)<<endl;			
				return false;
			}
		}
		else if(string(ai->type())=="Boolean")
		{
			if(srcObj.getBooleanAttr(*ai)!= (long)dstObj.getBooleanAttr(*ai)) 
			{
				cout<<"True != False"<<endl;
				return false;
			}
		}
		else if(string(ai->type())=="Real")
		{
			if(srcObj.getRealAttr(*ai)!= (long)dstObj.getRealAttr(*ai))
			{
				cout<<(long)srcObj.getRealAttr(*ai)<<"\t"<<(long)dstObj.getRealAttr(*ai)<<endl;			
				return false;
			}
		}
	}

	return true;
}

bool GetAssociationEnds(Udm::Object& assoObj, Udm::Object& end1, Udm::Object& end2)
{
	Uml::Class assocls = assoObj.type();
	Uml::Association asso = assocls.association();
	if(asso==Udm::null) return false;
	set< ::Uml::AssociationRole> roles=asso.roles();
	set< ::Uml::AssociationRole>::iterator ir = roles.begin();
	string role1 = (*ir).name();
	ir++;
	string role2 = (*ir).name();
	end1 = assoObj.GetAssociationEnd(role1);
	end2 = assoObj.GetAssociationEnd(role2);

	return true;
}

bool CompareAssociationObjects()
{
	for(set<Udm::Object>::const_iterator i=src_assoObjs.begin();i!=src_assoObjs.end();++i)
	{
		Udm::Object src_assoObj = *i;
		Udm::Object s_end1, s_end2;
		if(!GetAssociationEnds(src_assoObj, s_end1, s_end2)) return false;

		Udm::Object dst_assoObj;
		for(set<Udm::Object>::const_iterator j=dst_assoObjs.begin();j!=dst_assoObjs.end();++j)
		{
			if(!CompareType(src_assoObj, *j)) continue;
			Udm::Object d_end1, d_end2;
			if(!GetAssociationEnds((Udm::Object)*j, d_end1, d_end2)) return false;
			if(CompareName(s_end1, d_end1) && CompareName(s_end2, d_end2)) 
			{
				dst_assoObj = *j; 
				break;
			}
		}
		if(dst_assoObj==Udm::null) return false;
		dst_assoObjs.erase(dst_assoObj);
		if(!CompareAttributes(src_assoObj, dst_assoObj)) return false;
	}
	if(!dst_assoObjs.empty()) return false;
	return true;
}

bool CompareSimpleAssociations(Udm::Object srcObj, Udm::Object dstObj)
{
	set<Udm::Object> sObjs = srcObj.GetAdjacentUniqueObjects();
	set<Udm::Object> dObjs = dstObj.GetAdjacentUniqueObjects();
	if(sObjs.size()!=dObjs.size()) return false;
	
	for(set<Udm::Object>::const_iterator i=sObjs.begin();i!=sObjs.end();++i)
	{
		Udm::Object sObj = *i;
		if(assoClss.find(sObj.type())!=assoClss.end()) continue;
		bool found = false;
		set<Udm::Object>::const_iterator j=dObjs.begin();
		while(j!=dObjs.end())
		{
			Udm::Object dObj = *j;
			j++;
			if(assoClss.find(dObj.type())!=assoClss.end()) continue;
			if(CompareName(sObj,dObj)) 
			{
				found = true;
				break;
			}
		}
		if(!found) return false;
	}
	
	return true;
}

bool CompareObjects(Udm::Object srcObj, Udm::Object dstObj)
{
//	cout<<UdmUtil::ExtractName(srcObj)<<"\t"<<UdmUtil::ExtractName(dstObj)<<endl;
	if(!CompareType(srcObj, dstObj)) return false;
	if(!CompareAttributes(srcObj, dstObj)) return false;
	if(!CompareSimpleAssociations(srcObj, dstObj)) return false;

	set<Udm::Object> srcChildrens = srcObj.GetChildObjects();
	set<Udm::Object> dstChildrens = dstObj.GetChildObjects();
	
	if(srcChildrens.size() != dstChildrens.size()) return false;

	set<Udm::Object>::const_iterator j;
	for(j=dstChildrens.begin();j!=dstChildrens.end();++j)
	{
		if(assoClss.find((*j).type())!=assoClss.end()) 
			dst_assoObjs.insert((*j));
	}

	for(set<Udm::Object>::const_iterator i=srcChildrens.begin();i!=srcChildrens.end();++i)
	{
		Udm::Object srcChild = *i;
		if(assoClss.find(srcChild.type())!=assoClss.end()) 
		{
			src_assoObjs.insert(srcChild);
			continue;
		}
		Udm::Object dstChild;
		for(j=dstChildrens.begin();j!=dstChildrens.end();++j)
		{
			if(dst_assoObjs.find(*j)!=dst_assoObjs.end()) continue;
			if(CompareName(srcChild, *j))
			{
				dstChild = *j;
				break;
			}
		}
		if(dstChild==Udm::null) return false;
		if(!CompareObjects(srcChild, dstChild)) return false;
	}	
	return true;
};

bool CompareObjects(Udm::Object srcObj, Udm::Object dstObj, Uml::Diagram dgr)
{
	set<Uml::Association> assocs = dgr.associations();
	for(set<Uml::Association>::const_iterator i=assocs.begin();i!=assocs.end();++i)
	{
		Uml::Association assoc = *i;
		Uml::Class assocls = assoc.assocClass();
		if(assocls!=Udm::null) assoClss.insert(assocls);
	}

	if(!CompareObjects(srcObj, dstObj)) return false;

	if(src_assoObjs.size()!=dst_assoObjs.size()) return false;

	return CompareAssociationObjects();
}

void AddNameSurrogate(string classname, string namesurrogate)
{
	//cout << "In " << classname << " use " << namesurrogate << " for name" << endl;
	nameMap[classname] = namesurrogate;
}

void AddUnnamedRef(string refname)
{
	//cout << "unnamed ref " << refname << endl;
	unnamedRefs.insert(refname);
}